/*
 * ocfsproc.c
 *
 * ocfs proc interface
 *
 * Copyright (C) 2002, 2004 Oracle.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 021110-1307, USA.
 *
 * Authors: Kurt Hackel, Sunil Mushran, Manish Singh, Wim Coekaerts
 */

#define OCFSPROC_PRIVATE_DECLS

#ifdef __KERNEL__
#include <ocfs.h>
extern __u32 comm_voting;
#else
#include <libocfs.h>
#endif


/* Tracing */
#define OCFS_DEBUG_CONTEXT      OCFS_DEBUG_CONTEXT_PROC

ocfs_proc_list  top_dir[] = {
	{ "version", NULL, ocfs_proc_version },
	{ "nodename", NULL, ocfs_proc_nodename },
	{ "globalctxt", NULL, ocfs_proc_globalctxt },
	{ "lockstat", NULL, ocfs_proc_dlm_stats },
#ifdef OCFS_LINUX_MEM_DEBUG
	{ "memallocs", NULL, ocfs_proc_memallocs },
#endif
	{ NULL, } };

ocfs_proc_list  sub_dir[] = {
	{ "nodenum", NULL, ocfs_proc_nodenum },
	{ "mountpoint", NULL, ocfs_proc_mountpoint },
	{ "statistics", NULL, ocfs_proc_statistics },
	{ "hashstat", NULL, ocfs_proc_hash_stats },
	{ "lockstat", NULL, ocfs_proc_dlm_stats },
	{ "device", NULL, ocfs_proc_device },
	{ "locktypes", NULL, ocfs_proc_lock_type_stats },
	{ "nodes", NULL, ocfs_proc_nodes },
	{ "guid", NULL, ocfs_proc_guid },
	{ "label", NULL, ocfs_proc_label },
	{ NULL, } };

/*
 * ocfs_proc_init()
 *
 */
int ocfs_proc_init (void)
{
	struct proc_dir_entry *parent = NULL;
	ocfs_proc_list *p;

	LOG_ENTRY ();

	parent = proc_mkdir ("fs/ocfs", 0);
	if (parent) {
		OcfsGlobalCtxt.proc_root_dir = parent;
		proc_symlink ("ocfs", 0, "fs/ocfs");
		for (p = top_dir; p->name; p++)
			create_proc_read_entry (p->name, 0, parent,
						p->read_proc, p->data);
	}

	LOG_EXIT ();
	return 0;
}				/* ocfs_proc_init */

/*
 * ocfs_proc_deinit()
 *
 */
void ocfs_proc_deinit (void)
{
	struct proc_dir_entry *parent = OcfsGlobalCtxt.proc_root_dir;
	ocfs_proc_list *p;

	LOG_ENTRY ();

	if (parent) {
		for (p = top_dir; p->name; p++)
			remove_proc_entry (p->name, parent);
		remove_proc_entry ("ocfs", 0);
		remove_proc_entry ("fs/ocfs", 0);
	}

	LOG_EXIT ();
	return;
}				/* ocfs_proc_deinit */

/*
 * ocfs_proc_calc_metrics()
 *
 */
static int ocfs_proc_calc_metrics (char *page, char **start, off_t off,
				   int count, int *eof, int len)
{
	LOG_ENTRY ();

	if (len <= off + count)
		*eof = 1;

	*start = page + off;

	len -= off;

	if (len > count)
		len = count;

	if (len < 0)
		len = 0;

	LOG_EXIT ();
	return len;
}				/* ocfs_proc_calc_metrics */

#ifdef OCFS_LINUX_MEM_DEBUG
/*
 * ocfs_proc_memallocs()
 *
 */
static int ocfs_proc_memallocs (char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int ret = 0;
	int proc_overflow = 0;
	struct list_head *iter;
	struct list_head *temp_iter;
	alloc_item *item;
	int len = 0;
        char *slabname;
	char *tmpstr = NULL;

	LOG_ENTRY ();

#define	MEMDBG_LEN		255
	tmpstr = ocfs_malloc(MEMDBG_LEN);
	if (!tmpstr) {
		LOG_ERROR_STATUS (-ENOMEM);
		goto bail;
	}

	sprintf (tmpstr, "%-8s  %-9s  %s\n", "Pointer", "Size/Slab", "Line:File");
	printk("%s", tmpstr);
	ret = sprintf ((char *) (page + len), "%s", tmpstr);
	len += ret;

	list_for_each_safe (iter, temp_iter, &OcfsGlobalCtxt.item_list) {
		if (len >= 4096)
                        proc_overflow = 1;
		item = list_entry (iter, alloc_item, list);
		switch (item->type) {
		    case SLAB_ITEM:
			if (item->u.slab == OcfsGlobalCtxt.oin_cache)
				slabname = "oin";
			else if (item->u.slab == OcfsGlobalCtxt.ofile_cache)
				slabname = "ofile";
			else if (item->u.slab == OcfsGlobalCtxt.lockres_cache)
				slabname = "lockres";
			else if (item->u.slab == OcfsGlobalCtxt.fe_cache)
				slabname = "fe";
			else
				slabname = "unknown";

			if (item->u.slab == OcfsGlobalCtxt.lockres_cache) {
				ocfs_lock_res *p = item->address;
				sprintf(tmpstr,
				       	"%08x  %9s  %-40s  %5d  %u.%u\n",
				       	item->address, slabname, item->tag,
				       	atomic_read(&p->lr_ref_cnt),
				       	HILO(p->sector_num));
			} else
				sprintf(tmpstr, "%08x  %9s  %-40s\n", item->address,
				       	slabname, item->tag);

			printk("%s", tmpstr);
			if (!proc_overflow) {
				ret = snprintf ((char *) (page + len),
					       	(4096 - len), "%s", tmpstr);
				len += ret;
			}
			break;
		    case KMALLOC_ITEM: 
		    case VMALLOC_ITEM:
                    default:
			sprintf(tmpstr, "%08x  %9d  %s\n", item->address,
			       	item->u.length, item->tag);
			printk("%s", tmpstr);
			if (!proc_overflow) {
				ret = snprintf ((char *) (page + len),
					       	(4096 - len), "%s", tmpstr);
				len += ret;
			}

			break;
		}
		if (ret < 0)
			proc_overflow = 1;
	}


	if (proc_overflow)
		LOG_ERROR_STR ("proc output truncated");

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

bail:
	ocfs_safefree (tmpstr);
	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_memallocs */
#endif

/*
 * ocfs_proc_globalctxt()
 *
 */
static int ocfs_proc_globalctxt(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int len = 0;
	int ret;

	LOG_ENTRY ();

	len += sprintf (page,       "lockres count: %d\n",
		       	atomic_read (&OcfsGlobalCtxt.cnt_lockres));
	len += sprintf (page + len, "comm voting  : %d\n", comm_voting);
	len += sprintf (page + len, "ip addr/port : %s/%d\n",
		       	OcfsGlobalCtxt.comm_info.ip_addr,
		       	OcfsGlobalCtxt.comm_info.ip_port);
	len += sprintf (page + len, "guid         : ");
	strncat (page + len, OcfsGlobalCtxt.guid.guid, GUID_LEN);
	len += GUID_LEN;
	strncat (page + len, "\n", 1);
	len++;

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_globalctxt */

/*
 * ocfs_proc_dlm_stats()
 *
 */
static int ocfs_proc_dlm_stats(char *page, char **start, off_t off,
			       int count, int *eof, void *data)
{
	int len;
	int ret;
	ocfs_super *osb;
	ocfs_dlm_stats *nrq;
	ocfs_dlm_stats *nrp;
	ocfs_dlm_stats *drq;
	ocfs_dlm_stats *drp;

	LOG_ENTRY ();

#define DLM_STATS_HDR	"%-25s %10s %10s %10s %10s %10s %10s %10s\n"
#define DLM_STATS	"%-25s %10d %10d %10d %10d %10d %10d %10d\n"

	osb = (ocfs_super *) data;
	if (osb) {
		nrq = &(osb->net_reqst_stats);
		nrp = &(osb->net_reply_stats);
		drq = &(osb->dsk_reqst_stats);
		drp = &(osb->dsk_reply_stats);
	} else {
		nrq = &(OcfsGlobalCtxt.net_reqst_stats);
		nrp = &(OcfsGlobalCtxt.net_reply_stats);
		drq = &(OcfsGlobalCtxt.dsk_reqst_stats);
		drp = &(OcfsGlobalCtxt.dsk_reply_stats);
	}

	len = sprintf (page, DLM_STATS_HDR, " ", "total", "okay", "etimedout",
		       "eagain", "efail", "enoent", "others");

	len += sprintf (page + len, DLM_STATS, "netdlm vote requests",
		        atomic_read (&nrq->total), atomic_read (&nrq->okay),
		        atomic_read (&nrq->etimedout), atomic_read (&nrq->eagain),
		        atomic_read (&nrq->efail), atomic_read (&nrq->enoent),
		        atomic_read (&nrq->def));

	len += sprintf (page + len, DLM_STATS, "netdlm vote responses",
		       	atomic_read (&nrp->total), atomic_read (&nrp->okay),
		       	atomic_read (&nrp->etimedout), atomic_read (&nrp->eagain),
		       	atomic_read (&nrp->efail), atomic_read (&nrp->enoent),
		       	atomic_read (&nrp->def));

	len += sprintf (page + len, DLM_STATS, "diskdlm vote requests",
		       	atomic_read (&drq->total), atomic_read (&drq->okay),
		       	atomic_read (&drq->etimedout), atomic_read (&drq->eagain),
		       	atomic_read (&drq->efail), atomic_read (&drq->enoent),
		       	atomic_read (&drq->def));

	len += sprintf (page + len, DLM_STATS, "diskdlm vote responses",
		       	atomic_read (&drp->total), atomic_read (&drp->okay),
		       	atomic_read (&drp->etimedout), atomic_read (&drp->eagain),
		       	atomic_read (&drp->efail), atomic_read (&drp->enoent),
		       	atomic_read (&drp->def));

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_dlm_stats */


/*
 * ocfs_proc_version()
 *
 */
static int ocfs_proc_version (char *page, char **start, off_t off,
			      int count, int *eof, void *data)
{
	int len;
	int ret;
	extern char *ocfs_version;
	extern char *ocfs_date;
	extern char *ocfs_md5;

	LOG_ENTRY ();
  
	len = sprintf (page, "%s %s (build %s)\n", ocfs_version, ocfs_date,
		       ocfs_md5);

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_version */

/*
 * ocfs_proc_nodenum()
 *
 */
static int ocfs_proc_nodenum (char *page, char **start, off_t off,
			      int count, int *eof, void *data)
{
	int len;
	int ret;
	ocfs_super *osb;

	LOG_ENTRY ();

	osb = (ocfs_super *) data;
	len = sprintf (page, "%d\n", osb->node_num);

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_nodenum */

/*
 * ocfs_proc_nodename()
 *
 */
static int ocfs_proc_nodename (char *page, char **start, off_t off,
			       int count, int *eof, void *data)
{
	int len;
	int ret;

	LOG_ENTRY ();

	strcpy (page, OcfsGlobalCtxt.node_name);
	strcat (page, "\n");
	len = strlen (page);

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_nodename */

/*
 * ocfs_proc_add_volume()
 *
 */
void ocfs_proc_add_volume (ocfs_super * osb)
{
	char dir[15];
	char lnk[10];
	struct proc_dir_entry *parent = NULL;
	ocfs_proc_list *p;

	LOG_ENTRY ();

	snprintf (dir, sizeof(dir), "%u_%u", MAJOR(osb->sb->s_dev),
		  MINOR(osb->sb->s_dev));

	parent = proc_mkdir (dir, OcfsGlobalCtxt.proc_root_dir);

	if (parent) {
		osb->proc_sub_dir = parent;
		for (p = sub_dir; p->name; p++)
			create_proc_read_entry (p->name, 0, parent,
						p->read_proc, (char *)osb);
		snprintf (lnk, sizeof(lnk), "%u", osb->osb_id);
		proc_symlink (lnk, OcfsGlobalCtxt.proc_root_dir, dir);
	}

	LOG_EXIT ();
	return;
}				/* ocfs_proc_add_volume */

/*
 * ocfs_proc_remove_volume()
 *
 */
void ocfs_proc_remove_volume (ocfs_super * osb)
{
	ocfs_proc_list *p;
	char dir[15];

	LOG_ENTRY ();

	if (osb->proc_sub_dir) {
		for (p = sub_dir; p->name; p++)
			remove_proc_entry (p->name, osb->proc_sub_dir);

		snprintf (dir, sizeof(dir), "%u", osb->osb_id);
		remove_proc_entry (dir, OcfsGlobalCtxt.proc_root_dir);

		snprintf (dir, sizeof(dir), "%u_%u", MAJOR(osb->sb->s_dev),
			  MINOR(osb->sb->s_dev));
		remove_proc_entry (dir, OcfsGlobalCtxt.proc_root_dir);
	}

	LOG_EXIT ();
	return;
}				/* ocfs_proc_remove_volume */

/*
 * ocfs_proc_mountpoint()
 *
 */
static int ocfs_proc_mountpoint (char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len;
	int ret;
	ocfs_super *osb;

	LOG_ENTRY ();

	osb = (ocfs_super *) data;

	strncpy (page, osb->vol_layout.mount_point, MAX_MOUNT_POINT_LEN);
	strcat (page, "\n");
	len = strlen (page);

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_mountpoint */

/*
 * ocfs_proc_statistics()
 *
 */
static int ocfs_proc_statistics (char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len;
	char *pubmap = NULL;
	ocfs_super *osb;
	ocfs_vol_layout *vol;
	int ret = 0, i;
	char *ptr;

	LOG_ENTRY ();

	osb = (ocfs_super *) data;
	vol = &(osb->vol_layout);

	pubmap = ocfs_malloc(100);
	if (!pubmap) {
		LOG_ERROR_STATUS (-ENOMEM);
		goto bail;
	}

	ptr = pubmap;
	for (i = 0; i < 32; i++) {
		if (osb->publ_map & (1 << i))
			ptr += sprintf (ptr, "%d ", i);
	}
	if (pubmap != ptr)
		*(ptr - 1) = '\0';

#define PROC_STATS		\
  "File open count = %d.%u\n"	\
  "Publish map = %s\n"		\
  "Cluster size = %u\n"		\
  "Volume size = %u.%u\n"	\
  "Failed large allocs = %u\n"	\
  "Retry large allocs = %u\n"	\
  "Disk heartbeat = %u\n"	\
  "Node timeout = %u\n"		\
  "Max miss count = %u\n"

	len = sprintf (page, PROC_STATS, HILO (osb->file_open_cnt), pubmap,
		       vol->cluster_size, HILO (vol->size),
		       osb->cluster_bitmap.failed, osb->cluster_bitmap.ok_retries,
		       vol->disk_hb, vol->hb_timeo, osb->max_miss_cnt);

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

bail:
	ocfs_safefree (pubmap);
	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_statistics */

/*
 * ocfs_proc_hash_stats()
 *
 */
static int ocfs_proc_hash_stats (char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len;
	int ret;
	ocfs_super *osb;

	LOG_ENTRY ();

	osb = (ocfs_super *) data;

	*page = '\0';
	ocfs_hash_stat (&(osb->root_sect_node), page, 1024);

	len = strlen (page);

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_hash_stats */

/*
 * ocfs_proc_device()
 *
 */
static int ocfs_proc_device (char *page, char **start, off_t off,
			     int count, int *eof, void *data)
{
	int len;
	int ret;
	ocfs_super *osb;

	LOG_ENTRY ();

	osb = (ocfs_super *) data;
	len = snprintf (page, sizeof(osb->dev_str), "%s\n", osb->dev_str);

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_device */

/*
 * ocfs_proc_lock_type_stats()
 *
 */
static int ocfs_proc_lock_type_stats (char *page, char **start, off_t off,
				      int count, int *eof, void *data)
{
	int len;
	int ret;
	ocfs_super *osb;
	ocfs_lock_type_stats *lts;

	LOG_ENTRY ();

#define LTS_STATS_HDR	"%18s %18s %18s %18s %18s\n"
#define LTS_STATS	"%18d %18d %18d %18d %18d\n"

	len = sprintf (page, LTS_STATS_HDR, "update_lock_state",
		       "make_lock_master", "disk_release_lock",
		       "break_cache_lock", "others");

	osb = (ocfs_super *) data;
	if (osb) {
		lts = &(osb->lock_type_stats);

		len += sprintf (page + len, LTS_STATS,
			       	atomic_read (&lts->update_lock_state),
			       	atomic_read (&lts->make_lock_master),
			       	atomic_read (&lts->disk_release_lock),
			       	atomic_read (&lts->break_cache_lock),
				atomic_read (&lts->others));
	}

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_lock_type_stats */

/*
 * ocfs_proc_nodes()
 *
 */
static int ocfs_proc_nodes (char *page, char **start, off_t off,
			    int count, int *eof, void *data)
{
	int len = 0;
	int i;
	int ret;
	ocfs_super *osb;
	ocfs_node_config_info *node;
	__u32 publ_map;
	char mount;

	LOG_ENTRY ();

	osb = (ocfs_super *) data;

	if (osb) {
		down (&(osb->cfg_lock));
		publ_map = (__u32)osb->publ_map;
		for (i = 0; i < OCFS_MAXIMUM_NODES; ++i, publ_map >>= 1) {
			node = osb->node_cfg_info[i];
			if (!node)
				continue;
			mount = (publ_map & 0x1) ? 'M' : ' ';
			len += sprintf (page + len,
				       	"%2d %c %-32s %-15s %-6d ",
				       	i, mount, node->node_name,
				       	node->ipc_config.ip_addr,
				       	node->ipc_config.ip_port);
			strncat (page + len, node->guid.guid, GUID_LEN);
			len += GUID_LEN;
			len += sprintf (page + len, "\n");
		}
		up (&(osb->cfg_lock));
	}

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_nodes */

/*
 * ocfs_proc_guid()
 *
 */
static int ocfs_proc_guid (char *page, char **start, off_t off,
			   int count, int *eof, void *data)
{
	int len;
	int ret;
	ocfs_super *osb;
	ocfs_vol_layout *vol;
	char *p;
	int i;

	LOG_ENTRY ();

	osb = (ocfs_super *) data;
	vol = &(osb->vol_layout);

	for (i = 0, p = page; i < MAX_VOL_ID_LENGTH; i++, p += 2)
		sprintf(p, "%02X", vol->vol_id[i]);
	*p = '\n'; ++p; *p = '\0';

	len = strlen (page);

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_guid */


/*
 * ocfs_proc_label()
 *
 */
static int ocfs_proc_label (char *page, char **start, off_t off,
			    int count, int *eof, void *data)
{
	int len;
	int ret;
	ocfs_super *osb;
	ocfs_vol_layout *vol;

	LOG_ENTRY ();

	osb = (ocfs_super *) data;
	vol = &(osb->vol_layout);

	len = snprintf (page, sizeof(vol->label), "%s\n", vol->label);

	ret = ocfs_proc_calc_metrics (page, start, off, count, eof, len);

	LOG_EXIT_LONG (ret);
	return ret;
}				/* ocfs_proc_label */

