/*
 * Standard kernel functions for Linux crash dumps.
 *
 * Created by: Matt Robinson (yakker@sourceforge.net)
 * Contributions from SGI, IBM, HP, MCL, and others.
 *
 * Copyright (C) 1999 - 2003 Silicon Graphics, Inc. All rights reserved.
 * Copyright (C) 2000 - 2002 TurboLinux, Inc.  All rights reserved.
 * Copyright (C) 2001 - 2002 Matt D. Robinson.  All rights reserved.
 * Copyright (C) 2002 Free Software Foundation, Inc. All rights reserved.
 *
 * This code is released under version 2 of the GNU GPL.
 */

/*
 * -----------------------------------------------------------------------
 *
 * DUMP HISTORY
 *
 * This dump code goes back to SGI's first attempts at dumping system
 * memory on SGI systems running IRIX.  A few developers at SGI needed
 * a way to take this system dump and analyze it, and created 'icrash',
 * or IRIX Crash.  The mechanism (the dumps and 'icrash') were used
 * by support people to generate crash reports when a system failure
 * occurred.  This was vital for large system configurations that
 * couldn't apply patch after patch after fix just to hope that the
 * problems would go away.  So the system memory, along with the crash
 * dump analyzer, allowed support people to quickly figure out what the
 * problem was on the system with the crash dump.
 *
 * In comes Linux.  SGI started moving towards the open source community,
 * and upon doing so, SGI wanted to take its support utilities into Linux
 * with the hopes that they would end up the in kernel and user space to
 * be used by SGI's customers buying SGI Linux systems.  One of the first
 * few products to be open sourced by SGI was LKCD, or Linux Kernel Crash
 * Dumps.  LKCD comprises of a patch to the kernel to enable system
 * dumping, along with 'lcrash', or Linux Crash, to analyze the system
 * memory dump.  A few additional system scripts and kernel modifications
 * are also included to make the dump mechanism and dump data easier to
 * process and use.
 *
 * As soon as LKCD was released into the open source community, a number
 * of larger companies started to take advantage of it.  Today, there are
 * many commmunity members that contribute to LKCD, and it continues to
 * flourish and grow as an open source project.
 *
 * -----------------------------------------------------------------------
 *
 * SYSTEM DUMP LAYOUT
 * 
 * System dumps are currently the combination of a dump header and a set
 * of data pages which contain the system memory.  The layout of the dump
 * (for full dumps) is as follows:
 *
 *             +-----------------------------+
 *             |     generic dump header     |
 *             +-----------------------------+
 *             |   architecture dump header  |
 *             +-----------------------------+
 *             |         page header         |
 *             +-----------------------------+
 *             |          page data          |
 *             +-----------------------------+
 *             |         page header         |
 *             +-----------------------------+
 *             |          page data          |
 *             +-----------------------------+
 *             |              |              |
 *             |              |              |
 *             |              |              |
 *             |              |              |
 *             |              V              |
 *             +-----------------------------+
 *             |        PAGE_END header      |
 *             +-----------------------------+
 *
 * There are two dump headers, the first which is architecture
 * independent, and the other which is architecture dependent.  This
 * allows different architectures to dump different data structures
 * which are specific to their chipset, CPU, etc.
 *
 * After the dump headers come a succession of dump page headers along
 * with dump pages.  The page header contains information about the page
 * size, any flags associated with the page (whether it's compressed or
 * not), and the address of the page.  After the page header is the page
 * data, which is either compressed (or not).  Each page of data is
 * dumped in succession, until the final dump header (PAGE_END) is
 * placed at the end of the dump, assuming the dump device isn't out
 * of space.
 *
 * This mechanism allows for multiple compression types, different
 * types of data structures, different page ordering, etc., etc., etc.
 * It's a very straightforward mechanism for dumping system memory.
 * -----------------------------------------------------------------------
 *
 * DUMP IMPLEMENTATION
 *
 * Dumps are implemented using a "start at the top and work your way
 * to the bottom" method.  The starting location of kernel memory is
 * determined, and each successive page is passed through to the
 * dump_add_page() function, which determines whether to compress the
 * page, throw it out, add the page header, etc.  This mechanism is
 * going to change over time as non-disruptive dumps are created, so
 * it is best to read through the code (it is commented pretty well),
 * and let the developers know if it isn't clear enough.  We believe
 * in well-documented, well-commented code.
 *
 * -----------------------------------------------------------------------
 *
 * DUMP TUNABLES
 *
 * This is the list of system tunables (via /proc) that are available
 * for Linux systems.  All the read, write, etc., functions are listed
 * here.  Currently, there are a few different tunables for dumps:
 *
 * dump_device (used to be dumpdev):
 *     The device for dumping the memory pages out to.  This is almost
 *     always the primary swap partition for disruptive dumps.
 *
 * dump_compress (used to be dump_compress_pages):
 *     This is the flag which indicates which compression mechanism
 *     to use.  This is a BITMASK, not an index (0,1,2,4,8,16,etc.).
 *     This is the current set of values:
 *
 *     0: DUMP_COMPRESS_NONE -- Don't compress any pages.
 *     1: DUMP_COMPRESS_RLE  -- This uses RLE compression.
 *     2: DUMP_COMPRESS_GZIP -- This uses GZIP compression.
 *
 * dump_level:
 *     The amount of effort the dump module should make to save
 *     information for post crash analysis.  This value is now
 *     a BITMASK value, not an index:
 *
 *     0:   Do nothing, no dumping. (DUMP_LEVEL_NONE)
 *
 *     1:   Print out the dump information to the dump header, and
 *          write it out to the dump_device. (DUMP_LEVEL_HEADER)
 *
 *     2:   Write out the dump header and all kernel memory pages.
 *          (DUMP_LEVEL_KERN)
 *
 *     4:   Write out the dump header and all kernel and user
 *          memory pages.  (DUMP_LEVEL_USED)
 *
 *     8:   Write out the dump header and all conventional/cached 
 *	    memory (RAM) pages in the system (kernel, user, free).  
 *	    (DUMP_LEVEL_ALL_RAM)
 *
 *    16:   Write out everything, including non-conventional memory
 *	    like firmware, proms, I/O registers, uncached memory.
 *	    (DUMP_LEVEL_ALL)
 *
 *     The dump_level will default to 1.
 *
 *      REMIND: How about we change these to 1, 3, 7, 15, 31.
 *	        If we reserve the bits for individual passes
 *		it's easer to test things like non-conventional
 *		memory dump on systems with limited disk space.
 *		If would be more consistant with dump_level being 
 *		a bitmask. We might also consider changeing the name 
 *		to 'dump_passes' to make it more clear that the bitmask 
 *		is not a index.
 * 
 *
 * dump_flags:
 *     These are the flags to use when talking about dumps.  There
 *     are lots of possibilities.  This is a BITMASK value, not an index.
 * 
 *     1:  Try to keep the system running _after_ we are done
 *         dumping -- for non-disruptive dumps.  (DUMP_FLAGS_NONDISRUPT)
 *
 * -----------------------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------
 *                      H E A D E R   F I L E S
 * -----------------------------------------------------------------------
 */

/* header files */
#include <asm/system.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/delay.h>
#include <linux/reboot.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/swap.h>
#include <asm/uaccess.h>
#include <linux/dump.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/file.h>
#include <linux/sysctl.h>
#include <linux/iobuf.h>
#include <linux/mman.h>
#include <linux/init.h>
#include <linux/ctype.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/utsname.h>
#include <linux/highmem.h>
#include <asm/hardirq.h>
#include <linux/version.h>
#include <asm/system.h>
#ifdef CONFIG_KDB
#include <linux/kdb.h>
#include <linux/kdbprivate.h>
#endif /* CONFIG_KDB */
#include <linux/mm_inline.h>



/*
 * -----------------------------------------------------------------------
 *                       D E F I N I T I O N S
 * -----------------------------------------------------------------------
 */
#define DUMP_MODULE_NAME "dump"

#if DUMP_DEBUG
#define DUMP_PREFIX __func__
#else
#define DUMP_PREFIX DUMP_MODULE_NAME
#endif


#ifdef CONFIG_KDB
#define DUMP_PRINT_FUNC(args...) do { kdb_nextline = 0; kdb_printf(args); } while (0)
#else
#define DUMP_PRINT_FUNC(args...) do { printk(args); } while(0)
#endif /* CONFIG_KDB */

#define DUMP_PRINTN(args...)  {						\
	DUMP_PRINT_FUNC("\n");							\
	DUMP_PRINTF(args);						\
}
#define DUMP_PRINTF(args...)  {						\
	DUMP_PRINT_FUNC("%s: ", DUMP_PREFIX);					\
	DUMP_PRINT_FUNC(args);							\
}
#define DUMP_PRINT(args...) {						\
	DUMP_PRINT_FUNC(args);							\
}

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) (((a) << 16) | ((b) << 8) | (c))
#endif

/* 
 * Handle printing of 64-bit values  
 *
 * NOTE: on ia64 %llx is recomend for ia32. 
 *       on RedHat 7.2 	%llx work in user space but not in the kernel.
 *	 Perhaps this is dependent on the kernel version.
 */
#if BITS_PER_LONG == 64
#define PU64X "%lx"
#else
#define PU64X "%Lx"
#endif


/*
 * -----------------------------------------------------------------------
 *                         V A R I A B L E S
 * -----------------------------------------------------------------------
 */

/* Dump tunables */
kdev_t dump_device;                /* the actual kdev_t device number      */
int dump_level;                    /* the current dump level               */
int dump_compress;                 /* whether to try to compress each page */
int dump_flags;                    /* whether to try to compress each page */

/* Other global fields */
char dumpdev_name[PATH_MAX];       /* the name of the dump device          */
void *dump_page_buf;               /* dump page buffer for memcpy()!       */
void *dump_page_buf_0;             /* dump page buffer returned by kmalloc */
int dump_blk_size;                 /* sector size for dump_device          */
int dump_blk_shift;                /* shift need to convert to sector size */
dump_header_t dump_header;         /* the primary dump header              */
dump_header_asm_t dump_header_asm; /* the arch-specific dump header        */
struct kiobuf *dump_iobuf;         /* kiobuf for raw I/O to disk           */
loff_t dump_fpos;                  /* the offset in the output device      */
int dump_mbanks;		   /* number of  physical memory banks     */
dump_mbank_t dump_mbank[MAXCHUNKS];/* describes layout of physical memory  */
long dump_unreserved_mem = 0;      /* Save Pages even if it isn't reserved */
long dump_unreferenced_mem = 0;    /* Save Pages even if page_count == 0   */
long dump_nonconventional_mem = 0;  /* Save non-conventional mem (firmware) */

static int dump_compress_none(char *old, int oldsize, char *new, int newsize);

static dump_compress_t dump_none_compression = {
	compress_type:	DUMP_COMPRESS_NONE,
	compress_func:	dump_compress_none,
};

/* our device operations and functions */
static int dump_open(struct inode *i, struct file *f);
static int dump_release(struct inode *i, struct file *f);
static int dump_ioctl(struct inode *i, struct file *f,
	unsigned int cmd, unsigned long arg);

static struct file_operations dump_fops = {
	open:		dump_open,
	release:	dump_release,
	ioctl:		dump_ioctl,
};

/* function pointers and prototypes */
int (*dump_compress_func)(char *old, int oldsize, char *new, int newsize);

/* proc entries */
static struct proc_dir_entry *dump_root;    /* /proc/sys/dump root dir     */
static struct proc_dir_entry *dump_dd;      /* dump_device tunable         */
static struct proc_dir_entry *dump_cp;      /* dump_compress tunable       */
static struct proc_dir_entry *dump_l;       /* dump_level tunable          */
static struct proc_dir_entry *dump_f;       /* dump_flags tunable          */

/* static variables                                                        */
static int dump_okay = FALSE;      	   /* can we dump out to disk?     */
static char dpcpage[DUMP_DPC_PAGE_SIZE];  /* buffer used for compression   */
static unsigned long dump_save_flags;     /* save_flags()/restore_flags()  */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,4))
static unsigned long dump_blocks[KIO_MAX_SECTORS];
#endif

/* for dumping from interrupt context (Fixme)  */
static int saved_irq_count; 	/* remember the current irq nesting level   */
static int saved_bh_count;	/* remember if we were in soft irq context  */

/* used for dump compressors */
static struct list_head dump_compress_list = LIST_HEAD_INIT(dump_compress_list);

/* external variables                                                      */
extern volatile int dumping_cpu;   /* cpu on which dump is progressing     */
extern volatile int dump_in_progress;         /* note when we're dumping   */
extern int panic_timeout;          /* time before reboot                   */
extern int *blk_size[];            /* device block size calculations       */
extern struct new_utsname system_utsname;     /* system information        */

#if !defined(irq_count) && !defined(bh_count)
extern irq_cpustat_t irq_stat[];
#endif

/* external functions                                                      */
extern void si_meminfo(struct sysinfo *);
extern void *kmalloc(size_t, int);
extern void kfree(const void *);
extern unsigned long simple_strtoul(const char *,char **,unsigned int);

/* lkcd info structure -- this is used by lcrash for basic system data     */
lkcdinfo_t lkcdinfo = {
	0,
	(sizeof(void *) * 8),
#if defined(__LITTLE_ENDIAN) 
	__LITTLE_ENDIAN,
#else
	__BIG_ENDIAN,
#endif
	0,
	PAGE_SHIFT,
	PAGE_SIZE,
	PAGE_MASK,
	PAGE_OFFSET,
	0
};

#if DUMP_DEBUG
void dump_bp(void) {}
#endif


/*
 * -----------------------------------------------------------------------
 *            / P R O C   T U N A B L E   F U N C T I O N S
 * -----------------------------------------------------------------------
 */

/*
 * Name: dump_read_proc()
 * Func: Read the proc data for dump tunables.
 */
static int
dump_read_proc(char *page, char **start, off_t off,
	int count, int *eof, void *data)
{
	int len;
	char *out = page;
	struct proc_dir_entry *p = (struct proc_dir_entry *)data;


	if (0 == strcmp(p->name, DUMP_LEVEL_NAME)) {
		out += sprintf(out, "%d\n", dump_level);
		len = out - page;
	} else if (0 == strcmp(p->name, DUMP_FLAGS_NAME)) {
		out += sprintf(out, "0x%x\n", dump_flags);
		len = out - page;
	} else if (0 == strcmp(p->name, DUMP_COMPRESS_NAME)) {
		out += sprintf(out, "%d\n", dump_compress);
		len = out - page;
	} else if (0 == strcmp(p->name, DUMP_DEVICE_NAME)) {
#if 0
		len = strlen(dumpdev_name);
		memcpy((void *)page, (const void *)dumpdev_name, len);
#else
		out += sprintf(out, "%s\n", dumpdev_name);
		len = out - page;
#endif
	} else {
		return (0);
	}
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0) return 0;
	} else {
		len = count;
	}
	*start = page + off;
	return (len);
}

/*
 * -----------------------------------------------------------------------
 *              C O M P R E S S I O N   F U N C T I O N S
 * -----------------------------------------------------------------------
 */

/*
 * Name: dump_compress_none()
 * Func: Don't do any compression, period.
 */
static int
dump_compress_none(char *old, int oldsize, char *new, int newsize)
{
	/* just return the old size */
	return (oldsize);
}

/*
 * -----------------------------------------------------------------------
 *                  U T I L I T Y   F U N C T I O N S
 * -----------------------------------------------------------------------
 */

/*
 * Name: dump_kernel_write()
 *
 * Func: Write out kernel information, check the device limitations,
 *       block sizes, etc.  The thing I don't like about this function
 *       is that most of it is a duplicate of what rw_raw_dev() does,
 *       but sct doesn't want rw_raw_dev() to do kernel kiobuf stuff, and
 *       brw_kiovec() isn't a good place to do device checks, since we'd
 *       have to check all this stuff anyway, and device block calculations
 *       are ugly in there for each call to the function.  So we do it all
 *       here, call brw_kiovec() on our own, and return the results.
 *
 * 	 Writes DUMP_BUFFER_SIZE bytes in page buffer
 *
 * Returns: number of bytes written or -ERRNO. 
 *	    At EOF it returns 0.
 */
static ssize_t
dump_kernel_write(int *eof)
{
	int		err = 0, iosize, i;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,4))
	unsigned long	*b = dump_blocks;
#else
	unsigned long	*b = dump_iobuf->blocks;
#endif
	unsigned long blocknr, blocks, limit;
	
	/* check the device size to make sure we are in-line */
	if (blk_size[MAJOR(dump_device)]) {
		limit = (((loff_t) blk_size[MAJOR(dump_device)][MINOR(dump_device)])
				<< BLOCK_SIZE_BITS) >> dump_blk_shift;
	} else {
		limit = INT_MAX;
	}

	/* not aligned? */
	if (dump_fpos & (dump_blk_size - 1)) {
		DUMP_PRINTN("invalid buffer alignment!");
		return (-EINVAL);
	}

	/* beyond the end of the device -- not really an error, per se */
	if ((dump_fpos >> dump_blk_shift) > limit) {
		DUMP_PRINTN("no space left on device!");
		DUMP_BP();
		return (0);
	}

	/* 
	 * If near the end of the dump device we set *eof (if ptr provided) to
	 * let the caller know that it's near the EOF on the dump device.
	 */
	if ( ((dump_fpos + (4 * DUMP_BUFFER_SIZE)) >> dump_blk_shift) > limit) {
		if(eof != NULL) {
			*eof = 1;
		}
		DUMP_BP();
	}
	/* 
	 * If near the end of the dump device we set *eof (if ptr provided) to
	 * let the caller know that it's near the EOF on the dump device.
	 */
	else if ( ((dump_fpos + (2 * DUMP_BUFFER_SIZE)) >> dump_blk_shift) > limit) {
		if(eof != NULL) {
			*eof = 1;
		}
		DUMP_BP();
	}

	/* reset the block number values */
	blocknr = dump_fpos >> dump_blk_shift;

	/* calculate the number of blocks */
	blocks = DUMP_BUFFER_SIZE >> dump_blk_shift;
	if (blocks > (KIO_MAX_SECTORS >> (dump_blk_shift - 9))) {
		blocks = KIO_MAX_SECTORS >> (dump_blk_shift - 9);
	}

	/* don't write more blocks than our max limit */
	if (blocks > limit - blocknr) {
		blocks = limit - blocknr;
	}

	/* make sure we have blocks to write! */
	if (!blocks) {
		DUMP_PRINTN("no blocks left on dump device!");
		return (0);
	}

	/* map the kernel kiobuf based on the iosize */
	iosize = blocks << dump_blk_shift;

	/* map the block numbers */
	for (i = 0; i < blocks; i++) {
		b[i] = blocknr++;
	}

	/* write out the data to disk */
	err = brw_kiovec(WRITE, 1, &dump_iobuf, dump_device, b, dump_blk_size);

	/* make sure our I/O size matches return from brw_kiovec() */
	if (err < 0) {
		DUMP_PRINTN("brw_kiovec() returned %d!", err);
		return (-EFAULT);
	}

	if (err != DUMP_BUFFER_SIZE) {
		DUMP_PRINTN("brw_kiovec() did not write out DUMP_BUFFER_SIZE!");
		return (-EFAULT);
	}

	/* err is the transferred bytes -- return those */
	dump_fpos += err;
	return (err);
}

/*
 * Name: dump_add_end_marker()
 * Func: add last dump page marker to the end of the dump buffer.
 */
static void
dump_add_end_marker(unsigned long *toffset, int flags)
{
	dump_page_t dp;

	/* set the end marker */
	dp.dp_address = dp.dp_size = 0x0;
	dp.dp_flags = flags;

#if DUMP_DEBUG >= 6
	dp.dp_byte_offset = dump_header.dh_num_bytes + DUMP_BUFFER_SIZE + PAGE_SIZE;
	dp.dp_page_index = dump_header.dh_num_dump_pages;
#endif

	/* clear the buffer and copy the page header */
	memcpy((void *)(dump_page_buf + *toffset),
		(const void *)&dp, sizeof(dump_page_t));
	*toffset += sizeof(dump_page_t);

#if DUMP_DEBUG >= 6
	dump_header.dh_num_bytes += sizeof(dump_page_t);
#endif
	dump_header.dh_num_dump_pages++;

	return;
}

#define PageInuse(page)		(!page_count((page)))

inline int kernel_page(struct page *p)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,15))
	if (PageReserved(p) == 0)
		return 0;
	return 1;
#else
	if (PageReserved(p) || (!PageLRU(p) && PageInuse(p)))
		return 1;
	else
		return 0;
#endif
}

inline int referenced_page(struct page *p)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,15))
	if(PageReserved(p) || page_count(p) == 0)
		return 0;
	return 1;
#else
	if (PageInuse(p)) {
		if (!PageReserved(p) && PageLRU(p))
			return 1;
		else
			return 0;
	} else
		return 0;
#endif
}

inline int unreferenced_page(struct page *p)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,15))
	if(PageReserved(p) ||  page_count(p) > 0)
		return 0;
	return 1;
#else
	if (!PageInuse(p)) {
		if (!PageReserved(p))
			return 1;
		else
			return 0;
	} else
		return 0;
#endif
}

long dump_add_page_debug = 0;
long dump_add_page_test_pattern = 0;

/*
 * Name: dump_add_page()
 * Func: Add a hardware page to the dump buffer.
 *
 * NB: page is a DUMP_PAGE_SIZE page, 
 *     not a system page (PAGE_SIZE).
 * 
 * REMIND:
 *	We should likely move some of this to
 *	arch specific code and/or clean it up.
 */
static int
dump_add_page(unsigned long page_index, unsigned long *toffset, int pass)
{
#if defined(CONFIG_X86_64)
	extern int __dump_page_valid(unsigned long);
#endif
#if defined(CONFIG_X86) || defined(CONFIG_ALPHA)
	extern int page_is_ram(unsigned long);
#endif
#ifdef CONFIG_PPC64
	extern int ppc64_page_is_ram(u64);
#endif
	unsigned long size;
	dump_page_t dp;
	void *vaddr;
	u64 paddr = (((u64)page_index) << DUMP_PAGE_SHIFT);
	struct page *p;

#if !defined(CONFIG_DISCONTIGMEM) && !defined(CONFIG_IA64)
	p = (struct page *) &(mem_map[page_index]);

	/*
	 * If the address is in highmem, map it before copying to
	 * the dump buffer.  In essence, the dump buffer is the
	 * bounce buffer.
	 */
#ifdef CONFIG_HIGHMEM
	if (PageHighMem(p)) {
		/*
		 * Since this can be executed from IRQ context,
		 * reentrance on the same CPU must be avoided:
		 */
		vaddr = kmap_atomic(p, KM_USER0);
	}
	else
#endif
#if defined(CONFIG_X86_64)
	if (!__dump_page_valid(page_index)) {
		goto ret1;
	}
	else
#endif
#if defined(CONFIG_X86) || defined(CONFIG_ALPHA)
	if (!page_is_ram(page_index)) {
		goto ret1;
	}
	else
#endif
#ifdef CONFIG_PPC64
	if (!ppc64_page_is_ram(paddr)) {
		/*
		 * This function is defined in dump_ppc64.c
		 */
		return (1);
	}
	else
#endif
	/* low memory */
	vaddr = page_address(p);

#else /*  CONFIG_DISCONTIGMEM || CONFIG_IA64 */
#if defined(CONFIG_X86_64)
	if (!__dump_page_valid(page_index)) {
		goto ret1;
	}
#endif
	vaddr = __va(paddr);
	p = virt_to_page(vaddr);
#endif
	dp.dp_address = paddr;
	dp.dp_flags = DUMP_DH_RAW;


#if DUMP_DEBUG >= 6
	/*
	 * Helpfull when looking at hexdump of /dev/vdump.
	 * 
	 *                                               Dump Header         Swap Header
	 */
	dp.dp_byte_offset = dump_header.dh_num_bytes + DUMP_BUFFER_SIZE + DUMP_HEADER_OFFSET;
	dp.dp_page_index = dump_header.dh_num_dump_pages;

	switch(pass) {
	    case 1:	break;
	    case 2:	goto ret1;					/* Already Dumped */
	    case 3:	goto ret1;					/* Already Dumped */
	    case 4:	break;
	}
#else
	/*
 	 * Selective dump:
	 *  	Some systems, have huge memories, NUMA for example, where
	 *	a dump of all of memory isn't likely to fit on a swap partition.
	 *	This is a simple 1st attempt at ordering the dump so the most 
	 *	important pages are dumped first.
	 *
	 *	pass1: Kernel Pages
	 *	pass2: Remaining referenced pages
	 *	pass3: the rest of conventional memory
	 *	pass4: non-conventional memory,
	 */
	switch (pass) {
	case 1:
		if (kernel_page(p))
			break;
		else
			goto ret1;
	case 2: 
		if (referenced_page(p))
			break;
		else
			goto ret1;
	case 3:
		if (unreferenced_page(p))
			break;
		else
			goto ret1;
	case 4:
		break;

	}
#endif

#if DUMP_DEBUG
	if(dump_add_page_debug) {
		printk(KERN_ALERT "dump_add_page(page_index:%lx, *toffset:%lx):\n",  
				                 page_index,     *toffset);
	}
#endif

	/*
	 * Don't compress the page if any part of it overlaps
	 * with the current task_struct or current stack.
	 */
	if ((((unsigned long)vaddr < (unsigned long)current + THREAD_SIZE)  &&
		((unsigned long)vaddr + DUMP_PAGE_SIZE >
		(unsigned long)current))) {
			size = DUMP_PAGE_SIZE;
	}
	else {
		size = dump_compress_func((char *)vaddr, DUMP_PAGE_SIZE,
			(char *)dpcpage, DUMP_DPC_PAGE_SIZE);
		if (!size) {
			/* dump compression failed -- default to raw mode */
			size = DUMP_PAGE_SIZE;
		}
		/* set the compressed flag if the page did compress */
		if (size < DUMP_PAGE_SIZE) {
			dp.dp_flags = DUMP_DH_COMPRESSED;
		}
	}

	/* set the universal size */
	dp.dp_size = size;

	/* copy the page header */
	memcpy((void *)(dump_page_buf + *toffset), (const void *)&dp,
		sizeof(dump_page_t));

	*toffset += sizeof(dump_page_t);

	if (dp.dp_flags & DUMP_DH_COMPRESSED) {
		/* copy the compressed page */
		memcpy((void *)(dump_page_buf + *toffset),
			(const void *)dpcpage, size);
	} else {
		/* copy directly from memory */
		DUMP_memcpy((void *)(dump_page_buf + *toffset),
			(const void *)vaddr, size);
	}

#if DUMP_DEBUG
	/*
	 * Write a Test Pattern:
	 *	Page data is full of bytes with the page number.
	 */	
	if (dump_add_page_test_pattern) {
		unsigned char num_pages = (char) (dump_header.dh_num_dump_pages & 0Xff);
		char *cp = (dump_page_buf + *toffset);
		int i;

		for( i = 0; i < size; i++) {
			*cp++ = num_pages;	
		}
	}
#endif

#ifdef CONFIG_HIGHMEM
	if (PageHighMem(p)) {
		/*
		 * Since this can be executed from IRQ context,
		 * reentrance on the same CPU must be avoided:
		 */
		kunmap_atomic(vaddr, KM_USER0);
	}
#endif
	*toffset += size;

#if DUMP_DEBUG >= 6
	if (dump_add_page_debug) {
		printk(KERN_ALERT "dump_add_page: *toffset:%lx += size:%lx:\n",  
						  *toffset,       size);
	}
	dump_header.dh_num_bytes += (size + sizeof(dump_page_t));
#endif

	dump_header.dh_num_dump_pages++;

	return (0);

ret1:
#ifdef CONFIG_HIGHMEM
	if (PageHighMem(p)) {
		kunmap_atomic(vaddr, KM_USER0);
	}
#endif
	return 1;
}


/*
 * Name: dump_silence_system()
 * Func: Silence the system, or make CPUs spin, etc.  The intention
 *       here is to get things to a "quiet" state, so we can dump.
 */
void
dump_silence_system(void)
{
	unsigned int stage = 0;
	int cpu = smp_processor_id();

#ifdef CONFIG_KDB
	if (KDB_IS_RUNNING()) {
		/* kdb is in control, the system is already silenced */
		kdb_printf("LKCD entered from KDB\n");
	}
#endif /* CONFIG_KDB */

	if (in_interrupt()) {
		printk(KERN_ALERT "Dumping from interrupt handler !\n");
		printk(KERN_ALERT "Uncertain scenario - but will try my best\n");
		/* 
		 * Must be an unrelated interrupt, not in the middle of io ! 
		 * If we've panic'ed in the middle of io we should take 
		 * another approach 
		 */
	}

#ifdef CONFIG_KDB
	if (!KDB_IS_RUNNING())
#endif /* CONFIG_KDB */
	/* see if there's something to do before we re-enable interrupts */
	(void)__dump_silence_system(stage);

	/* we set this to FALSE so we don't ever re-enter this code! */
	dump_okay = FALSE;
	dumping_cpu = cpu;
	dump_in_progress = TRUE;
	save_flags(dump_save_flags);

	/* -------------------------------------------------- */
	/* Kludge - dump from interrupt context is unreliable (Fixme)
	 *
	 * We do this so that softirqs initiated for dump i/o 
	 * get processed and we don't hang while waiting for i/o
	 * to complete or in any irq synchronization attempt.
	 *
	 * This is not quite legal of course, as it has the side 
	 * effect of making all interrupts & softirqs triggered 
	 * while dump is in progress complete before currently 
	 * pending softirqs and the currently executing interrupt 
	 * code. 
	 */
	/*
	 * We use irq_stat directly instead of using local_irq_count
	 * and local_bh_count as they can't be used as Lvalues on
	 * uniprocessor machines. 
	 */
#ifdef __powerpc64__
	saved_irq_count = local_irq_count(cpu);
	saved_bh_count =  local_bh_count(cpu);
	local_irq_count(cpu) = 0;
	local_bh_count(cpu) = 0;
#else
#if defined(irq_count) && defined(bh_count)
	saved_irq_count = irq_count(cpu);
	saved_bh_count =  bh_count(cpu);
	irq_count(cpu) = 0;
	bh_count(cpu) = 0;
#else
	saved_irq_count = irq_stat[cpu].__local_irq_count;
	saved_bh_count = irq_stat[cpu].__local_bh_count;
	irq_stat[cpu].__local_irq_count = 0;
	irq_stat[cpu].__local_bh_count = 0;
#endif
#endif
	/* -----------------------------------------------------*/

	sti(); /* enable interrupts just in case ... */

	/* now increment the stage and do stuff after interrupts are enabled */
	stage++;

#ifdef CONFIG_KDB
	if (!KDB_IS_RUNNING())
#endif /* CONFIG_KDB */
	(void)__dump_silence_system(stage);

	/* time to leave */
	return;
}

/*
 * Name: dump_resume_system()
 * Func: Resume the system state.
 */
void
dump_resume_system(void)
{
	unsigned int stage = 0;

	/* let the architectures return their proper system state */
	(void)__dump_resume_system(stage);

	/* restore flags and other dump state fields */
	restore_flags(dump_save_flags);

#ifdef __powerpc64__
	local_irq_count(dumping_cpu) =  saved_irq_count;
	local_bh_count(dumping_cpu) = saved_bh_count;
#else
#if defined(irq_count) && defined(bh_count)
	irq_count(dumping_cpu) =  saved_irq_count;
	bh_count(dumping_cpu) = saved_bh_count;
#else
	irq_stat[dumping_cpu].__local_irq_count = saved_irq_count;
	irq_stat[dumping_cpu].__local_bh_count = saved_bh_count;
#endif
#endif

	dump_okay = TRUE;

	/* reboot the system if this isn't a disrupted dump */
	if ((panic_timeout > 0) && (!(dump_flags & DUMP_FLAGS_NONDISRUPT))) {
		DUMP_PRINTF("Dump: Rebooting in %d seconds ...", panic_timeout);
		mdelay(panic_timeout * 1000);
		machine_restart(NULL);
	}
	dump_in_progress = FALSE;

	return;
}

static void
dump_speedo(void)
{
	static int i = 0;

	switch (++i%4) {
	case 0:
		printk("|\b");
		break;
	case 1:
		printk("\\\b");
		break;
	case 2:
		printk("-\b");
		break;
	case 3:
		printk("/\b");
		break;
	}
}

/*
 * -----------------------------------------------------------------------
 *                     H E A D E R   F U N C T I O N S
 * -----------------------------------------------------------------------
 */

/*
 * Name: dump_configure_header()
 * Func: Update the header with the appropriate information.
 */
static int
dump_configure_header(char *panic_str, struct pt_regs *regs)
{
	struct timeval dh_time;
	/* make sure the dump header isn't TOO big */
	if ((sizeof(dump_header_t) +
		sizeof(dump_header_asm_t)) > DUMP_BUFFER_SIZE) {
			DUMP_PRINTN("dump_configure_header(): combined "
				"headers larger than DUMP_BUFFER_SIZE!");
			return (0);
	}

	/* configure dump header values */
	dump_header.dh_magic_number = DUMP_MAGIC_NUMBER;
	dump_header.dh_version = DUMP_VERSION_NUMBER;
	dump_header.dh_memory_start = PAGE_OFFSET;
	dump_header.dh_memory_end = DUMP_MAGIC_NUMBER;
	dump_header.dh_header_size = sizeof(dump_header_t);
	dump_header.dh_page_size = PAGE_SIZE;
	dump_header.dh_dump_level = dump_level;
	dump_header.dh_current_task = (unsigned long) current;
	dump_header.dh_dump_compress = dump_compress;
	dump_header.dh_dump_flags = dump_flags;
	dump_header.dh_dump_device = dump_device;

#if DUMP_DEBUG >= 6
	dump_header.dh_num_bytes = 0;
#endif
	dump_header.dh_num_dump_pages = 0;
	do_gettimeofday(&dh_time);
	dump_header.dh_time.tv_sec = dh_time.tv_sec;
	dump_header.dh_time.tv_usec = dh_time.tv_usec;

#ifndef UTSNAME_ENTRY_SZ
#define UTSNAME_ENTRY_SZ 65
#endif
	memcpy((void *)&(dump_header.dh_utsname_sysname), 
		(const void *)&(system_utsname.sysname), UTSNAME_ENTRY_SZ);
	memcpy((void *)&(dump_header.dh_utsname_nodename), 
		(const void *)&(system_utsname.nodename), UTSNAME_ENTRY_SZ);
	memcpy((void *)&(dump_header.dh_utsname_release), 
		(const void *)&(system_utsname.release), UTSNAME_ENTRY_SZ);
	memcpy((void *)&(dump_header.dh_utsname_version), 
		(const void *)&(system_utsname.version), UTSNAME_ENTRY_SZ);
	memcpy((void *)&(dump_header.dh_utsname_machine), 
		(const void *)&(system_utsname.machine), UTSNAME_ENTRY_SZ);
	memcpy((void *)&(dump_header.dh_utsname_domainname), 
		(const void *)&(system_utsname.domainname), UTSNAME_ENTRY_SZ);

	if (panic_str) {
		memcpy((void *)&(dump_header.dh_panic_string),
			(const void *)panic_str, DUMP_PANIC_LEN);
	}

        dump_header_asm.dha_magic_number = DUMP_ASM_MAGIC_NUMBER;
        dump_header_asm.dha_version = DUMP_ASM_VERSION_NUMBER;
        dump_header_asm.dha_header_size = sizeof(dump_header_asm_t);

	/* copy the registers if they are valid */
	if (regs) {
		memcpy((void *)&(dump_header_asm.dha_regs),
			(const void *)regs, sizeof(struct pt_regs));
	}

	/* configure architecture-specific dump header values */
	if (!__dump_configure_header(regs)) {
		return (0);
	}
	return (1);
}

/*
 * Name: dump_write_header()
 * Func: Write out the dump header.
 *
 * Returns:
 *	-1: failed
 *	 0: wrote o bytes (truncated)
 *	+1: wrote header
 */
static int
dump_write_header(void)
{
	int state;
	loff_t toffset;

	/* clear the dump page buffer */
	memset(dump_page_buf, 0, DUMP_BUFFER_SIZE);

	/* copy the dump header directly into the dump page buffer */
	memcpy(dump_page_buf, (const void *)&dump_header,
		sizeof(dump_header_t));

	memcpy((void *)(dump_page_buf + sizeof(dump_header_t)),
		(const void *)&dump_header_asm, sizeof(dump_header_asm_t));

	/* save our file pointer */
	toffset = dump_fpos;

	/* 
	 * ALWAYS write out the dump header at DUMP_HEADER_OFFSET, 
	 * this is after the longest swap header possibly written by mkswap;
	 * likely the largest PAGE_SIZE supported by the archecture.
	 */
	dump_fpos = DUMP_HEADER_OFFSET;

	/* do the real write here */
	state = dump_kernel_write(NULL);
	dump_fpos = toffset;
	if (state < 0) {
		DUMP_PRINTF("dump_kernel_write() failed!\n");
		return (-1);
	} else if (!state) {
		/* wrote zero bytes - failed */
		return (0);
	}
	/* Wrote the header - success */
	return (1);
}

/*
 * -----------------------------------------------------------------------
 *                  E X E C U T E   F U N C T I O N S
 * -----------------------------------------------------------------------
 */

/*
 * Name: dump_execute_memdump()
 * Func: Perform the actual memory dump.  This walks through the
 *       memory pages and dumps the data to disk (using other functions).
 *
 * Returns:
 *	-1: failed
 *	 0: dump truncated	
 *	 1: success
 */
static int
dump_execute_memdump(void)
{
	int counter = 0, n_bytes = 0, i;
	unsigned long buf_loc;
	u64 mem_loc;
	int dump_truncated = 0;
	int pass = 0;

	DUMP_PRINTN("Compression value is 0x%x, Writing dump header ", dump_compress);

	/* update the header to disk the first time */
	n_bytes = dump_write_header();
	DUMP_PRINT("\n");

	if (n_bytes < 0) {
		DUMP_PRINTF("Initial dump header update failed!\n");
		return (-1);
	} else if (n_bytes == 0) {
		return (0);
	}

	/* if we only want the header, return */
	if (dump_level & DUMP_LEVEL_HEADER) {
		return (1);
	}

	/* set beginning offset to keep compats right with swap devices */
	dump_fpos = DUMP_HEADER_OFFSET + DUMP_BUFFER_SIZE;

	/* clear the dump page buffer */
	memset(dump_page_buf, 0, DUMP_BUFFER_SIZE);

	/* set the buffer location */
	buf_loc = 0;

	for (pass = 1; pass <= 4; pass++) {

		if( (pass == 2) && dump_unreserved_mem == 0) {
			 continue;
		}
		if( (pass == 3) && dump_unreferenced_mem == 0) {
			 continue;
		}
		if (pass == 4) { 
			if (dump_nonconventional_mem == 0) break;
			if (dump_mbanks == 1) break;				/* Contigious Mem Sys have 1 mbank */
		} 
		/* Algorithm suggested by Jack Steiner <steiner@sgi.com */
		switch(pass) {
			case 1:	DUMP_PRINTN("Pass 1: Saving Kernel Pages: "); 					break;
			case 2:	DUMP_PRINTN("Pass 2: Saving Remaining Referenced Pages: "); 			break;
			case 3:	DUMP_PRINTN("Pass 3: Saving Remaining Unreferenced Pages: ");			break;
			case 4:	DUMP_PRINTN("Pass 4: Saving Unconventional Memory: ");				break;
		}
		for (i = 0; i < dump_mbanks; i++) {
		  u64 mem_bank_start = dump_mbank[ i ].start;
		  u64 mem_bank_end = dump_mbank[ i ].end;
		  int type = dump_mbank[ i ].type;

		  if ( (type != DUMP_MBANK_TYPE_CONVENTIONAL_MEMORY) && (pass != 4) ) continue;
		  if ( (type == DUMP_MBANK_TYPE_CONVENTIONAL_MEMORY) && (pass == 4) ) continue;
	
		  DUMP_PRINTN("Memory Bank[%d]: " PU64X " ... " PU64X ": ", i, mem_bank_start, mem_bank_end);
		  for (mem_loc = mem_bank_start; mem_loc < mem_bank_end; mem_loc += DUMP_PAGE_SIZE) {
	
			/* add the page (if it's real RAM) */
			if (dump_add_page((mem_loc >> DUMP_PAGE_SHIFT), 
						&buf_loc, pass)) {
				/* didn't add a page to buffer */
				continue;
			}
	
			/* see if we've filled the buffer */
			if (buf_loc >= DUMP_BUFFER_SIZE) {
				int near_eof = 0;
	
				/* write out the dump page buffer */
				n_bytes = dump_kernel_write(&near_eof);
				if (n_bytes < 0) {
					DUMP_PRINTN("Write of dump pages failed!");
					return(-1);
				} else if (n_bytes == 0) {
					DUMP_PRINTN("EOF on Write of dump pages; dump terminated!");
					return(0);
				}
				if( near_eof ) {
					DUMP_PRINTN("Near EOF on Write of dump pages; truncating dump,");
					dump_truncated = 1;
				}
	
				/* bump the counter for writing out the header */
				counter++;
	
				/*
				 * Update the header every once in a while -- this
				 * _must_ be done before we write the overflow end of
				 * the dump_page_buf into the top of dump_page_buf,
				 * as dump_write_header() uses dump_page_buf to
				 * write the default/asm dump headers.  After we are
				 * done updating the header, _then_ we can move the
				 * leftover dump data into the top of dump_page_buf.
				 *
				 * NOTE: a period is printed to indicate that the header
				 *	 was updated. Speedo shows page outactivity.
				 */
				if ((counter & 0x3f) == 0) {
					n_bytes = dump_write_header();
					if (n_bytes < 0) {
						DUMP_PRINTN("Dump header update failed!");
						return (-1);
					} else if (n_bytes == 0) {
						DUMP_PRINTN("Dump header update failed; bizzare");
						return (0);
					}
					DUMP_PRINT(".");
				} else {
					if ((counter & 0x07) == 0) {
						dump_speedo();
					}
				}
	
				/* clear the dump page buffer */
				memset(dump_page_buf, 0, DUMP_BUFFER_SIZE);
	
				/* adjust leftover data back to the top of the page */
				if (buf_loc > DUMP_BUFFER_SIZE) {
	        			/* copy the dump page buffer remnants */
	        			memcpy((void *)dump_page_buf,
	                			(const void *)(dump_page_buf +
							DUMP_BUFFER_SIZE),
	                			buf_loc - DUMP_BUFFER_SIZE);
	
	        			/* set the new buffer location */
	        			buf_loc -= DUMP_BUFFER_SIZE;
				} else {
					/* reset the buffer location counter */
					buf_loc = 0;
				}
			}
			if (dump_truncated) {
				break;
			}
		    }
		    DUMP_PRINT(" ");	
		    if (dump_truncated) {
			break;
		    }
		}
		if (dump_truncated) {
			break;
		}
		DUMP_PRINT("\n");
	}

	/* 
	 * we have written out most of the dump pages, a few may still be in the
	 * page buffer and it's possible we had to truncate the dump because we
	 * got very close to the end of the dump partition. Now we add a EOF
	 * marker to the page out buffer and flush out the remaining pages.
	 *
	 * We need to write the DUMP_DH_END even for truncated data because
	 * the reader won't get an EOF if he reads the data in smaller chunks
	 * than the DUMP_BUFFER_SIZE it's written in.
	 */
	if (dump_truncated) {
		dump_add_end_marker(&buf_loc, (DUMP_DH_END | DUMP_DH_TRUNCATED));
	} else {
		dump_add_end_marker(&buf_loc, DUMP_DH_END);
	}
	n_bytes = dump_kernel_write(NULL);

	if (n_bytes < 0) {
                DUMP_PRINTN("Final write of last of page buffer and DUMP_DH_END failed!");
                return (-1);
        } else if (n_bytes == 0) {
		DUMP_PRINTN("Hit EOF writing DUMP_DH_END; bad luck\n");
                return (0);
        }

	/*
	 * Writing out DUMP_DH_END may have pushed us past the end of the first
	 * part of the page buffer, if it did we have to write out one last 
	 * DUMP_BUFFER with the spill over page.
	 */
	if (buf_loc > DUMP_BUFFER_SIZE) {
		/* clear the first part of the buffer */
		memset(dump_page_buf, 0, DUMP_BUFFER_SIZE);

		/*
		 * Copy the dump page buffer remnants in the second 
		 * part of the buffer to the first part.
		 */
		memcpy((void *)dump_page_buf,
			(const void *)(dump_page_buf + DUMP_BUFFER_SIZE),
			buf_loc - DUMP_BUFFER_SIZE);

		n_bytes = dump_kernel_write(NULL);
		if (n_bytes < 0) {
			DUMP_PRINTN("Final write of spillover page failed!");
			return (-1);
		} else if (n_bytes == 0) {
			DUMP_PRINTN("Hit EOF writing spill over page with DUMP_DH_END; very bad luck!\n");
			return (0);
		}
	}
	if (dump_truncated) {
		return(0);
	}
	/* success */
	return (1);
}

/*
 * Name: dump_execute()
 * Func: Execute the dumping process.  This makes sure all the appropriate
 *       fields are updated correctly, and calls dump_execute_memdump(),
 *       which does the real work.
 * 
 * if( dump_flags & DUMP_FLAGS_NONDISRUPT ) {
 *    Returns:
 *      -1: failed
 *       0: dump truncated
 *       1: success
 * }
 */
int
dump_execute(char *panic_str, struct pt_regs *regs)
{
	int state = 1;

	/* make sure we can dump */
	if (dump_okay == FALSE) {
		return(-1);
	}

	if(!dump_configure_header(panic_str, regs)) {
		DUMP_PRINTN("dump header could not be configured!");
		return(-1);
	}

	/* silence the system */
	dump_silence_system();

	/* tell interested parties that a dump is happening */
	notifier_call_chain(&dump_notifier_list, DUMP_BEGIN, &dump_device);

	/* bail out if we're not going to do any dumping */
	if (dump_level != DUMP_LEVEL_NONE) {
		/* inform users of what we are about to do */
		DUMP_PRINTN("Dumping to device 0x%x [%s] on CPU %d ...",
			dump_device, bdevname(dump_device),
			smp_processor_id());

		/* start walking through the page tables */
		state = dump_execute_memdump();

		DUMP_PRINT("\n");

		/* update header to disk for the last time */
		if (dump_write_header() < 0) {
			DUMP_PRINTF("Final dump header update failed!\n");
		}

		if (state < 0) {
			DUMP_PRINTF("Dump Failed!\n");
		} else if (state == 0) {
			DUMP_PRINTF("Dump Truncated (likely out of space).\n");
		} else {
			DUMP_PRINTF("Dump Complete; %d dump pages saved.\n", dump_header.dh_num_dump_pages);
		}
	}

	/* tell interested parties that a dump has completed */
	notifier_call_chain(&dump_notifier_list, DUMP_END, &dump_device);

	/* put the system state back */
	dump_resume_system();

	return (state);
}

/*
 * Name: dump_register_compression()
 * Func: Register a dump compression mechanism.
 */
void
dump_register_compression(dump_compress_t *item)
{
	/* let's make sure our list is valid */
	if (!item) {
		return;
	}

	/* add our item */
	list_add(&(item->list), &dump_compress_list);

	/* print information to callers */
	DUMP_PRINTF("Registering dump compression type 0x%x\n", item->compress_type);
}

/*
 * Name: dump_unregister_compression()
 * Func: Remove a dump compression mechanism, and re-assign the dump
 *       compression pointer if necessary.
 */
void
dump_unregister_compression(int compression_type)
{
	struct list_head *tmp;
	dump_compress_t *dc;

	/* let's make sure our list is valid */
	if (compression_type == DUMP_COMPRESS_NONE) {
		DUMP_PRINTN("Compression list is invalid!");
		return;
	}

	/* try to remove the compression item */
	list_for_each(tmp, &dump_compress_list) {
		dc = list_entry(tmp, dump_compress_t, list);
		if (dc->compress_type == compression_type) {
			list_del(&(dc->list));
			DUMP_PRINTN("De-registering dump compression type 0x%x\n",
				compression_type);
			return;
		}
	}
}

/*
 * Name: dump_compress_init()
 * Func: Initialize (or re-initialize) compression scheme.
 */
static int
dump_compress_init(int compression_type)
{
	struct list_head *tmp;
	dump_compress_t *dc;

	/* try to remove the compression item */
	list_for_each(tmp, &dump_compress_list) {
		dc = list_entry(tmp, dump_compress_t, list);
		if (dc->compress_type == compression_type) {
			dump_compress_func = dc->compress_func;
			dump_compress = compression_type;

			DUMP_PRINTF("dump_compress = %d\n", dump_compress);
			DUMP_BP();
			return (0);
		}
	}

	/* 
	 * nothing on the list -- return ENODATA to indicate an error 
	 *
	 * NB: 
	 *	EAGAIN: reports "Resource temporarily unavailable" which
	 *		isn't very enlightening.
	 */
	DUMP_PRINTF("compression_type:%d not found\n", compression_type);
	DUMP_BP();

	return (-ENODATA);
}

void dump_iobuf_end_io(struct kiobuf *iobuf)
{
	/* No wakeup needed since we've stopped scheduling */
	return;
}
/*
 * Name: dump_open_kdev()
 * Func: Try to open the kdev_t argument as the real dump device.
 *       This is where all the work is done for setting up the dump
 *       device.  It's assumed at this point that by passing in the
 *       dump device's major/minor number, we can open it up, check
 *       it out, and use it for whatever purposes.
 */
static int
dump_open_kdev(kdev_t tmp_dump_device)
{
	int i;
	unsigned long a;
	unsigned long dump_page_addr;

	/* make sure this is a valid block device */
	if (!tmp_dump_device) {
		return (-EINVAL);
	}

	/* we'd better have a name here, or this isn't valid ... */
	if (!get_blkfops(MAJOR(tmp_dump_device))) {
		return (-ENODEV);
	}

	/* if this is the second call to this function, clean up ... */
	if ((dump_okay == TRUE) && (dump_page_buf_0 != (void *)NULL)) {
		kfree((const void *)dump_page_buf_0);
	}

	/* 
	 * Allocate buffer to be used for copying pages (only once ...) 
	 *
	 * An extra:
	 *	 2 * DUMP_PAGE_SIZE:
	 * 		is needed for the page overflow of last page, it's page header
	 * 		and the EOF page header.
	 *
	 * and an  extra:
	 *	1 * PAGE_SIZE:
	 *		is needed for rounding up the start of the dump_page_buf to
	 *		a system page (PAGE_SIZE) boundry.
	 * 
	 */
	dump_page_buf = dump_page_buf_0 = (void *)kmalloc(
						DUMP_BUFFER_SIZE + (DUMP_PAGE_SIZE * 2) + PAGE_SIZE,
						GFP_KERNEL);


	if (dump_page_buf_0 == (void *)0) {
		DUMP_PRINTF("Cannot kmalloc() dump page buffer!\n");
		dump_okay = FALSE;
		return (-ENOMEM);
	}

	/*
	 * Allocate our kiobuf -- we used to do this in dump_execute(),
	 * but since the memory allocation schemes and the kiobuf code
	 * keeps changing out underneath us, we'll do it when we open
	 * up the dump device.
	 */
	dump_iobuf = (struct kiobuf *)NULL;

	/* allocate a new one for use */
	if (alloc_kiovec(1, &dump_iobuf)) {
		DUMP_PRINTF("alloc_kiovec() call failed!\n");
		dump_okay = FALSE;
		return (-ENOMEM);
	}

	/* align the dump page addresses */
	dump_page_addr = (unsigned long) dump_page_buf;
	if (dump_page_addr % PAGE_SIZE) {
		dump_page_buf = (void *) PAGE_ALIGN(dump_page_addr);
	}

	a = (unsigned long) dump_page_buf;
	/* get the base address, and copy the number of total pages */
	for (i = 0; i < (DUMP_BUFFER_SIZE >> PAGE_SHIFT); i++, a += PAGE_SIZE) {
		dump_iobuf->maplist[i] = (struct page *)virt_to_page(a);
	}
	dump_iobuf->locked = 1;
	dump_iobuf->offset = 0;
	dump_iobuf->length = DUMP_BUFFER_SIZE;
	dump_iobuf->nr_pages = (DUMP_BUFFER_SIZE >> PAGE_SHIFT);
	dump_iobuf->end_io = dump_iobuf_end_io;

	/* assign the new dump file structure */
	dump_device = tmp_dump_device;
	memcpy(dumpdev_name, bdevname(dump_device), PATH_MAX);

	/* set the sector size information */
	dump_blk_size = PAGE_SIZE;
	set_blocksize(dump_device, dump_blk_size);

	/* set the sector bits */
	i = dump_blk_size;
	for (dump_blk_shift = 0; !(i & 1);) {
		i >>= 1, dump_blk_shift++;
	}
	 DUMP_PRINTF("dump_blk_shift:%d,  PAGE_SHIFT:%d\n", 
		      dump_blk_shift,     PAGE_SHIFT);

	/* set the dump okay flag to 1, the dump device is valid */
	dump_okay = TRUE;

	DUMP_PRINTN("dump device 0x%x opened; Ready to take a save a core dump\n", dump_device);

	/* after opening the block device, return */
	return (0);
}

/*
 * Name: dump_release()
 * Func: Release the dump device -- it's never necessary to call
 *       this function, but it's here regardless.
 */
static int
dump_release(struct inode *i, struct file *f)
{
	return (0);
}

/*
 * Name: dump_ioctl()
 * Func: Allow all dump tunables through a standard ioctl() mechanism.
 *       This is far better than before, where we'd go through /proc,
 *       because now this will work for multiple OS and architectures.
 */
static int
dump_ioctl(struct inode *i, struct file *f,
	unsigned int cmd, unsigned long arg)
{
	/* check capabilities */
	if (!capable(CAP_SYS_ADMIN)) {
		return (-EPERM);
	}

	/*
	 * This is the main mechanism for controlling get/set data
	 * for various dump device parameters.  The real trick here
	 * is setting the dump device (DIOSDUMPDEV).  That's what
	 * triggers everything else.
	 */
	switch (cmd) {
		/* set dump_device */
		case DIOSDUMPDEV:
			/* check flags */
			if (!(f->f_flags & O_RDWR)) {
				return (-EPERM);
			}
			__dump_open();
			return (dump_open_kdev((kdev_t)arg));

		/* get dump_device */
		case DIOGDUMPDEV:
			return (put_user((long)dump_device, (long *)arg));

		/* set dump_level */
		case DIOSDUMPLEVEL:
			/* check flags */
			if (!(f->f_flags & O_RDWR)) {
				return (-EPERM);
			}
			/* make sure we have a positive value */
			if (arg < 0) {
				return (-EINVAL);
			}
			dump_level = (int)arg;

			/*
			 * REMIND: Still in development:
			 *
			 *	We will consider reserved pages a initial proxy for kernel pages.
			 */
			if (dump_level > DUMP_LEVEL_KERN) {
				dump_unreserved_mem = 1;
			} else {
				dump_unreserved_mem = 0;
			}

			/*
			 * REMIND: Still in development:
			 *
			 * 	Using refcount > 1 as a proxy for kernel & user pages.
			 */
			if (dump_level > DUMP_LEVEL_USED) {
				dump_unreferenced_mem = 1;
			} else {
				dump_unreferenced_mem = 0;
			}

			/*
			 *  REMIND: Still in development:
			 *
			 *	This may not be 100% stable on all ia64
			 *	memory banks. Accessing uncached memory
			 *	with cached accesses can cause subsequent
			 *	bus errors when the cache is flushed.
			 */
			if( dump_level > DUMP_LEVEL_ALL_RAM ) {
				dump_nonconventional_mem = 1;
			} else {
				dump_nonconventional_mem = 0;
			}
			break;

		/* get dump_level */
		case DIOGDUMPLEVEL:
			return (put_user((long)dump_level, (long *)arg));

		/* set dump_flags */
		case DIOSDUMPFLAGS:
			/* check flags */
			if (!(f->f_flags & O_RDWR)) {
				return (-EPERM);
			}
			/* make sure we have a positive value */
			if (arg < 0) {
				return (-EINVAL);
			}
			dump_flags = (int)arg;
			break;

		/* get dump_flags */
		case DIOGDUMPFLAGS:
			return (put_user((long)dump_flags, (long *)arg));

		/* set the dump_compress status */
		case DIOSDUMPCOMPRESS:
			/* check flags */
			if (!(f->f_flags & O_RDWR)) {
				return (-EPERM);
			}
			return (dump_compress_init((int)arg));

		/* get the dump_compress status */
		case DIOGDUMPCOMPRESS:
			return (put_user((long)dump_compress, (long *)arg));

#ifdef CONFIG_IA64
		/* get kernel address for IA64 (and others), needed by lcrash */
		case DIOGKERNELADDR:
			return (put_user(__dump_kernel_addr(), (unsigned long *)arg));
#endif

		default:
			/*
			 * In 2.5 network dump specific ioctls are passed
			 * to network dump driver from here.
			 */
			return (-EINVAL);

	}
	return (0);
}

/*
 * Name: dump_open()
 * Func: Open the dump device for use when the system crashes.
 */
static int
dump_open(struct inode *i, struct file *f)
{
	/* opening a device is straightforward -- nothing to do here */
	return (0);
}

/*
 * -----------------------------------------------------------------------
 *                     I N I T   F U N C T I O N S
 * -----------------------------------------------------------------------
 */

/*
 * Name: dump_init_proc_entry()
 * Func: Create a dump proc entry based on the /proc dump root.
 *       Returns 1 on failure, 0 on success.
 */
int
dump_init_proc_entry(char *name, struct proc_dir_entry *dirent)
{
	if (!(dirent = create_proc_entry(name,
		S_IFREG|S_IRUGO|S_IWUSR, dump_root))) {
			DUMP_PRINT("unable to initialize "
				"/proc/%s/%s!\n", DUMP_ROOT_NAME, name);
		return (1);
	}
	dirent->data = (void *)dirent;
	dirent->read_proc = &dump_read_proc;
	dirent->write_proc = NULL;
	dirent->size = 16;
	return (0);
}

/*
 * Name: dump_proc_init()
 * Func: Initialize the /proc interfaces for dumping.
 * 
 * Typically:
 *	/proc/sys/dump/
 *		dump_compress  dump_device  dump_flags  dump_level
 */
int
dump_proc_init(void)
{
	/* create the proc entries for the various tunables */
	dump_root = create_proc_entry(DUMP_ROOT_NAME, S_IFDIR, 0);
	if (dump_root) {
		dump_root->owner = THIS_MODULE;
	} else {
		DUMP_PRINTF("unable to initialize /proc/%s!\n", DUMP_ROOT_NAME);
		return (-EBUSY);
	}

	if (dump_init_proc_entry(DUMP_DEVICE_NAME, dump_dd)) {
		remove_proc_entry(DUMP_ROOT_NAME, 0);
		return (-EBUSY);
	}
	if (dump_init_proc_entry(DUMP_LEVEL_NAME, dump_l)) {
		remove_proc_entry(DUMP_ROOT_NAME, 0);
		remove_proc_entry(DUMP_DEVICE_NAME, dump_root);
		return (-EBUSY);
	}
	if (dump_init_proc_entry(DUMP_FLAGS_NAME, dump_f)) {
		remove_proc_entry(DUMP_ROOT_NAME, 0);
		remove_proc_entry(DUMP_DEVICE_NAME, dump_root);
		remove_proc_entry(DUMP_LEVEL_NAME, dump_root);
		return (-EBUSY);
	}
	if (dump_init_proc_entry(DUMP_COMPRESS_NAME, dump_cp)) {
		remove_proc_entry(DUMP_ROOT_NAME, 0);
		remove_proc_entry(DUMP_DEVICE_NAME, dump_root);
		remove_proc_entry(DUMP_LEVEL_NAME, dump_root);
		remove_proc_entry(DUMP_FLAGS_NAME, dump_root);
		return (-EBUSY);
	}
	return (0);
}

/*
 * Name: dump_proc_cleanup()
 * Func: Cleanup the /proc interfaces for dumping.
 */
void
dump_proc_cleanup(void)
{
	/* remove the proc entries */
	remove_proc_entry(DUMP_DEVICE_NAME, dump_root);
	remove_proc_entry(DUMP_LEVEL_NAME, dump_root);
	remove_proc_entry(DUMP_FLAGS_NAME, dump_root);
	remove_proc_entry(DUMP_COMPRESS_NAME, dump_root);
	remove_proc_entry(DUMP_ROOT_NAME, 0);
	return;
}

/*
 * Name: dump_init()
 * Func: Initialize the dump process.  This will set up any architecture
 *       dependent code.  The big key is we need the memory offsets before
 *       the page table is initialized, because the base memory offset
 *       is changed after paging_init() is called.
 */
int
dump_init(void)
{
	struct sysinfo info;
#if DUMP_DEBUG
	int i; 
#endif

	/* try to initialize /proc interfaces */
	if (dump_proc_init() < 0) {
		DUMP_PRINTF("dump_proc_init failed!; dump not initialized\n");
		return (-EBUSY);
	}

	/* try to create our dump device */
	if (devfs_register_chrdev(DUMP_MAJOR, "dump", &dump_fops)) {
		DUMP_PRINTF("cannot register dump character device!\n");
		return (-EBUSY);
	}

	/* initialize the dump headers to zero */
	memset(&dump_header, 0, sizeof(dump_header));
	memset(&dump_header_asm, 0, sizeof(dump_header_asm));

#if    !defined(CONFIG_DISCONTIGMEM) && !defined(CONFIG_IA64)
	/* 
	 * CONFIG_DISCONTIGMEM and CONFIG_IA64 systems are responsible 
	 * for initializing dump_mbank[] in __dump_init().
	 */
	dump_mbanks = 1;
	dump_mbank[ 0 ].start = 0;
	dump_mbank[ 0 ].end  = (((u64) max_mapnr) << PAGE_SHIFT) - 1;
	dump_mbank[ 0 ].type = DUMP_MBANK_TYPE_CONVENTIONAL_MEMORY;
#endif

	/* 
	 * initialize the dump device at the arch level.
 	 *
	 * if defined(CONFIG_DISCONTIGMEM) {
	 *       __dump_init() must set up dump_mbank[].
	 * }
	 */
	__dump_init((u64)PAGE_OFFSET);

	/* initialize the dump page buffer */
	dump_page_buf = (void *)0;

	/* set the dump function pointer for dump execution */
	dump_function_ptr = dump_execute;

	/* set the dump_compression_list structure up */
	dump_compress = DUMP_COMPRESS_NONE;
	dump_compress_func = dump_compress_none;
	dump_register_compression(&dump_none_compression);

	/* initialize the dump flags, dump level and dump_compress fields */
	dump_flags = DUMP_FLAGS_NONE;
	dump_level = DUMP_LEVEL_ALL;

	/* grab the total memory size now (not if/when we crash) */
	si_meminfo(&info);

	/* set the memory size */
	dump_header.dh_memory_size = (u64)info.totalram;

#if DUMP_DEBUG
	for (i = 0; i < dump_mbanks; i++) {
		DUMP_PRINTF("mbank[%d]: type:%d, phys_addr: " PU64X " ... " PU64X "\n",
			i,
			dump_mbank[i].type,
			dump_mbank[i].start,
			dump_mbank[i].end);

		if (dump_mbank[i].start % DUMP_PAGE_SIZE) {
			DUMP_PRINTF("oops, start is not DUMP_PAGE_SIZE:%x aligned!\n", (int) DUMP_PAGE_SIZE);
		}
		if ((dump_mbank[i].end + 1) % DUMP_PAGE_SIZE) {
			DUMP_PRINTF("oops, end is not DUMP_PAGE_SIZE:%x aligned!\n", (int) DUMP_PAGE_SIZE);
		}
	}
#endif

	DUMP_PRINTF("Crash dump driver initialized.\n");
	return (0);
}

void
dump_cleanup(void)
{
	/* get rid of the allocated dump page buffer */
	if (dump_page_buf_0) {
		kfree((const void *)dump_page_buf_0);
	}

	/* arch-specific cleanup routine */
	__dump_cleanup();

	/* remove the proc entries */
	dump_proc_cleanup();

	/* try to create our dump device */
	if (devfs_unregister_chrdev(DUMP_MAJOR, "dump")) {
		DUMP_PRINT("cannot unregister dump character device!\n");
	}

	/* reset the dump function pointer */
	dump_function_ptr = NULL;

	return;
}

EXPORT_SYMBOL(dump_register_compression);
EXPORT_SYMBOL(dump_unregister_compression);

#ifdef MODULE
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,16))
#if !defined(MODULE_LICENSE)
#define MODULE_LICENSE(str)
#endif
#endif

MODULE_AUTHOR("Matt D. Robinson <yakker@sourceforge.net>");
MODULE_DESCRIPTION("Linux Kernel Crash Dump (LKCD) driver");
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,16))
MODULE_LICENSE("GPL");
#endif
module_init(dump_init);
module_exit(dump_cleanup);
#endif /* MODULE */
