/*
 * GZIP Compression functions for kernel crash dumps.
 *
 * Created by: Matt Robinson (yakker@sourceforge.net)
 * Copyright 2001 Matt D. Robinson.  All rights reserved.
 *
 * This code is released under version 2 of the GNU GPL.
 */

/* header files */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/dump.h>

/*
 * The memory requirements for deflate are (in bytes):
 *          1 << (windowBits+2)   +  1 << (memLevel+9) + sizeof(deflate_state)
 * that is: 128K for windowBits=15  +  128K for memLevel = 8  (default values)
 *
 * It makes no sense to use a window larger than the page size, right?
 */
#define     MAX_WBITS       12  /* default of 15 */
#define     MAX_MEM_LEVEL   7   /* default of 9 */


#include "dump_zlib.c"

/*
 * -----------------------------------------------------------------------
 *                       D E F I N I T I O N S
 * -----------------------------------------------------------------------
 */
#define DUMP_MODULE_NAME "dump_gzip"
#define DUMP_PRINTN(format, args...) \
	printk("%s: " format , DUMP_MODULE_NAME , ## args);
#define DUMP_PRINT(format, args...) \
	printk(format , ## args);


static struct zalloc_pool {
    char *za_buf;
    int za_ind, za_size;
} dump_compress_gzip_pool;

/*
 * Name: dump_compress_gzip_alloc()
 * Func: The alloc function for gzip dump compression.
 *
 * We use a fixed buffer which we allocate while the system is
 * running normally.
 */
static void *
dump_compress_gzip_alloc(void *arg, unsigned int items, unsigned int size)
{
        void *rv;
        struct zalloc_pool *za = arg;
        int req = items * size;
        if (!za->za_buf || za->za_ind + req > za->za_size) {
                DUMP_PRINTN("alloc failed: %p %d %d %d\n",
                        za->za_buf, req, za->za_size, za->za_ind);
                return NULL;
        }
        rv = &za->za_buf[za->za_ind];
        za->za_ind += req;
	return (rv);
}

/*
 * Name: dump_compress_gzip_free()
 * Func: The free function for gzip dump compression.
 */
static void
dump_compress_gzip_free(void *arg, void *ptr)
{
	return;
}

/*
 * Name: dump_compress_gzip_pool_init(struct zalloc_pool *za)
 * Func: Initializes alloc_pool for a given gzip session
 */
static void
dump_compress_gzip_pool_init(struct zalloc_pool *za)
{
        za->za_ind = 0;
}

/*
 * Name: dump_compress_gzip()
 * Func: Compress a DUMP_PAGE_SIZE page using gzip-style algorithms (the.
 *       deflate functions similar to what's used in PPP).
 */
static int
dump_compress_gzip(char *old, int oldsize, char *new, int newsize)
{
	/* error code and dump stream */
	int err;
	z_stream dump_stream;

	/* setup the alloc/free functions */
	dump_stream.zalloc = dump_compress_gzip_alloc;
	dump_stream.zfree = dump_compress_gzip_free;
	dump_stream.opaque = &dump_compress_gzip_pool;

        dump_compress_gzip_pool_init(&dump_compress_gzip_pool);

	if ((err = deflateInit(&dump_stream, Z_BEST_COMPRESSION)) != Z_OK) {
		/* fall back to RLE compression */
		DUMP_PRINT("dump_compress_gzip(): deflateInit() "
			"failed (%d)!\n", err);
		return (0);
	}

	/* use old (page of memory) and size (DUMP_PAGE_SIZE) as in-streams */
	dump_stream.next_in = old;
	dump_stream.avail_in = oldsize;

	/* out streams are new (dpcpage) and new size (DUMP_DPC_PAGE_SIZE) */
	dump_stream.next_out = new;
	dump_stream.avail_out = newsize;

	/* deflate the page -- check for error */
	err = deflate(&dump_stream, Z_FINISH);
	if (err != Z_STREAM_END) {
		/* zero is return code here */
		(void)deflateEnd(&dump_stream);
		DUMP_PRINT("dump_compress_gzip(): deflate() failed (%d)!\n",
			err);
		return (0);
	}

	/* let's end the deflated compression stream */
	if ((err = deflateEnd(&dump_stream)) != Z_OK) {
		DUMP_PRINT("dump_compress_gzip(): deflateEnd() "
			"failed (%d)!\n", err);
	}

	/* return the compressed byte total (if it's smaller) */
	if (dump_stream.total_out >= oldsize) {
		return (oldsize);
	}
	return (dump_stream.total_out);
}

/* setup the gzip compression functionality */
static dump_compress_t dump_gzip_compression = {
	compress_type:	DUMP_COMPRESS_GZIP,
	compress_func:	dump_compress_gzip,
};

/*
 * Name: dump_compress_gzip_init()
 * Func: Initialize gzip as a compression mechanism.
 */
int __init
dump_compress_gzip_init(void)
{
        struct zalloc_pool *za = &dump_compress_gzip_pool;
        /* fetch our memory now */
        za->za_size = (1 << (MAX_WBITS + 2)) + (1 << (MAX_MEM_LEVEL+9))
                        + sizeof(deflate_state);
        za->za_buf = kmalloc(za->za_size, GFP_ATOMIC);
        za->za_ind = 0;
	dump_register_compression(&dump_gzip_compression);
	return (0);
}

/*
 * Name: dump_compress_gzip_cleanup()
 * Func: Remove gzip as a compression mechanism.
 */
void __init
dump_compress_gzip_cleanup(void)
{
        struct zalloc_pool *za = &dump_compress_gzip_pool;
	dump_unregister_compression(DUMP_COMPRESS_GZIP);
        if (za->za_buf)
                kfree(za->za_buf);
        za->za_buf = NULL;
        za->za_size = 0;
}

/* module initialization */
module_init(dump_compress_gzip_init);
module_exit(dump_compress_gzip_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("LKCD Development Team <lkcd-devel@lists.sourceforge.net>");
MODULE_DESCRIPTION("Gzip compression module for crash dump driver");
