/*
 * Architecture specific (ppc64) functions for Linux crash dumps.
 *
 * Created by: Matt Robinson (yakker@sgi.com)
 *
 * Copyright 1999 Silicon Graphics, Inc. All rights reserved.
 * 
 * 2.3 kernel modifications by: Matt D. Robinson (yakker@turbolinux.com)
 * Copyright 2000 TurboLinux, Inc.  All rights reserved.
 * Copyright 2003, 2004 IBM Corporation
 * 
 * This code is released under version 2 of the GNU GPL.
 */

/*
 * The hooks for dumping the kernel virtual memory to disk are in this
 * file.  Any time a modification is made to the virtual memory mechanism,
 * these routines must be changed to use the new mechanisms.
 */
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/dump.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <asm/hardirq.h>
#include <linux/irq.h>
#include <asm/machdep.h>
#include <asm/lmb.h>
#if defined(CONFIG_KDB) && !defined(CONFIG_DUMP_MODULE)
#include <linux/kdb.h>
#endif

extern unsigned long irq_affinity[];
extern dump_header_t dump_header;

static unsigned long saved_affinity[NR_IRQS];

static unsigned int last_percent;

static int alloc_dha_stack(void)
{
        int i;
        void *ptr;

        if (dump_header_asm.dha_stack[0])
                return 0;

        ptr = (void *)vmalloc(THREAD_SIZE * smp_num_cpus);
        if (!ptr) {
                return -ENOMEM;
        }

        for (i = 0; i < smp_num_cpus; i++) {
                dump_header_asm.dha_stack[i] = 
			(uint64_t)((unsigned long)ptr + (i * THREAD_SIZE));
        }
        return 0;
}

static int free_dha_stack(void)
{
        if (dump_header_asm.dha_stack[0])
                vfree((void*)dump_header_asm.dha_stack[0]);
        return 0;
}

static void
save_this_cpu_state(int cpu, struct pt_regs *regs, struct task_struct *tsk)
{
	if (regs) {
		dump_header_asm.dha_smp_regs[cpu] = *regs;
	}
	dump_header_asm.dha_smp_current_task[cpu] = (uint64_t)tsk;
	if (dump_header_asm.dha_stack[cpu]) {
		memcpy((void *)dump_header_asm.dha_stack[cpu], tsk, THREAD_SIZE);
	}
}

static int dump_expect_ipi[NR_CPUS];
static atomic_t waiting_for_dump_ipi;

static int
dump_ipi_handler(struct pt_regs *regs) 
{
	int cpu = smp_processor_id();

	if (!dump_expect_ipi[cpu])
		return 0;
	
	save_this_cpu_state(cpu, regs, current);

	dump_expect_ipi[cpu] = 0;
	atomic_dec(&waiting_for_dump_ipi);
	return 1;
}

/* save registers on other processors
 * If the other cpus don't respond we simply do not get their states.
 */
void 
save_other_cpu_states(void)
{
	int i;

	if (smp_num_cpus > 1) {
		atomic_set(&waiting_for_dump_ipi, smp_num_cpus-1);
		for (i = 0; i < NR_CPUS; i++)
			dump_expect_ipi[i] = 1;

		printk(KERN_ALERT "sending IPI to other cpus...\n");
		dump_send_ipi(dump_ipi_handler);
		i = 5000;	/* wait max of 5 seconds */
		while (atomic_read(&waiting_for_dump_ipi) && (--i > 0)) {
			barrier();
			mdelay(1);
		}
		printk(KERN_ALERT "done waiting: %d remain\n",
		       atomic_read(&waiting_for_dump_ipi));
		dump_send_ipi(NULL);	/* clear handler */
	}
}

/*
 * Name: __dump_configure_header()
 * Func: Configure the dump header with all proper values.
 */
int
__dump_configure_header(struct pt_regs *regs)
{
	int cpu = smp_processor_id();

	dump_header_asm.dha_smp_num_cpus = smp_num_cpus;
	dump_header_asm.dha_dumping_cpu = cpu;

	save_this_cpu_state(cpu, regs, current);

	save_other_cpu_states();

	return (1);
}

/*
 * Name: __dump_write_header()
 * Func: Update the header information with all architecture specific
 *       information.
 */
int
__dump_write_header(char *dpage)
{
	memcpy((void *)(dpage + sizeof(dump_header_t)),
		(const void *)&dump_header_asm, sizeof(dump_header_asm_t));

	return (1);
}

#if defined(CONFIG_KDB) && !defined(CONFIG_DUMP_MODULE)
int
kdb_sysdump(int argc, const char **argv, const char **envp, struct pt_regs *regs)
{
	kdb_printf("Dumping to disk...\n");
	dump("dump from kdb", regs);
	kdb_printf("Dump Complete\n");
	return 0;
}
#endif

/*
 * Name: __dump_init()
 * Func: Initialize the dumping routine process.  This is in case
 *       it's necessary in the future.
 */
void
__dump_init(uint64_t local_memory_start)
{

#if defined(FIXME) && defined(CONFIG_KDB) && !defined(CONFIG_DUMP_MODULE)
	/* This won't currently work because interrupts are off in kdb
	 * and the dump process doesn't understand how to recover.
	 */
	/* ToDo: add a command to query/set dump configuration */
	kdb_register_repeat("sysdump", kdb_sysdump, "", "use lkcd to dump the system to disk (if configured)", 0, KDB_REPEAT_NONE);
#endif
	/* return */
	return;
}

/*
 * Name: __dump_open()
 * Func: Open the dump device (architecture specific).  This is in
 *       case it's necessary in the future.
 */
void
__dump_open(void)
{
	alloc_dha_stack();
}


/*
 * Name: __dump_cleanup()
 * Func: Free any architecture specific data structures. This is called
 *       when the dump module is being removed.
 */
void
__dump_cleanup(void)
{
	free_dha_stack();
}

/*
 * Non dumping cpus will spin here. If a cpu is handling an irq when ipi is
 * received, we let go of it here while making sure that it hits schedule
 * on the way up and make it spin there instead.
 */
static void
__dump_spin(void *arg)
{
	if (in_irq()) {
		current->need_resched = 1;
	} else {
		while (dump_in_progress) ;
	}
}


/*
 * Routine to save the old irq affinities and change affinities of all irqs to
 * the dumping cpu.
 * 
 * NB: Need to be expanded to multiple nodes.
 */
static void
__dump_set_irq_affinity(void)
{
	int i;
	int cpu = smp_processor_id();

	memcpy(saved_affinity, irq_affinity, NR_IRQS * sizeof(unsigned long));

	for (i = 0; i < NR_IRQS; i++) {
		if (irq_desc[i].handler == NULL) {
			continue;
		}
		irq_affinity[i] = 1UL << cpu;
		if (irq_desc[i].handler->set_affinity != NULL) {
			irq_desc[i].handler->set_affinity(i, irq_affinity[i]);
		}
	}
}

/*
 * Restore old irq affinities.
 */
static void
__dump_reset_irq_affinity(void)
{
	int i;

	memcpy(irq_affinity, saved_affinity, NR_IRQS * sizeof(unsigned long));

	for (i = 0; i < NR_IRQS; i++) {
		if (irq_desc[i].handler == NULL) {
			continue;
		}
		if (irq_desc[i].handler->set_affinity != NULL) {
			irq_desc[i].handler->set_affinity(i, saved_affinity[i]);
		}
	}
}

/*
 * Name: __dump_silence_system()
 * Func: Do an architecture-specific silencing of the system.
 */
unsigned int
__dump_silence_system(unsigned int stage)
{
	ppc64_dump_msg(0x1, "Dump Start");
	last_percent = 0;
	if (stage) {	/* Do this after interrupts are enabled */
		/* read the comments above ... */
		__dump_set_irq_affinity();
		synchronize_irq();
		smp_call_function(__dump_spin, NULL, 0, 0);
	}
	return 0;
}

/*
 * Name: __dump_resume_system()
 * Func: Resume the system state in an architecture-specific way.
 */
unsigned int
__dump_resume_system(unsigned int stage)
{
	ppc64_dump_msg(0x3, "Dump Done");
	/* put the irq affinity tables back */
	__dump_reset_irq_affinity();
	return 0;
}

/* Cheap progress hack.  It estimates pages to write and
 * assumes all pages will go -- so it may get way off.
 * As the progress is not displayed for other architectures, not used at this 
 * moment.
 */
void
__dump_progress_add_page(void)
{
	unsigned long total_pages = nr_free_pages() + nr_inactive_pages + nr_active_pages;
	unsigned int percent = (dump_header.dh_num_dump_pages * 100) / total_pages;
	char buf[30];

	if (percent > last_percent && percent <= 100) {
		sprintf(buf, "Dump %3d%%     ", percent);
		ppc64_dump_msg(0x2, buf);
		last_percent = percent;
	}

}

int
ppc64_page_is_ram(u64 paddr)
{
	int i;
	unsigned long base;
	for (i=0; i < lmb.memory.cnt ;i++) {
#ifdef CONFIG_MSCHUNKS
		base = lmb.memory.region[i].physbase;
#else
		base = lmb.memory.region[i].base;
#endif
		if ((paddr >= base) && 
			(paddr < (base + lmb.memory.region[i].size))) { 
			return 1;
		}
	}

	return 0;
