/*
 * Architecture specific (x86-64) functions for Linux crash dumps.
 *
 * Created by: Matt Robinson (yakker@sgi.com)
 *
 * Copyright 1999 Silicon Graphics, Inc. All rights reserved.
 *
 * 2.3 kernel modifications by: Matt D. Robinson (yakker@turbolinux.com)
 * Copyright 2000 TurboLinux, Inc.  All rights reserved.
 *
 * x86-64 port Copyright 2002 Andi Kleen, SuSE Labs
 * x86-64 port Sachin Sant ( sachinp@in.ibm.com )
 * This code is released under version 2 of the GNU GPL.
 */

/*
 * The hooks for dumping the kernel virtual memory to disk are in this
 * file.  Any time a modification is made to the virtual memory mechanism,
 * these routines must be changed to use the new mechanisms.
 */
#include <linux/init.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/smp.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/dump.h>
#include <linux/mm.h>
#include <asm/processor.h>
#include <asm/hardirq.h>
#include <asm/uaccess.h>
#include <asm/nmi.h>

void (*dump_trace_ptr)(struct pt_regs *);

static int alloc_dha_stack(void)
{
	int i;
	void *ptr;
	
	if (dump_header_asm.dha_stack[0])
		return 0;

       	ptr = vmalloc(THREAD_SIZE * smp_num_cpus);
	if (!ptr) {
		printk("vmalloc for dha_stacks failed\n");
		return -ENOMEM;
	}

	for (i = 0; i < smp_num_cpus; i++) {
		dump_header_asm.dha_stack[i] = 
			(uint64_t)((unsigned long)ptr + (i * THREAD_SIZE));
	}
	return 0;
}

static int free_dha_stack(void) 
{
	if (dump_header_asm.dha_stack[0]) {
		vfree((void *)dump_header_asm.dha_stack[0]);	
		dump_header_asm.dha_stack[0] = 0;
	}	
	return 0;
}

static void
save_this_cpu_state(int cpu, struct pt_regs* regs, struct task_struct *tsk)
{
	if (regs)
		memcpy(&dump_header_asm.dha_smp_regs[cpu],
			regs, sizeof(struct pt_regs));
	dump_header_asm.dha_smp_current_task[cpu] = (unsigned long)tsk;
	/* take a snapshot of the stack */
	/* doing this enables us to tolerate slight drifts on this cpu */
	if (dump_header_asm.dha_stack[cpu]) {
		memcpy((void*)dump_header_asm.dha_stack[cpu], tsk, THREAD_SIZE);
	}
	return;
}

#ifdef CONFIG_SMP
extern unsigned long irq_affinity[];
extern irq_desc_t irq_desc[];
extern void dump_send_ipi(void);
static int dump_expect_ipi[NR_CPUS];
static atomic_t waiting_for_dump_ipi;
static unsigned long saved_affinity[NR_IRQS];

static int
dump_nmi_callback(struct pt_regs *regs, int cpu) 
{
	if (!dump_expect_ipi[cpu]) {
		return 0;
	}
	
	dump_expect_ipi[cpu] = 0;

	save_this_cpu_state(cpu, regs, current);
	atomic_dec(&waiting_for_dump_ipi);

	return 1;
}

/* save registers on other processors */
void 
save_other_cpu_states(void) 
{
	int i, cpu = smp_processor_id();
	int other_cpus = smp_num_cpus - 1;

	if (other_cpus > 0) {
		atomic_set(&waiting_for_dump_ipi, other_cpus);

		for (i = 0; i < NR_CPUS; i++)
			dump_expect_ipi[i] = (i != cpu && cpu_online(i));
		
		set_nmi_callback(dump_nmi_callback);
		wmb();

		dump_send_ipi();

		/* may be we dont need to wait for NMI to be processed. 
		   just write out the header at the end of dumping, if
		   this IPI is not processed untill then, there probably
		   is a problem and we just fail to capture state of 
		   other cpus. */
		while(atomic_read(&waiting_for_dump_ipi) > 0)
			cpu_relax();

		unset_nmi_callback();
	}
	return;
}

/*
 * Non dumping cpus will spin here. If a cpu is handling an irq when ipi is
 * received, we let go of it here while making sure that it hits schedule
 * on the way up and make it spin there instead.
 */
static void
__dump_spin(void *arg)
{
	if (in_irq()) {
		current->need_resched = 1;
	} else {
		while (dump_in_progress)
			touch_nmi_watchdog();
	}
	return;
}

/*
 * Routine to save the old irq affinities and change affinities of all irqs to
 * the dumping cpu.
 */
static void
__dump_set_irq_affinity(void)
{
	int i;
	int cpu = smp_processor_id();

	memcpy(saved_affinity, irq_affinity, NR_IRQS * sizeof(unsigned long));
	for (i = 0; i < NR_IRQS; i++) {
		if (irq_desc[i].handler == NULL)
			continue;
		irq_affinity[i] = 1UL << cpu;
		if (irq_desc[i].handler->set_affinity != NULL)
			irq_desc[i].handler->set_affinity(i, irq_affinity[i]);
	}
}

/*
 * Restore old irq affinities.
 */
static void
__dump_reset_irq_affinity(void)
{
	int i;

	memcpy(irq_affinity, saved_affinity, NR_IRQS * sizeof(unsigned long));
	for (i = 0; i < NR_IRQS; i++) {
		if (irq_desc[i].handler == NULL)
			continue;
		if (irq_desc[i].handler->set_affinity != NULL)
			irq_desc[i].handler->set_affinity(i, saved_affinity[i]);
	}
}

#else /* !CONFIG_SMP */
#define __dump_set_irq_affinity()	do { } while (0)
#define __dump_reset_irq_affinity()	do { } while (0)
#define save_other_cpu_states() do { } while (0)
#endif /* !CONFIG_SMP */

/*
 * Name: __dump_silence_system()
 * Func: Do an architecture-specific silencing of the system.
 *     - Change irq affinities
 *     - Wait for other cpus to come out of irq handling
 *     - Send CALL_FUNCTION_VECTOR ipi to other cpus to put them to spin
 */
unsigned int
__dump_silence_system(unsigned int stage)
{
#ifdef CONFIG_SMP
	if (stage) {    /* Do this after interrupts are enabled */
		/* read the comments above ... */
		__dump_set_irq_affinity();
		synchronize_irq();
		smp_call_function(__dump_spin, NULL, 0, 0);
	}
#endif

	/* return */
	return (0);
}

/*
 * Name: __dump_resume_system()
 * Func: Resume the system state in an architecture-specific way.
 */
unsigned int
__dump_resume_system(unsigned int stage)
{
#ifdef CONFIG_SMP
	/* put the irq affinity tables back */
	__dump_reset_irq_affinity();
#endif

	/* return */
	return (0);
}

/*
 * Name: __dump_configure_header()
 * Func: Configure the dump header with all proper values.
 */
int
__dump_configure_header(struct pt_regs *regs)
{
	int cpu = smp_processor_id();

	dump_header_asm.dha_smp_num_cpus = smp_num_cpus;
	dump_header_asm.dha_dumping_cpu = cpu;

	save_this_cpu_state(cpu, regs, current);

	save_other_cpu_states();

	return (1);
}

void
__init_mem_banks(void)
{
#if defined(CONFIG_DISCONTIGMEM)
/* copied from arch/x86_64/mm/numa.c */
#define for_all_nodes(x) for ((x) = 0; (x) <= maxnode; (x)++) \
				if ((1UL << (x)) & nodes_present)
	int i;

	memset (dump_mbank, 0, sizeof dump_mbank);
	dump_mbanks = 0;

	for_all_nodes(i) {
		dump_mbank[dump_mbanks].start
			= (u64) PLAT_NODE_DATA(i)->start_pfn << PAGE_SHIFT;
		dump_mbank[dump_mbanks].end
			= (((u64) PLAT_NODE_DATA(i)->end_pfn + 1) << PAGE_SHIFT) - 1;
		dump_mbank[dump_mbanks].type
			= DUMP_MBANK_TYPE_CONVENTIONAL_MEMORY;
		printk("dump_mbank[ %d ] = start: %lx, end %lx\n",
			dump_mbanks, dump_mbank[dump_mbanks].start,
				     dump_mbank[dump_mbanks].end);
		dump_mbanks++;
	}
#endif
}

/*
 * Name: __dump_init()
 * Func: Initialize the dumping routine process.  This is in case
 *       it's necessary in the future.
 */
void
__dump_init(uint64_t local_memory_start)
{
	__init_mem_banks();		/* Initialize Memory Banks */
	return;
}

/*
 * Name: __dump_open()
 * Func: Open the dump device (architecture specific).  This is in
 *       case it's necessary in the future.
 */
void
__dump_open(void)
{
	alloc_dha_stack();
	/* return */
	return;
}

/*
 * Name: __dump_cleanup()
 * Func: Free any architecture specific data structures. This is called
 *       when the dump module is being removed.
 */
void
__dump_cleanup(void)
{
	free_dha_stack();
	return;
}

extern int page_is_ram(unsigned long);

/*
 * Name: __dump_page_valid()
 * Func: Check if page is valid to dump.
 */
int
__dump_page_valid(unsigned long index)
{
	if (!pfn_valid(index))
		return 0;

	return page_is_ram(index);
}

