/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2003
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/lkst_private.h>
#include <linux/stat.h>
#include <linux/device.h>
#include <linux/lkst_class.h>
#include "lkst_core.h"

/* inter-module symbols */
EXPORT_SYMBOL(lkst_eh_device_alloc);
EXPORT_SYMBOL(lkst_eh_device_free);
EXPORT_SYMBOL(lkst_eh_device_register);
EXPORT_SYMBOL(lkst_eh_device_unregister);

/* intra-module symbols */
int lkst_class_init(void);
void lkst_class_cleanup(void);

/* for device model - sysfs - kobject */
static void lkst_class_release(struct class_device *cd);

struct class lkst_class = {
	.name = "lkst",
	.release = lkst_class_release,
};
EXPORT_SYMBOL(lkst_class);

/* lkst class */
static ssize_t lkst_class_show_version(struct class *cls, char * buf) 
{
	sprintf(buf, "%s\n", LKST_VERSION_STR);
	return strlen(buf) + 1;
}
static CLASS_ATTR(version,S_IRUGO,lkst_class_show_version,NULL);

static ssize_t lkst_class_show_current_maskset(struct class *cls, char * buf) 
{
	sprintf(buf, "%d\n", (int)lkst.current_maskset_id);
	return strlen(buf) + 1;
}
static ssize_t lkst_class_store_current_maskset(struct class *cls, 
						const char * buf, size_t count)
{
	lkst_maskset_id mask;
	long tmp;
	tmp = simple_strtol(buf, NULL, 0);
	if ( tmp >= 0 && tmp <= LKST_MASKSET_ID_MAX) {
		mask = (lkst_maskset_id)tmp;
		lkst_maskset_xchg(&mask);
	}
	return count; /* zero means wait forever, as it seems */
}
static CLASS_ATTR(current_maskset, S_IRUSR | S_IWUSR,
		  lkst_class_show_current_maskset,
		  lkst_class_store_current_maskset);

/* lkst class device */

/* allocate from slab */
struct lkst_eh_device * lkst_eh_device_alloc(void)
{
	struct lkst_eh_device * ehdev;
	ehdev = kmalloc(sizeof(struct lkst_eh_device), GFP_KERNEL);
	if (ehdev) {
		memset(ehdev->name, 0, LKST_EVHANDLER_NAME_LEN);
		ehdev->id = LKST_EVHANDLER_ID_VOID;
		ehdev->ctrl_store = NULL;
		ehdev->ctrl_show = NULL;
		ehdev->evhandler = NULL;
		ehdev->evhandler_ctrl = NULL;
		ehdev->class_dev.class = &lkst_class;
		ehdev->class_dev.dev = NULL;
	}
	return ehdev;
}

/* free to slab */
void lkst_eh_device_free(struct lkst_eh_device *ehdev)
{
	/*TODO : comfirm ehdev correctly derefered*/
	if (ehdev)
		kfree(ehdev);
}

static ssize_t lkst_eh_device_ctrl_show(struct class_device *cd, char *buf)
{
	struct lkst_eh_device *ehdev;
	ehdev = to_lkst_eh_device(cd);
	if (ehdev->ctrl_show) {
		return ehdev->ctrl_show(ehdev, buf);
	}
	return 0;
}
static ssize_t lkst_eh_device_ctrl_store(struct class_device *cd,
					 const char *buf, size_t count)
{
	struct lkst_eh_device *ehdev;
	ehdev = to_lkst_eh_device(cd);
	if (ehdev->ctrl_store) {
		return ehdev->ctrl_store(ehdev, buf, count);
	}
	return count;
}
static CLASS_DEVICE_ATTR(ctrl, S_IRUSR | S_IWUSR,
			 lkst_eh_device_ctrl_show,
			 lkst_eh_device_ctrl_store);

static ssize_t lkst_eh_device_name_show(struct class_device *cd, char *buf)
{
	struct lkst_eh_device *ehdev;
	ehdev = to_lkst_eh_device(cd);
	sprintf(buf, "%s\n", ehdev->name);
	return strlen(buf) + 1;
}
static CLASS_DEVICE_ATTR(name, S_IRUSR | S_IWUSR,
			 lkst_eh_device_name_show, NULL);

static ssize_t lkst_eh_device_id_show(struct class_device *cd, char *buf)
{
	struct lkst_eh_device *ehdev;
	ehdev = to_lkst_eh_device(cd);
	sprintf(buf, "%d\n", (int)ehdev->id);
	return strlen(buf) + 1;
}
static CLASS_DEVICE_ATTR(id, S_IRUSR | S_IWUSR,
			 lkst_eh_device_id_show, NULL);

static int lkst_eh_device_register_core(struct lkst_eh_device *ehdev)
{
	int ret;
	ehdev->class_dev.class = &lkst_class;
	strncpy(ehdev->class_dev.class_id, ehdev->name, BUS_ID_SIZE-1);
	ehdev->class_dev.class_id[BUS_ID_SIZE-1] = '\0';
	ret = class_device_register(&ehdev->class_dev);
	if (ret < 0) {
		LKST_DPRINTK("failed to register class device\n");
		return ret;
	}
	class_device_create_file(&ehdev->class_dev, &class_device_attr_id);
	class_device_create_file(&ehdev->class_dev, &class_device_attr_ctrl);
	class_device_create_file(&ehdev->class_dev, &class_device_attr_name);
	return ehdev->id;
}

/* register & unregister (including lkst_evhandler_register/unregister)*/
int lkst_eh_device_register(struct lkst_eh_device *ehdev)
{
	int ret;
	if ( !ehdev ) {
		return -EINVAL;
	}
	ret = lkst_evhandler_register(ehdev->id, ehdev->name,
				      ehdev->evhandler, 
				      ehdev->evhandler_ctrl);
	if ( ret < 0 ) {
		return ret;
	}
	ehdev->id = ret;

	ret = lkst_eh_device_register_core(ehdev);
	if ( ret < 0 ) {
		lkst_evhandler_unregister(ehdev->id);
	}
	return ret;
}

void lkst_eh_device_unregister(struct lkst_eh_device *ehdev)
{
	class_device_unregister(&ehdev->class_dev);
	/*unregister evhandler in lkst_class_release()*/
}

static void lkst_class_release(struct class_device *cd)
{
	int ret;
	struct lkst_eh_device *ehdev;
	ehdev = to_lkst_eh_device(cd);
	do { 
		ret = lkst_evhandler_unregister(ehdev->id);
	} while (ret == -EAGAIN) ;
}

static LKST_EH_DEV_DEF(DEFAULT, lkst_evhandler_default, 
		       lkst_evhandler_default_ctrl, NULL, NULL);
static LKST_EH_DEV_DEF(BUFFER_SHIFT, lkst_evhandler_buffer_shift, 
		       lkst_evhandler_default_ctrl, NULL, NULL);
static LKST_EH_DEV_DEF(LOG_STATIC, lkst_evhandler_log_static, 
		       lkst_evhandler_default_ctrl, NULL, NULL);

int lkst_class_init(void)
{
	int retval;
	LKST_DPRINTK("Initializing LKST class ...\n");
	
	retval = class_register(&lkst_class);
	if (retval < 0) goto out;
	retval = class_create_file(&lkst_class, &class_attr_version);
	if (retval < 0)	goto err_out;
	retval = class_create_file(&lkst_class, &class_attr_current_maskset);
	if (retval < 0)	goto err_out;

	/*register default handlers*/
	LKST_EH_DEV(DEFAULT).id = LKST_EVHANDLER_ID_DEFAULT;
	LKST_EH_DEV(BUFFER_SHIFT).id = LKST_EVHANDLER_ID_BUFFER_SHIFT;
	LKST_EH_DEV(LOG_STATIC).id = LKST_EVHANDLER_ID_LOG_STATIC;
	lkst_eh_device_register_core(&LKST_EH_DEV(DEFAULT));
	lkst_eh_device_register_core(&LKST_EH_DEV(BUFFER_SHIFT));
	lkst_eh_device_register_core(&LKST_EH_DEV(LOG_STATIC));	
	
	LKST_DPRINTK("LKST class init success\n");
	return 0;
 err_out:
	class_unregister(&lkst_class);
 out:
	printk(KERN_ERR "LKST class init fail!(%d)\n",retval);
	return retval;
}

void lkst_class_cleanup(void)
{
	LKST_DPRINTK("Cleanup LKST class\n");
	class_unregister(&lkst_class);
}

