/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2001-2004
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by K.Serizawa <serizawa@sdl.hitachi.co.jp>
 * Updated by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/

#include <linux/init.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/lkst_private.h>
#include <linux/spinlock.h>
#include <linux/utsname.h>
#include <linux/sysrq.h>
#include <linux/interrupt.h>
#include <linux/unistd.h>
// for registering notifier (DUMP, hotplug-cpu)
#include <linux/notifier.h>
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
#include <linux/dump.h>
#endif /*defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)*/
#include "lkst_core.h"

// TODO: check architecture independency
#include <asm/uaccess.h>
#include <asm/atomic.h>

/* force to call event handler */
EXPORT_SYMBOL(lkst_entry_log);

/* trace control functions */
EXPORT_SYMBOL(lkst_trc_status);	/* get status */
EXPORT_SYMBOL(lkst_trc_start);	/* set maskset to saved one.*/
EXPORT_SYMBOL(lkst_trc_stop);	/* set maskset RNOTHING and save current.*/

/* maskset control functions */
EXPORT_SYMBOL(lkst_maskset_xchg);	/* set and return previous maskset id*/
EXPORT_SYMBOL(lkst_maskset_read);	/* read a maskset */
EXPORT_SYMBOL(lkst_maskset_write);	/* write/create a maskset */
EXPORT_SYMBOL(lkst_maskset_delete);	/* delete a maskset */
EXPORT_SYMBOL(lkst_maskset_list);	/* list all maskset *id*s */
EXPORT_SYMBOL(lkst_maskset_get_id);	/* get an id of maskset from name */

/* evhandler control functions */
EXPORT_SYMBOL(lkst_evhandler_list);	/* list all event-handler *id*s */
EXPORT_SYMBOL_GPL(lkst_evhandler_ctrl);		/* send a control message to a event-handler (GPL ONLY)*/
EXPORT_SYMBOL_GPL(lkst_evhandler_register);	/* register a event-hander(GPL ONLY)*/
EXPORT_SYMBOL_GPL(lkst_evhandler_unregister);	/* unregister a event-hander(GPL ONLY)*/
EXPORT_SYMBOL_GPL(lkst_evhandler_get_id);	/* get a event-hander id from name (GPL ONLY)*/

/* buffer control functions */
EXPORT_SYMBOL(lkst_buffer_create);	/* create a buffer */
EXPORT_SYMBOL(lkst_buffer_delete);	/* delete a buffer */
EXPORT_SYMBOL(lkst_buffer_read);	/* read a buffer */
EXPORT_SYMBOL(lkst_buffer_shift);	/* shift(jump to a buffer specified by 
					 * next field of event buffer 
					 * structure) current writing buffer */
EXPORT_SYMBOL(lkst_buffer_jump);	/* jump current writing buffer to 
					 * another one */


/* evhandler primitive functions */

/* get a event buffer that has specified id. */
EXPORT_SYMBOL_GPL(lkst_evhandlerprim_select_buffer_id);

/* get a event buffer that is currently writing. */
EXPORT_SYMBOL_GPL(lkst_evhandlerprim_select_buffer);

/* set reading process to ready to read.(and wake up at next time scheduling)*/
EXPORT_SYMBOL_GPL(lkst_evhandlerprim_owner_wakeup);

/* log event to current writing buffer. */
EXPORT_SYMBOL_GPL(lkst_evhandlerprim_entry_log);

/* log event to static event buffer. */
EXPORT_SYMBOL_GPL(lkst_evhandlerprim_entry_log_static);

/* exchange current maskset for specified maskset */
EXPORT_SYMBOL_GPL(lkst_evhandlerprim_maskset_xchg);

/* default event handler control function (do nothing)*/
EXPORT_SYMBOL(lkst_evhandler_default_ctrl);

/* embeded event handlers */
/* default event handler that log event to current writing buffer. */
EXPORT_SYMBOL(lkst_evhandler_default);
/* an event handler that jump to a buffer whose id is specified next 
 * field of current writing buffer. */
EXPORT_SYMBOL(lkst_evhandler_buffer_shift);
/* an event handler that log event to static event buffer.*/
EXPORT_SYMBOL(lkst_evhandler_log_static);

#define LKST_NOT_INIT     0 // not initialized
#define LKST_INIT_LEVEL_0 1 // stage 0 init
#define LKST_INIT_LEVEL_1 2 // stage 1 init
static volatile int lkst_init_flag = LKST_NOT_INIT; // TODO: Don't use
static size_t lkst_init_buffer_size = LKST_BUFFER_SIZE_DEFAULT;
int lkst_init_maskset = LKST_MASKSET_ID_RDEFAULT;

spinlock_t lkst_evhandler_lock = SPIN_LOCK_UNLOCKED;
spinlock_t lkst_maskset_lock = SPIN_LOCK_UNLOCKED;

/* definitions of special masksets. */
/*
 * maskset_rnothing (id: 0(LKST_MASKSET_ID_RNOTHING))
 * a maskset to record no events.
 */
static const lkst_evhandler_id maskset_rnothing[] = {
	[0 ... LKST_ETYPE_MAX] = LKST_EVHANDLER_ID_VOID
	    /* caution: LKST_ETYPE_MAX is significant
	     *   to decide the size of this maskset.
	     */
};

/*
 * maskset_rall (id: 1(LKST_MASKSET_ID_RALL))
 * a maskset to record all events.
 */
static lkst_evhandler_id maskset_rall[] = {
	[0 ... LKST_ETYPE_LKST_BUFF_OVFLOW - 1] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_LKST_BUFF_OVFLOW] = LKST_EVHANDLER_ID_BUFFER_SHIFT,
	[LKST_ETYPE_LKST_BUFF_OVFLOW + 1 ... LKST_ETYPE_MAX] = LKST_EVHANDLER_ID_DEFAULT
	    /* caution: LKST_ETYPE_MAX is significant, ditto. */
};

/*
 * maskset_rdefault (id: 2(LKST_MASKSET_ID_RDEFAULT))
 * a maskset selected initially.
 */
static const lkst_evhandler_id maskset_rdefault[] = {
	/* Default Maskset omits lock operation events */
	[0 ... LKST_ETYPE_MAX] = LKST_EVHANDLER_ID_VOID,
#ifndef CONFIG_X86_4G
	[0 ... LKST_ETYPE_MEM_DO_NOPAGE-1 ] = LKST_EVHANDLER_ID_DEFAULT,
#else
	[0 ... LKST_ETYPE_PROCESS_CONTEXTSWITCH - 1 ] = LKST_EVHANDLER_ID_VOID,
        [LKST_ETYPE_PROCESS_CONTEXTSWITCH ... LKST_ETYPE_INT_BH_ENTRY] = LKST_EVHANDLER_ID_DEFAULT,
        [LKST_ETYPE_INT_BH_ENTRY + 1 ... LKST_ETYPE_SYSCALL_ENTRY - 1] = LKST_EVHANDLER_ID_VOID,
        [LKST_ETYPE_SYSCALL_ENTRY ... LKST_ETYPE_MEM_DO_NOPAGE-1] = LKST_EVHANDLER_ID_DEFAULT,
#endif
	[LKST_ETYPE_NET_PKTSENDI] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_NET_PKTRECVI] = LKST_EVHANDLER_ID_DEFAULT,
#ifdef __ARCH_WANT_SYS_SOCKETCALL
	[LKST_ETYPE_NET_SOCKETIF] = LKST_EVHANDLER_ID_DEFAULT,
#endif /*__ARCH_WANT_SYS_SOCKETCALL*/
	[LKST_ETYPE_SYSV_IPC_SEMOP ... LKST_ETYPE_SYSV_IPC_SHMCTL] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_O_PANIC ... LKST_ETYPE_TIMER_RUN ] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_LKST_INIT ... LKST_ETYPE_LKST_BUFF_OVFLOW - 1] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_LKST_BUFF_OVFLOW] = LKST_EVHANDLER_ID_BUFFER_SHIFT,
#ifndef CONFIG_X86_4G
	[LKST_ETYPE_LKST_BUFF_OVFLOW + 1 ... LKST_ETYPE_MAX] = LKST_EVHANDLER_ID_DEFAULT
#else
        [LKST_ETYPE_LKST_BUFF_HIGH] = LKST_EVHANDLER_ID_DEFAULT,
        [LKST_ETYPE_LKST_BUFF_HIGH + 1 ... LKST_ETYPE_LKST_EXTEND - 1] = LKST_EVHANDLER_ID_VOID,
        [LKST_ETYPE_LKST_EXTEND ... LKST_ETYPE_MAX] = LKST_EVHANDLER_ID_DEFAULT
#endif
	/* caution: LKST_ETYPE_MAX is significant, ditto. */
};

int lkst_entry_log(int event_type,  lkst_arg_t arg1, lkst_arg_t arg2, 
		   lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_evhandler_id evhandler_id;
	volatile lkst_maskset *maskset;
	int ret;

	if (!LKST_ETYPE_VERIFY(event_type)) return -EINVAL;

	/*get the maskset lock*/
	if (!spin_trylock(&lkst_maskset_lock)) return -EBUSY;

	if (! (maskset = lkst.current_masksetp) ) {
		spin_unlock(&lkst_maskset_lock);
		return -EINVAL;
	}
	evhandler_id = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, event_type);

	/*release the maskset lock*/
	spin_unlock(&lkst_maskset_lock);

	if (evhandler_id > LKST_EVHANDLER_ID_MAX) return -EINVAL;

	/*get the evhandler lock*/
	if (!spin_trylock(&lkst_evhandler_lock)) return -EBUSY;

	if (lkst.evhandler_table[evhandler_id].evhandler) {
		(*lkst.evhandler_table[evhandler_id].evhandler) 
			(NULL, event_type, arg1, arg2, arg3, arg4);
		ret = 0;
	}else
		ret = -EINVAL;
	/*release the evhandler lock*/
	spin_unlock(&lkst_evhandler_lock);
	
	return ret;
}
/*********************************************************************
 * evhandler primitive functions
 *********************************************************************/

/*
 * lkst_buffer_jump_core:
 * Jump to specified buffer from current buffer.
 */
static inline int lkst_buffer_jump_core(lkst_buffer_id next) {
	struct lkst_event_buffer *nxtp, *curp;
	unsigned long flags;
	register int ret;

	local_irq_save(flags);
	
	if (next == lkst_current.write_buf ||
	    next == LKST_BUFFER_ID_VOID ||
	    ! LKST_BUFFER_LOAD(next, nxtp)) {
		/* Either next buffer does not exist, or same as current */
		ret = -EINVAL;
		goto endof_irq_saved_region;
	}
	if (! LKST_CURRENT_BUFFER_LOAD(curp)){
		ret = -EBUSY;
		goto endof_irq_saved_region;
	}

	/* LKST internal event */
	LKST_HOOK_INLINE(LKST_ETYPE_LKST_BUFF_SHIFT,
			 LKST_ARG(lkst_current.write_buf),
			 LKST_ARG(next),
			 LKST_ARGP(curp), 
			 LKST_ARGP(nxtp));

	if ( (ret = lkst_buffer_enqueue_inline(nxtp) ) == 0) {
		/* success to enqueue, then initialize */
		nxtp->write_offset = nxtp->baseid - LKST_BUFFER_INIT_POS;
		nxtp->init_baseid = nxtp->baseid;
		lkst_current.write_buf = next;
		lkst_current.owner.ready = 1; /* ready to read */
	}
 endof_irq_saved_region:
	local_irq_restore(flags);
	return ret;
}

/*
 * lkst_buffer_shift_core:
 * Jump to next buffer from current buffer.
 */
static inline int lkst_buffer_shift_core(void) {
	struct lkst_event_buffer * tmp;
	if ( ! LKST_CURRENT_BUFFER_LOAD(tmp) ) return -EBUSY;
	return lkst_buffer_jump_core(tmp->next);
}

/*
 * Event buffer selecting functions
 */
struct lkst_event_buffer * \
lkst_evhandlerprim_select_buffer_id(lkst_buffer_id id)
{
	struct lkst_event_buffer * buffer;
	if ( id != LKST_BUFFER_ID_VOID && LKST_BUFFER_LOAD(id, buffer) )
		return buffer;
	else 
		return NULL;
}

struct lkst_event_buffer * \
lkst_evhandlerprim_select_buffer(void)
{
	struct lkst_event_buffer * buffer;
	if ( LKST_CURRENT_BUFFER_LOAD(buffer) )
		return buffer;
	else 
		return NULL;
}

/*
 * function lkst_evhandlerprim_owner_wakeup
 * A primitive part of an Event-handler. This awake a owner process which
 * is waiting to read a specified buffer list.
 *
 *  Arguments
 *    NONE
 *
 *  returns
 *    NONE
 *
 */
void lkst_evhandlerprim_owner_wakeup(void)
{
	lkst_current.owner.ready = 1; /* ready to read */
}

/*
 * function lkst_evhandlerprim_entry_log_inline
 * A primitive part of an Event-handler. This function write specified record
 * to specified entry.
 * 
 *  Caution: This function does not verify value of event_entry_p.
 * TODO: remove cli ...
 * Finally, I had chosen safety rather then speed.
 * But, I would provide no-cli handler by addon-module.
 * 
 *  Arguments
 *    event_type {IN}:	corresponding event type.
 *    arg1-4 {IN}:	arguments acquired at a trace point
 *
 *  returns
 *    NONE
 * 
 */
static inline void \
lkst_evhandlerprim_entry_log_inline(int event_type,
				    lkst_arg_t arg1, lkst_arg_t arg2,
				    lkst_arg_t arg3, lkst_arg_t arg4) 
{
	struct lkst_event_buffer *buffer;
	int pos;
	unsigned long flags;
	int baseid;
	local_irq_save(flags);/* local irq disabling for atomicity */
	if (! LKST_CURRENT_BUFFER_LOAD(buffer)) goto endof_cli_region;
	
	/* get a entry position */ {
		baseid = buffer->baseid++;
	}
	/* adjust the entry position */
	if ( (pos = baseid - buffer->write_offset) >= buffer->size ) {
		pos %= buffer->size;
	}
	/* logging */ {
		register struct lkst_event_record * event_entry_p = 
			&(LKST_BUFFER_BODY(buffer)[pos]);
		event_entry_p->log_recid = baseid;
		event_entry_p->log_time = lkst_evhandlerprim_mc();
		event_entry_p->log_event_type = event_type;
		event_entry_p->log_pid = current->pid; /* current pid */ ;
		event_entry_p->log_arg1 = arg1;
		event_entry_p->log_arg2 = arg2;
		event_entry_p->log_arg3 = arg3;
		event_entry_p->log_arg4 = arg4;
	}
	/* Check buffer wrapping. */
	if ( pos == 0 ) {
		/* notify current writing buffer overflowed, 
		 * occurs LKST_ETYPE_LKST_BUFF_OVFLOW event, and 
		 * set reading process to ready to read. */
		buffer->write_offset += buffer->size;
		/* Wrapped. */
		/* Process buffer-over-flow event. */
		LKST_HOOK_INLINE(LKST_ETYPE_LKST_BUFF_OVFLOW,
				 LKST_ARG(smp_processor_id()),
				 LKST_ARG(buffer->next),
				 LKST_ARG(event_type),
				 LKST_ARGP(buffer));
		/* owner wakeup */
		lkst_current.owner.ready = 1; /* ready to read */
	}
 endof_cli_region:
	local_irq_restore(flags);
}

/*
 * function lkst_evhandlerprim_entry_log_inline
 * A primitive part of an Event-handler. This write an event log to 
 * a currently selected buffer correspond to a cpu on which the event
 * has occurred.
 * 
 *  Arguments
 *    event_type {IN}:	corresponding event type.
 *    arg1-4 {IN}:	arguments acquired at a trace point
 *
 *  returns
 *    NONE
 */
void lkst_evhandlerprim_entry_log(int event_type,
				  lkst_arg_t arg1, lkst_arg_t arg2,
				  lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_evhandlerprim_entry_log_inline(event_type,
					    arg1, arg2, arg3, arg4);
}

/*
 * function lkst_evhandlerprim_entry_log_static
 * A primitive part of an Event-handler. This write an event log to 
 * a shared static buffer on which the event has occurred.
 * 
 *  Arguments
 *    event_type {IN}:	corresponding event type.
 *    arg1-4 {IN}:	arguments acquired at a trace point
 *
 *  returns
 *    NONE
 */
void lkst_evhandlerprim_entry_log_static(int event_type,
					 lkst_arg_t arg1, lkst_arg_t arg2,
					 lkst_arg_t arg3, lkst_arg_t arg4)
{
	struct lkst_event_record *event_entry_p;
	__volatile__ int recid;
	
	recid = atomic_read_and_add(1, &lkst.static_buffer.recid);
	event_entry_p = &LKST_STATIC_BUFFER_BODY()[
		((unsigned)recid & LKST_STATIC_BUFFER_MASK)];

	/* logging */
	event_entry_p->log_recid = recid;
	event_entry_p->processor = smp_processor_id(); /*current processor*/
	event_entry_p->log_time = lkst_evhandlerprim_mc(); /* current time */
	event_entry_p->log_event_type = event_type;
	event_entry_p->log_pid = current->pid; /* current pid */
	event_entry_p->log_arg1 = arg1;
	event_entry_p->log_arg2 = arg2;
	event_entry_p->log_arg3 = arg3;
	event_entry_p->log_arg4 = arg4;

}

/*
 * function lkst_evhandlerprim_maskset_xchg
 * A primitive part of an Event-handler. This primitive
 * change current selected maskset to a specified id.
 * This primitive is for LKST internal use.
 * 
 *  Arguments
 *    id {IN}:	An id of a maskset to be selected next.
 *
 *  returns
 *    ret: 	error code (private).
 */
int lkst_evhandlerprim_maskset_xchg(lkst_maskset_id * id)
{
	int	ret = 0;
	lkst_maskset_id old_id = lkst.current_maskset_id;
	lkst_maskset_id new_id;
	volatile lkst_maskset *old_maskset = NULL, *new_maskset = NULL;

	if (id == NULL || (new_id = *id) > LKST_MASKSET_ID_MAX) {
		/* Verify new maskset id. */
		LKST_DPRINTK("Invalid maskset ID=%d\n", new_id);
		return -EINVAL;
	}
	/* if we can't get maskset lock, return an error. */
	if (!spin_trylock(&lkst_maskset_lock) ) {
		return -EAGAIN;
	}

 /*start of locked region */
	/* if new maskset has not been initialized, return as error. */
	if (!(new_maskset = lkst.maskset_table[new_id].masksetp) ) { 
		LKST_DPRINTK("Specified maskset(ID=%d) does not exist!\n", new_id);
		ret = -EINVAL;
		goto  endof_atomic_region;
	}

	/* Load current maskset pointer. */
	if ( (old_maskset = lkst.current_masksetp) ) {
		/* Verify old maskset id. */
		if (old_id > LKST_MASKSET_ID_MAX || 
		    old_maskset != lkst.maskset_table[old_id].masksetp) {
			LKST_DPRINTK("Invalid ID=%d of current maskset!\n", old_id);
			old_id = LKST_MASKSET_ID_RDEFAULT;
		}
	} else {
		/*if current maskset == NULL*/
		LKST_DPRINTK("Current maskset(ID=%d) is collapsed!\n", old_id);
		old_id = LKST_MASKSET_ID_RNOTHING;
	}

	/* if old maskset does not exist, return as error. */
	if ( !(old_maskset = lkst.maskset_table[old_id].masksetp) ) {
		LKST_DPRINTK("Current maskset(id=%d) is collapsed!\n", old_id);
		ret = -EBUSY;
		goto  endof_atomic_region;
	}

	/* LKST internal event */
	LKST_HOOK_INLINE(LKST_ETYPE_LKST_MSET_XCHG,
			 LKST_ARG(old_id), LKST_ARG(new_id),
			 LKST_ARGP(old_maskset), LKST_ARGP(new_maskset));

	LKST_MASKSET_SET_CURRENT(new_id);

	*id = old_id;

	lkst_hook_maskset_xchg(new_id, old_id);

	/* LKST internal event */
	LKST_HOOK_INLINE(LKST_ETYPE_LKST_MSET_XCHG,
			 LKST_ARG(old_id), LKST_ARG(new_id),
			 LKST_ARGP(old_maskset), LKST_ARGP(new_maskset));

 endof_atomic_region:
	spin_unlock(&lkst_maskset_lock);
	return ret;
}

/*********************************************************************
 * Event handle functions (embeded)
 *********************************************************************/

/* default event handler function */
void lkst_evhandler_default(void *phookrec, int event_type,
			    lkst_arg_t arg1, lkst_arg_t arg2,
			    lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_evhandlerprim_entry_log_inline(event_type,
					    arg1, arg2, arg3, arg4);
}

/* default event handler control function */
int lkst_evhandler_default_ctrl(void *buf, size_t bufsize)
{
	return 0;
}

/* default event handler function for the static buffer */
void lkst_evhandler_log_static(void *phookrec, int event_type,
			       lkst_arg_t arg1, lkst_arg_t arg2,
			       lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_evhandlerprim_entry_log_static(event_type,
					    arg1, arg2, arg3, arg4);
}

/* Event handler for buffer shifting */
void lkst_evhandler_buffer_shift(void *phookrec, int event_type,
				 lkst_arg_t arg1, lkst_arg_t arg2,
				 lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_buffer_shift_core();
}

/*********************************************************************
 * Lkst core functions
 *********************************************************************/

/* no-lock operation: you must get maskset lock before calling this!*/
static inline lkst_maskset_id lkst_maskset_get_id_core_nl(const char *name)
{
	/* !NOTE!: please check name != NULL && name[0] != '\0'*/
	lkst_maskset_id id;
	for (id = 0; id <= LKST_MASKSET_ID_MAX; id++) {
		if (lkst.maskset_table[id].masksetp) {
			if (strncmp(name, lkst.maskset_table[id].name,
				    LKST_MASKSET_NAME_LEN) == 0){
				return id;
			}
		}
	}
	return LKST_MASKSET_ID_VOID;
}

static inline \
int lkst_maskset_validate_evhandler_nl(struct lkst_maskset_param *param)
{
	int i;
	lkst_evhandler_id id;
	for (i = 0; i < param->maskset->len; i++ ) {
		int type = param->maskset->entry[i].event_type;
		id = param->maskset->entry[i].id;

		if (type > LKST_ETYPE_MAX || type < 0) {
			LKST_DPRINTK("Invalid event_type=%d\n", type);
			return -EINVAL;
		}

		if (id == LKST_EVHANDLER_ID_VOID)
			continue; /*don't handle this event*/

		if (id > LKST_EVHANDLER_ID_MAX ||
		    !(lkst.evhandler_table[id].evhandler)) {
			LKST_DPRINTK("Invalid handler ID=%d\n", id);
			return -EINVAL;
		}
	}
	return 0;
}

static inline \
lkst_evhandler_id * lkst_maskset_create_table_inline(void)
{
	int i;
	lkst_evhandler_id * table;
	if (lkst.used_memory_size + LKST_MASKSET_TABLE_BYTE >
	    lkst.available_memory_size ) {
		LKST_DPRINTK("Cannot use kernel memory any more!\n");
		return NULL;
	}
	table = (lkst_evhandler_id *)vmalloc(LKST_MASKSET_TABLE_BYTE);
	if (table == NULL) {
		LKST_DPRINTK("Cannot allocate kernel memory!\n");
		return NULL;
	}
	lkst.used_memory_size += LKST_MASKSET_TABLE_BYTE;
	for (i = 0; i <= LKST_ETYPE_MAX; i++)
		table[i] = LKST_EVHANDLER_ID_VOID;

	return table;
}

/*
 * lkst_evhandler_register_core_nl
 * return value : 
 * 	success 0
 *	fail	errno
 * Called from lkst_evhandler_register, lkst_evhandler_unregister_core, 
 * lkst_init_stage0
 * This function do !NOT! check which name is used. 
 * please check it before call.
 * execute without lock.
 */
static inline int 
lkst_evhandler_register_core_nl(lkst_evhandler_id id, const char *name,
				lkst_evhandler_func evhandler,
				lkst_evhandler_ctrl_func evhandler_ctrl)
{
	if ( name != NULL ) {
		strncpy(lkst.evhandler_table[id].name, name,
			(size_t) LKST_EVHANDLER_NAME_LEN);
		lkst.evhandler_table[id].name[LKST_EVHANDLER_NAME_LEN - 1] = '\0';
	} else
		lkst.evhandler_table[id].name[0] = '\0';

	/*register*/
	lkst.evhandler_table[id].evhandler_ctrl = evhandler_ctrl;
	lkst.evhandler_table[id].evhandler = evhandler;
	return 0;
}

/*
 * lkst_evhandler_unregister_core
 * return value :
 *	success	0
 *	fail	errno
 * Called from lkst_evhandler_unregister, lkst_trc_init_stage0, lkst_cleanup_module
 */
static inline int lkst_evhandler_unregister_core_nl(lkst_evhandler_id id)
{
	return lkst_evhandler_register_core_nl(id, NULL, NULL, NULL);
}

/*
 * lkst_evhandler_get_id_core_nl
 * return value : evhandler_id
 * Called from lkst_evhandler_register,lkst_evhandler_get_id, lkst_evhandler_get_id_ioctl(not implemented)
 */
static inline lkst_evhandler_id
lkst_evhandler_get_id_core_nl(const char *name)
{
	lkst_evhandler_id id;
	/* !NOTE!: please check (name != NULL) && (strlen(name)!=0) before calling.*/
	for ( id = 0; id <= LKST_EVHANDLER_ID_MAX; id++) {
		if (lkst.evhandler_table[id].evhandler != NULL) {
			if (strncmp(name, 
				    lkst.evhandler_table[id].name,
				    LKST_EVHANDLER_NAME_LEN) == 0){
				return id;
			}
		}
	}
	return LKST_EVHANDLER_ID_VOID;
}

/*
 * This function just makes a instance of a buffer.
 * And does never set in the table of buffers.
 */
static struct lkst_event_buffer * \
lkst_buffer_create_instance(size_t size, lkst_buffer_id next, int cpu_id) 
{
	struct lkst_event_buffer *buffer;
	int i;
	/* align to times of 4k */
	if ( (size & ~LKST_BUFFER_PAGE_MASK) ) size += LKST_BUFFER_PAGE_BYTE;
	size &= LKST_BUFFER_PAGE_MASK;
	if (size < LKST_BUFFER_SIZE_MIN) {
		size = LKST_BUFFER_SIZE_MIN;
	} else if (size > LKST_BUFFER_SIZE_MAX) {
		size = LKST_BUFFER_SIZE_MAX;
	}

	size += LKST_BUFFER_PAGE_BYTE; /* for margin and header */

	/* create new buffer area */
	if (lkst.used_memory_size + size > lkst.available_memory_size ) {
		LKST_DPRINTK("Cannot use kernel memory any more!\n");
		return NULL;
	}
	buffer = (struct lkst_event_buffer *)vmalloc(size);
	if (buffer == NULL) {
		LKST_DPRINTK("Cannot allocate kernel memory(size=%d)\n", size);
		return NULL;
	}
	lkst.used_memory_size += size;

	buffer->next = next;
	buffer->rq_next = LKST_BUFFER_ID_VOID;
	buffer->rq_prev = LKST_BUFFER_ID_VOID;

	buffer->size = (size - LKST_BUFFER_HEADER_BYTE)/LKST_EVENT_RECORD_BYTE;
	buffer->baseid = buffer->init_baseid = 0;
	buffer->write_offset = -LKST_BUFFER_INIT_POS;
	buffer->cpu = smp_processor_id();
	buffer->cmos_time_flag = lkst.cmos_time_flag;
	buffer->xtime.tv_sec = lkst.xtime.tv_sec;
	buffer->xtime.tv_usec = lkst.xtime.tv_usec;
	strncpy(buffer->eyecatcher, "LKBF", 4);
	buffer->buf_ver = LKST_BUF_VER;

	/* clear recid, cpu_id, and event_type */
	for (i = 0; i < buffer->size; i++) {
		LKST_BUFFER_BODY(buffer)[i].log_recid = 0;
		LKST_BUFFER_BODY(buffer)[i].log_event_type = LKST_ETYPE_VOID;
		LKST_BUFFER_BODY(buffer)[i].processor = cpu_id;
	}
	return buffer;
}

/*
 * lkst_evhandler_register
 * return value
 *	success 0
 *	fail	errno
 */
int lkst_evhandler_register(lkst_evhandler_id id, const char *name,
			    lkst_evhandler_func evhandler,
			    lkst_evhandler_ctrl_func evhandler_ctrl)
{
	int ret;

	if (evhandler == NULL) {
		return -EINVAL;
	}

	if (name == NULL || name[0] == '\0' ){
		LKST_DPRINTK("Invalid evhandler name!\n");
		return -EINVAL;
	}
	if (id <= LKST_EVHANDLER_ID_RESERVED) { 
		LKST_DPRINTK("Reserved handler ID = %d\n", id);
		return -EINVAL;
	}
	
	if (!spin_trylock(&lkst_evhandler_lock)) {
		LKST_DPRINTK("Cannot get the evhandler lock.\n");
		return -EAGAIN;
	}
 /*startof_atomic_region:*/
	/* check whether the name of event handler is unique.*/
	if ( lkst_evhandler_get_id_core_nl(name) != LKST_EVHANDLER_ID_VOID) {
		LKST_DPRINTK("The name <%s> is already used.\n",name);
		ret = -EINVAL;
		goto endof_atomic_region;
	}
	if (id == LKST_EVHANDLER_ID_VOID) {
		/* Get an unused ID. */
		for (id = LKST_EVHANDLER_ID_RESERVED + 1 ; 
		     lkst.evhandler_table[id].evhandler ;
		     id++ ) { 
			if (id > LKST_EVHANDLER_ID_MAX) {
				LKST_DPRINTK("No handler ID is available.\n");
				ret = -ENOSPC;
				goto endof_atomic_region;
			}
		}
	} else if ( lkst.evhandler_table[id].evhandler ) {
		LKST_DPRINTK("Specified handler ID is already used.\n");
		ret = -EINVAL;
		goto endof_atomic_region;
	}
	
	lkst_evhandler_register_core_nl(id, name, 
					evhandler,
					evhandler_ctrl);
	ret = lkst_hook_exit_update(id);

 endof_atomic_region:
	spin_unlock(&lkst_evhandler_lock);
	return ret == 0 ? id : ret;
}

int lkst_evhandler_unregister(lkst_evhandler_id id)
{
	int ret;

	if (id <= LKST_EVHANDLER_ID_RESERVED) { 
		LKST_DPRINTK("Reserved handler ID = %d\n", id);
		return -EINVAL;
	}

	if (!spin_trylock(&lkst_evhandler_lock)) {
		LKST_DPRINTK("Cannot get the evhandler lock.\n");
		return -EAGAIN;
	}
	if ( lkst.evhandler_table[id].evhandler == NULL) {
		LKST_DPRINTK("Specified handler(ID:%d) does not exist.\n",id);
		ret = -EINVAL;
	} else {
		lkst_evhandler_unregister_core_nl(id);
		ret = lkst_hook_exit_update(id);
	}
	spin_unlock(&lkst_evhandler_lock);
	return ret;
}

/* kernel trc_status wrap function */
int lkst_trc_status(struct lkst_status_param *param)
{
	int i;

	if (param == NULL) { 
		printk(KERN_ERR "Warning : argument param is NULL!\n");
		return -EINVAL;
	}

	param->current_maskset_id = lkst.current_maskset_id;
	param->static_buffer_recid = atomic_read(&lkst.static_buffer.recid);
	param->static_buffer_size = LKST_STATIC_BUFFER_BYTE;
	param->nr_cpus = LKST_NR_CPUS;
	for (i = 0; i < LKST_NR_CPUS; i++) {
		param->buffer_id[i].write = lkst_cpu[i].write_buf;
		param->buffer_id[i].read = lkst_cpu[i].owner.read_buf;
	}

	/*needless to get lock, just ignore to count modifying masksets*/
	param->maskset_num = 0;
	for (i = 0; i < LKST_MASKSET_TBL_MAX; i++) {
		if ( lkst.maskset_table[i].masksetp )
			param->maskset_num++;
	}
	/*needless to get lock, just ignore to count modifying evhandler*/
	param->evhandler_num = 0;
	for (i = 0; i < LKST_EVHANDLER_TBL_MAX; i++) {
		if (lkst.evhandler_table[i].evhandler)
			param->evhandler_num++;
	}
	/* get version */
	param->version.major    = LKST_VERSION_MAJOR;
	param->version.minor    = LKST_VERSION_MINOR;
	param->version.revision = LKST_VERSION_REVISION;
	return 0;
}

/* lkst_trc_start* and lkst_trc_stop* functions are obsolete.*/
/* kernel trc_start wrap function */
int lkst_trc_start(void)
{
	lkst_maskset_id current_id;

	if (!spin_trylock(&lkst_maskset_lock)) {
		LKST_DPRINTK("Cannot get maskset lock.\n");
		return -EAGAIN;
	}
	
	if (lkst.maskset_id_saved > LKST_MASKSET_ID_MAX) {
		LKST_DPRINTK("Save maskset ID=%d is invalid!\n", lkst.maskset_id_saved);
	        lkst.maskset_id_saved = LKST_MASKSET_ID_RDEFAULT;
	}

	if (!(lkst.maskset_table[lkst.maskset_id_saved].masksetp)) {
		LKST_DPRINTK("Saved maskset ID=%d has already been deleted!\n", lkst.maskset_id_saved);
		lkst.maskset_id_saved = LKST_MASKSET_ID_RDEFAULT;
	}

	current_id = lkst.current_maskset_id;
	if (current_id != lkst.maskset_id_saved){
		LKST_MASKSET_SET_CURRENT(lkst.maskset_id_saved);
		lkst_hook_maskset_xchg(lkst.current_maskset_id, current_id);
	}

	spin_unlock(&lkst_maskset_lock);
	return 0;
}

/* kernel trc_stop wrap function */
int lkst_trc_stop(void)
{
	if (!spin_trylock(&lkst_maskset_lock)) {
		LKST_DPRINTK("Cannot get maskset lock.\n");
		return -EAGAIN;
	}

	lkst.maskset_id_saved = lkst.current_maskset_id;

	lkst_hook_maskset_xchg(LKST_MASKSET_ID_RNOTHING, 
			       lkst.current_maskset_id);
	LKST_MASKSET_SET_CURRENT(LKST_MASKSET_ID_RNOTHING);

	spin_unlock(&lkst_maskset_lock);
	return 0;
}

/* kernel maskset_read wrap function */
int lkst_maskset_read(struct lkst_maskset_param *param)
{
	int i, retval;
	lkst_maskset *maskset;
	struct lkst_maskset_entry *entry;
	lkst_evhandler_id id;
	size_t size;

	if (param == NULL) { 
		printk(KERN_ERR "Warning : argument param is NULL!\n");
		return -EINVAL;
	}

	size = param->maskset_size;

	if (param->id == LKST_MASKSET_ID_VOID)
		param->id = lkst.current_maskset_id;

	if (param->id > LKST_MASKSET_ID_MAX) {
		LKST_DPRINTK("Invalid maskset ID=%d\n", param->id);
		return -EINVAL;
	}

	if (!spin_trylock(&lkst_maskset_lock)) {
		LKST_DPRINTK("Cannot get maskset lock.\n");
		return -EAGAIN;
	}
 /*startof_atomic_region:*/
	
	if (!(maskset = lkst.maskset_table[param->id].masksetp) ) {
		LKST_DPRINTK("Maskset(ID=%d) does not exist!\n", param->id);
		retval = -EINVAL;
		goto endof_atomic_region;
	}

	entry = param->maskset->entry;
	param->maskset->len = LKST_MASKSET_TABLE_LEN_MAX;

        size -= LKST_MASKSET_SIZE(0);

	for (i = 0; i < LKST_MASKSET_TABLE_LEN_MAX; i++) {
		id = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, i);
		entry->event_type = i;
		entry->id = id;
                size -= sizeof(struct lkst_maskset_entry);
                if (size < sizeof(struct lkst_maskset_entry)){
			param->maskset->len = i + 1;
			break;
                }
		entry++;
	}

	strncpy(param->maskset->name, lkst.maskset_table[param->id].name, (size_t) LKST_MASKSET_NAME_LEN);
	param->maskset->name[LKST_MASKSET_NAME_LEN - 1] = '\0';
	retval = 0;
	
 endof_atomic_region:
	spin_unlock(&lkst_maskset_lock);

	return retval;
}

/* kernel maskset_write wrap function */
int lkst_maskset_write(struct lkst_maskset_param *param)
{
	int i, ret;
	lkst_maskset *maskset;
	if (in_interrupt()) { /*bad request timing.*/
		return -EBUSY;
	}
	
	if (param == NULL ) { 
		printk(KERN_ERR "Warning : argument param is NULL!\n");
		return -EINVAL;
	}
	if (param->maskset == NULL ) { 
		printk(KERN_ERR "Warning : argument param->maskset is NULL!\n");
		return -EINVAL;
	}

	/*get the maskset lock*/
	if (!spin_trylock(&lkst_maskset_lock)) {
		LKST_DPRINTK("Cannot get maskset lock.\n");
		return -EAGAIN;
	}
 /*startof_outer_region:*/
	if ( param->maskset->name[0] == '\0' ) {
		if( param->id == LKST_MASKSET_ID_VOID || 
		    ! LKST_MASKSET_TABLE_LOAD(param->id, maskset)) {
			ret = -EINVAL;
			goto endof_outer_region;
		}
	} else {
		lkst_maskset_id mid;
		mid = lkst_maskset_get_id_core_nl(param->maskset->name);
		if ( mid == LKST_MASKSET_ID_VOID ) {
			/* name is unique */
			if(param->id == LKST_MASKSET_ID_VOID) {
				/* search empty entry */
				for (i = 0; i < LKST_MASKSET_TBL_MAX; i++) {
					if (!LKST_MASKSET_TABLE_LOAD(i, 
								     maskset))
						break;
				}
				if (i == LKST_MASKSET_TBL_MAX) {
					LKST_DPRINTK("Cannot register "
						     "maskset any more!\n");
					ret = -ENOSPC;
					goto endof_outer_region;
				}
				param->id = i;
			}
		} else {/* name is not unique */
			if ( param->id == LKST_MASKSET_ID_VOID ) {
				param->id = mid;
			} else if ( mid != param->id ){
				LKST_DPRINTK("Specified ID does not "
					     "associate with %s!\n",
					     param->maskset->name);
				ret = -EINVAL;
				goto endof_outer_region;
			}/*mid == param->id*/
		}
	}

	/* check whether specified id is reserved*/
	if (param->id <= LKST_MASKSET_ID_RDEFAULT ||
	    param->id > LKST_MASKSET_ID_MAX ) {
		LKST_DPRINTK("Invalid number of maskset ID=%d\n", param->id);
		ret = -EINVAL;
		goto endof_outer_region;
	}

	if (!spin_trylock(&lkst_evhandler_lock)) {
		LKST_DPRINTK("Cannot get evhandler lock.\n");
		ret = -EAGAIN;
		goto endof_outer_region;
	}
 /*startof_inner_region:*/
	/* check whether all of specified evhandlers are valid*/
	if ( (ret = lkst_maskset_validate_evhandler_nl(param)) < 0 ) {
		goto endof_inner_region;
	}
	ret = 0;

	if (lkst.maskset_table[param->id].masksetp == NULL) {
		/* new maskset-table allocate */
		maskset = (lkst_maskset *)lkst_maskset_create_table_inline();
		if (!maskset) {
			ret = -ENOMEM;
			goto endof_inner_region;
		}
		LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table, param->id, 
					     param->maskset->name, maskset);
	} else {
		/* table will be overwritten */
		if (!LKST_MASKSET_TABLE_LOAD(param->id, maskset)) {
			LKST_DPRINTK("Maskset(ID=%d) is collapsed!\n", param->id);
			ret = -EBUSY;
			goto endof_inner_region;
		}
		if (param->maskset->name[0] != '\0') {
			LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table, 
						     param->id, 
						     param->maskset->name,
						     maskset);
		}
	}

	if(param->id == lkst.current_maskset_id) {
		for (i = 0; i < param->maskset->len; i++) {
			lkst_event_type etype = 
				param->maskset->entry[i].event_type;
			if((LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, etype) !=
			    param->maskset->entry[i].id)&&
			   (etype < LKST_MASKSET_TBL_MAX)) {
				lkst_hook_set_exit(etype,
						   param->maskset->entry[i].id);
			}
		}
	}

	/* main */
	for (i = 0; i < param->maskset->len; i++)
		if (LKST_ETYPE_VERIFY(param->maskset->entry[i].event_type))
			LKST_MASKSET_SET_EVHANDLER_ID(maskset, 
						      param->maskset->entry[i].event_type,
						      param->maskset->entry[i].id);
 endof_inner_region:
	spin_unlock(&lkst_evhandler_lock);
 endof_outer_region:
	spin_unlock(&lkst_maskset_lock);
	return ret;
}

/* kernel maskset_get_id wrap function */
int lkst_maskset_get_id(const char *name)
{
	int ret;
	if (name == NULL || name[0] == '\0') {
		return -EINVAL;
	}
	if (!spin_trylock(&lkst_maskset_lock)) {
		return -EAGAIN;
	} else {
		ret = lkst_maskset_get_id_core_nl(name);
		spin_unlock(&lkst_maskset_lock);
	}
	return ret;
}

/* kernel maskset_xchg wrap function */
int lkst_maskset_xchg(lkst_maskset_id * id)
{
	if (id == NULL) { 
		printk(KERN_ERR "Warning : argument id is NULL!\n");
		return -EINVAL;
	}
	return lkst_evhandlerprim_maskset_xchg(id);
}

/* kernel maskset_list wrap function */
int lkst_maskset_list(struct lkst_maskset_listparam *lp)
{
	int i, cnt, cnt_max;

	if (lp == NULL) { 
		printk(KERN_ERR "Warning : argument lp is NULL!\n");
		return -EINVAL;
	}

	if(lp->listent_size < LKST_MASKSET_LISTENT_SIZE(1) ||
	   lp->listent_size > LKST_MASKSET_LISTENT_SIZE(LKST_MASKSET_TBL_MAX)) {
		LKST_DPRINTK("Invalid listent_size=%d\n", lp->listent_size);
		return -EINVAL;
	}

	cnt = 0;
	cnt_max = lp->listent_size / sizeof(struct lkst_maskset_listent);

	for (i = 0; i < cnt_max; i++) {
		lp->listent[i].id = LKST_MASKSET_ID_VOID;
		lp->listent[i].name[0]='\0';
		lp->listent[i].len = 0;
	}
		
	if (!spin_trylock(&lkst_maskset_lock)) {
		LKST_DPRINTK("Cannot get maskset lock.\n");
		return -EAGAIN;
	}
 /*startof_atomic_region:*/
	for (i = 0; i < LKST_MASKSET_TBL_MAX; i++) {
		if (lkst.maskset_table[i].masksetp) {
			lp->listent[cnt].id = i;
			strcpy(lp->listent[cnt].name, 
			       lkst.maskset_table[i].name); /*TODO*/
			lp->listent[cnt].len = LKST_MASKSET_TABLE_LEN_MAX;
			cnt++;
			if(cnt >= cnt_max) break;
		}
	}
	lp->current_id = lkst.current_maskset_id;

 /*endof_atomic_region:*/
	spin_unlock(&lkst_maskset_lock);

	return 0;
}

/* kernel maskset_delete wrap function */
int lkst_maskset_delete(lkst_maskset_id id)
{
	int ret;
	lkst_maskset *maskset;

	if (id == LKST_MASKSET_ID_RALL ||
	    id == LKST_MASKSET_ID_RNOTHING ||
	    id == LKST_MASKSET_ID_RDEFAULT) {
		LKST_DPRINTK("Cannot delete special maskset!\n");
		return -EINVAL;
	}

	if (id > LKST_MASKSET_ID_MAX) {
		LKST_DPRINTK("Invalid maskset ID=%d\n", id);
		return -EINVAL;
	}

	if (!spin_trylock(&lkst_maskset_lock)) {
		LKST_DPRINTK("Cannot get maskset lock.\n");
		return -EAGAIN;
	}
	ret = 0;
 /*startof_atomic_region:*/
	if (id == lkst.maskset_id_saved) {
		LKST_DPRINTK("Specified maskset(ID=%d) is currently saved!\n Change to DEFAULT maskset.", id);
		lkst.maskset_id_saved = LKST_MASKSET_ID_RDEFAULT;
	}
	if (id == lkst.current_maskset_id) {
		LKST_DPRINTK("Specified maskset(ID=%d) is currently used!\n", id);
		ret = -EBUSY;
	} else {
		if (!(maskset = lkst.maskset_table[id].masksetp )) {
			LKST_DPRINTK("Maskset(ID=%d) does not exist!\n", id);
			ret = -EINVAL;
		} else {
			lkst.maskset_table[id].masksetp = NULL;
			lkst.used_memory_size -= LKST_MASKSET_TABLE_BYTE;
			vfree(maskset);
		}
	}
 /*endof_atomic_region:*/
	spin_unlock(&lkst_maskset_lock);

	return ret;
}

/* kernel evhandler_list wrap function */
int lkst_evhandler_list(struct lkst_evhandler_listparam *lp)
{
	int i, cnt, cnt_max;

	if (lp == NULL) { 
		printk(KERN_ERR "Warning : argument lp is NULL!\n");
		return -EINVAL;
	}

	if(lp->listent_size < sizeof(struct lkst_evhandler_listent) ||
	   lp->listent_size > LKST_EVHANDLER_LISTENT_SIZE(LKST_EVHANDLER_TBL_MAX)) {
		LKST_DPRINTK("Invalid listent_size=%d\n", lp->listent_size);
		return -EINVAL;
	}

	cnt = 0;
	cnt_max = lp->listent_size / sizeof(struct lkst_evhandler_listent);

	for (i = 0; i < cnt_max; i++) {
		lp->listent[i].id = LKST_EVHANDLER_ID_VOID;
		lp->listent[i].name[0] = '\0';
	}
		
	if (!spin_trylock(&lkst_evhandler_lock)) {
		LKST_DPRINTK("Cannot get the evhandler lock.\n");
		return -EAGAIN;
	}
	for (i = 0; i < LKST_EVHANDLER_TBL_MAX; i++) {
		if (lkst.evhandler_table[i].evhandler ) {
			lp->listent[cnt].id = i;
			strcpy(lp->listent[cnt].name,
			       lkst.evhandler_table[i].name);/*TODO*/
			if(++cnt > cnt_max) break;
		}
	}
	spin_unlock(&lkst_evhandler_lock);

	return 0;
}

/* kernel evhandler_ctrl wrap function */
int lkst_evhandler_ctrl(struct lkst_evhandler_ctrl_param *param)
{
	lkst_evhandler_ctrl_func evhandler_ctrl;
	if (param == NULL)
		return -EINVAL;

	if (param->id > LKST_EVHANDLER_ID_MAX) {
		LKST_DPRINTK("Invalid handler ID = %d\n", id);
		return -EINVAL;
	}

	if (!(evhandler_ctrl = lkst.evhandler_table[param->id].evhandler_ctrl)) {
		LKST_DPRINTK("Ctrl-function(ID=%d) does not exist!\n", id);
		return -EINVAL;
	}

	param->ret = (*evhandler_ctrl) (param->buf, param->bufsize);
	return 0;
}

/* kernel evhandler_get_id wrap function */
int lkst_evhandler_get_id(const char *name)
{
	int ret;
	/* this might be called from evhandler.
	 * thus don't check whether caller process has admin-privilege.*/
	
	if ( name == NULL || name[0] == '\0' ) {
		return -EINVAL;
	}
	if (!spin_trylock(&lkst_evhandler_lock))
		return -EAGAIN;

	ret =lkst_evhandler_get_id_core_nl(name);
	spin_unlock(&lkst_evhandler_lock);
	return ret;
}

/*
 * function lkst_buffer_jump
 * A primitive part of an Event-handler. This shift the writing buffer to 
 * the specified buffer.
 * 
 *  Arguments
 *    next:		ID of the distination buffer.
 *
 *  returns
 *    errno:		error code
 * 
 *  This will be called from event handlers those are invoked by any processes.
 *  So, we can't restrict any unpermitted accesses.
 */
int lkst_buffer_jump(lkst_buffer_id next)
{
	return lkst_buffer_jump_core(next);
}

/*
 * function lkst_evhandlerprim_buffer_shift
 * A primitive part of an Event-handler. This shift the writing buffer to 
 * the next buffer.
 * 
 *  Arguments
 *    NONE
 *
 *  returns
 *    errno:		error code
 * 
 *  This will be called from event handlers those are invoked by any processes.
 *  So, we can't restrict any unpermitted accesses.
 */
int lkst_buffer_shift(void)
{
	return lkst_buffer_shift_core();
}

/* kernel buffer_create wrap function */
int lkst_buffer_create(lkst_buffer_id id, size_t size, lkst_buffer_id next)
{
	struct lkst_event_buffer *buffer;
	int i;
	if (in_interrupt()) { /*bad request timing.*/
		return -EBUSY;
	}
	buffer = lkst_buffer_create_instance(size, next, 
					     smp_processor_id());
	if ( buffer == NULL ) return -ENOMEM;

	if (id == LKST_BUFFER_ID_VOID) {
		/* search empty entry */
		for (i = 0; i < LKST_BUFFER_TBL_MAX; i++) {
			if (LKST_BUFFER_CHECK_NULL(i))
				break;
		}
		if (i == LKST_BUFFER_TBL_MAX) {
			LKST_DPRINTK("Cannot create buffer any more!\n");
			lkst.used_memory_size -= LKST_BUFFER_REAL_SIZE(buffer);
			vfree(buffer);
			return -ENOSPC;
		}
		id = i;
	}else if (! LKST_BUFFER_CHECK_NULL(id)) {
		LKST_DPRINTK("Buffer(ID=%d) already exist!\n", id);
		lkst.used_memory_size -= LKST_BUFFER_REAL_SIZE(buffer);
		vfree(buffer);
		return -EINVAL;
	}

	/* register to buffer_table */ {
		unsigned long flags;
		buffer->id = id;
		local_irq_save(flags);
		LKST_BUFFER_STORE(id, buffer);
		local_irq_restore(flags);
	}
	LKST_DPRINTK("Create a buffer cpu=%d id=%d addr=%p size=%ul\n",
		     smp_processor_id(), id, buffer, size);

	return id;
}

/* kernel buffer_delete wrap function */
int lkst_buffer_delete(lkst_buffer_id id)
{
	struct lkst_event_buffer *buffer = NULL;
	unsigned long flags;
	int ret = 0;
	
	if (id == LKST_BUFFER_ID_VOID) {
		LKST_DPRINTK("Buffer(cpu=%d ID=%d) does not exist!\n", 
			     smp_processor_id(), id);
		return -EINVAL;
	}
	local_irq_save(flags);
	if (lkst_current.write_buf == id) {
		LKST_DPRINTK("Buffer(cpu=%d ID=%d) is currently writing!\n",
			     smp_processor_id(), id);
		ret = -EBUSY;
	} else if ( lkst_current.owner.read_buf == id && lkst_current.reading ){
		LKST_DPRINTK("Buffer(cpu=%d ID=%d) is currently reading!\n",
			     smp_processor_id(), id);
		ret = -EBUSY;
	} else {
		if (!LKST_BUFFER_LOAD(id, buffer)) {
			LKST_DPRINTK("Buffer(cpu=%d ID=%d) does not exist!\n",
				     smp_processor_id(), id);
			ret = -EINVAL;
		} else {
			/* unregister buffer pointer */
			if(buffer->rq_next != LKST_BUFFER_ID_VOID ||
			   lkst_current.rq_tail == id)
				lkst_buffer_dequeue_inline(buffer);
			/* Check whether this buffer is currently using. */
			LKST_BUFFER_STORE(id, NULL);
			if ( lkst_current.owner.read_buf == id ){
				lkst_current.owner.read_buf = 
					LKST_BUFFER_ID_VOID;
			}
		}
	}
	local_irq_restore(flags);

	/* free buffer area */
	if (ret == 0) {
		lkst.used_memory_size -= LKST_BUFFER_REAL_SIZE(buffer);
		vfree(buffer);
	}
	LKST_DPRINTK("Delete a Buffer(cpu=%d ID=%d)\n",smp_processor_id(), id);

	return ret;
}

/*-------------------------------------------------------------------
 * Initialize/cleanup functions
 *-------------------------------------------------------------------*/

static int lkst_trc_init_percpu(int cpu) 
{
	struct lkst_event_buffer *buffer;
	if ( cpu > LKST_CPU_MAX ) return -EINVAL;
	if ( lkst_cpu[cpu].write_buf != LKST_BUFFER_ID_VOID ) return 0;

	buffer = lkst_buffer_create_instance(lkst_init_buffer_size,
					     LKST_BUFFER_ID_VOID, cpu);
	if ( buffer == NULL ) { 
		LKST_DPRINTK("Can't allocate a buffer for CPU%d\n", cpu);
		return -EINVAL;
	}
	buffer->id = 0;
	lkst_cpu[cpu].buffer_table[0].ptr = buffer;

	/* set current buffer */
	lkst_cpu[cpu].write_buf = 0;
	/* enqueue current buffer */
	lkst_cpu[cpu].rq_head = 0;
	lkst_cpu[cpu].rq_tail = 0;

	LKST_DPRINTK( "Initialized CPU%d \n", cpu);
	return 0;
}

/* while dumping, stop logging.*/
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
static int lkst_dump_notify_handler(struct notifier_block *self, 
				    unsigned long stat,
				    void *dev)
{
	static lkst_maskset_id maskset_id_saved = 0;
	if ( stat == DUMP_BEGIN ) {
		maskset_id_saved = lkst.current_maskset_id;
		lkst_hook_maskset_xchg(LKST_MASKSET_ID_RNOTHING, 
				       lkst.current_maskset_id);
		LKST_MASKSET_SET_CURRENT(LKST_MASKSET_ID_RNOTHING);
	} else {
		/* after dumping, recover previous maskset */
		lkst_hook_maskset_xchg(maskset_id_saved, 
				       LKST_MASKSET_ID_RNOTHING);
		LKST_MASKSET_SET_CURRENT(maskset_id_saved);	
	}	
	lkst_evhandlerprim_entry_log_static(LKST_ETYPE_LKST_KERNEL_DUMP,
					    LKST_ARGP(dev), LKST_ARG(stat),
					    0, 0);
	return NOTIFY_DONE;
}

static struct notifier_block dump_nb = {
	.notifier_call = lkst_dump_notify_handler,
	.next = NULL,
	.priority = 0
};
#endif /*defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)*/

/*
 * function lkst_trc_init_main
 * Initialize basic part of Kernel State Tracer.
 */
int __init lkst_trc_init_main()
{
	int i, ret;

	if ( lkst_init_flag != LKST_NOT_INIT) LKST_ENCOUNTER_A_BUG; 

	/* The following values are initialized when kernel was loaded,
	 *   lkst_cpu[...].recid = 1;
	 *   maskset_rnothing[], maskset_rall[], maskset_rdefault[]
	 */

	LKST_DPRINTK("Starting initialization\n");
	/* initialize global header */{

		/* Initialize LKST header time stamp */
		lkst_time_set(0);

		spin_lock_init(&lkst_maskset_lock);
		spin_lock_init(&lkst_evhandler_lock);

		/* Clean up maskset table */
		lkst.current_masksetp = NULL;
		lkst.current_maskset_id = 0;
		lkst.maskset_id_saved = LKST_MASKSET_ID_RDEFAULT;
		for ( i = 0 ;i < LKST_MASKSET_TBL_MAX ; i++) {
			lkst.maskset_table[i].name[0] = '\0';
			lkst.maskset_table[i].masksetp = NULL;
		}

		/* Clean up evhandler table */
		for ( i = 0 ;i < LKST_EVHANDLER_TBL_MAX ; i++) {
			lkst.evhandler_table[i].name[0] = '\0';
			lkst.evhandler_table[i].evhandler = NULL;
			lkst.evhandler_table[i].evhandler_ctrl = NULL;
		}

		/* set current buffer */
		atomic_set(&lkst.static_buffer.recid, 0);
		/* clear recid, event_type */
		for (i = 0; i < LKST_STATIC_BUFFER_COUNT; i++) {
			LKST_STATIC_BUFFER_BODY()[i].log_recid = 0;
			LKST_STATIC_BUFFER_BODY()[i].log_event_type = 
				LKST_ETYPE_VOID;
		}

		/* Set architecture name */
		strncpy(lkst.arch, system_utsname.machine, 65); 

		/* Set cpu frequency */
		lkst.cpu_freq = lkst_get_cpu_khz();
		lkst_time_set(0);
	}
	
	/* initialize percpu header */
	for (i = 0; i < LKST_NR_CPUS; i++) {
		lkst_cpu[i].cpu = i;
		lkst_cpu[i].owner.pid = LKST_OWNER_PID_VOID;
		init_waitqueue_head(&lkst_cpu[i].owner.wait);
		lkst_cpu[i].owner.read_buf = LKST_BUFFER_ID_VOID;
		lkst_cpu[i].rq_head = LKST_BUFFER_ID_VOID;
		lkst_cpu[i].rq_tail = LKST_BUFFER_ID_VOID;
		lkst_cpu[i].write_buf = LKST_BUFFER_ID_VOID;
	}

	/* allocate event buffer for all CPUs */
	for (i = 0; i < smp_num_cpus; i++) {
		lkst_trc_init_percpu(i);
	}

	/* registering masksets & evhandlers */{
		spin_lock(&lkst_evhandler_lock);
		spin_lock(&lkst_maskset_lock);
		
		/* Setup nothing recording maskset */
		LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
					     LKST_MASKSET_ID_RNOTHING,
					     LKST_MASKSET_NAME_RNOTHING,
					     maskset_rnothing);
		/* And set that currently using maskset */
		LKST_MASKSET_SET_CURRENT(LKST_MASKSET_ID_RNOTHING);
		
		/* Initialize Default Event Handler */
		lkst_evhandler_register_core_nl(LKST_EVHANDLER_ID_DEFAULT,
						LKST_EVHANDLER_NAME_DEFAULT,
						lkst_evhandler_default,
						lkst_evhandler_default_ctrl);
		
		/* Initialize Buffer Shift Event Handler */
		lkst_evhandler_register_core_nl(LKST_EVHANDLER_ID_BUFFER_SHIFT,
						LKST_EVHANDLER_NAME_BUFFER_SHIFT,
						lkst_evhandler_buffer_shift,
						NULL);
		
		/* Initialize Logging static buffer Event Handler */
		lkst_evhandler_register_core_nl(LKST_EVHANDLER_ID_LOG_STATIC,
						LKST_EVHANDLER_NAME_LOG_STATIC,
						lkst_evhandler_log_static,
						NULL);
		
		/* Setup other special masksets */
		LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
					     LKST_MASKSET_ID_RALL,
					     LKST_MASKSET_NAME_RALL,
					     maskset_rall);
		LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
					     LKST_MASKSET_ID_RDEFAULT,
					     LKST_MASKSET_NAME_RDEFAULT,
					     maskset_rdefault);
		
		/* Clear entries for other masksets */
		for (i = 3; i < LKST_MASKSET_TBL_MAX; i++) {
			lkst.maskset_table[i].masksetp = NULL;
		}

		/* Kernelhooks initialization will invoked automatically now. */
		/* Initialize hook table used by lkst */
		lkst_hook_init();

		/* Set specified maskset */
		lkst_hook_maskset_xchg(lkst_init_maskset, 
				       LKST_MASKSET_ID_RNOTHING);
		LKST_MASKSET_SET_CURRENT(lkst_init_maskset);

		spin_unlock(&lkst_maskset_lock);
		spin_unlock(&lkst_evhandler_lock);
	}

	/* register Dump notifier (if available)*/
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
	register_dump_notifier(&dump_nb);
#endif /*defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)*/
	
	LKST_DPRINTK("Initialized masksets and evhandlers\n");
	LKST_HOOK_INLINE(LKST_ETYPE_LKST_INIT, LKST_ARG(0x0001), 
			 LKST_ARG(0), LKST_ARG(0), LKST_ARG(0));

	/* lkst device */
	if ( (ret = lkst_device_init()) ) {
		LKST_DPRINTK("failed to initialize lkst device(%d)\n",ret);
		return ret;
	}

#ifdef CONFIG_DEBUG_LKST_ETYPES_PROCFS
	/* lkst hook (procfs entry)*/
	if ( (ret = lkst_hook_proc_init()) ) {
		LKST_DPRINTK("failed to initialize lkst proc (%d)\n",ret);
		lkst_device_cleanup();
		return ret;
	}
#endif /*CONFIG_DEBUG_LKST_ETYPES_PROCFS*/
	/* LKST initialize flag on */
	lkst_init_flag = LKST_INIT_LEVEL_1;

	LKST_DPRINTK("Initialization finished\n");

	/* LKST internal event */
	LKST_HOOK_INLINE(LKST_ETYPE_LKST_INIT, LKST_ARG(0x10ff), 
			 LKST_ARG(0), LKST_ARG(0), LKST_ARG(0));

	return 0;
}

#ifdef MODULE
static int __init lkst_init_module(void)
{
	size_t size;

	printk(KERN_INFO "LKST: Installing driver.\n");

	/* buffer size check */
	printk(KERN_INFO "LKST: initial event buffer size=" LKST_FMT_SZT "\n", (size_t)lkst_init_buffer_size);
	if (lkst_init_buffer_size < LKST_BUFFER_SIZE_MIN) {
		printk(KERN_ERR "LKST warning(init): buffer size(=" LKST_FMT_SZT ") is too small, changed to " LKST_FMT_SZT "\n", (size_t)lkst_init_buffer_size, (size_t)LKST_BUFFER_SIZE_MIN);
		lkst_init_buffer_size = LKST_BUFFER_SIZE_MIN;
	} else if (lkst_init_buffer_size > LKST_BUFFER_SIZE_MAX) {
		printk(KERN_ERR "LKST warning(init): buffer size(=" LKST_FMT_SZT ") is too large, changed to " LKST_FMT_SZT "\n", (size_t)lkst_init_buffer_size, (size_t)LKST_BUFFER_SIZE_MAX);
		lkst_init_buffer_size = LKST_BUFFER_SIZE_MAX;
	}
	size = lkst_init_buffer_size;
	if ( size & ~LKST_BUFFER_PAGE_MASK ) size+=LKST_BUFFER_PAGE_BYTE;
	size &= LKST_BUFFER_PAGE_MASK;
	if (lkst_init_buffer_size != size)
		printk(KERN_ERR "LKST warning(init): buffer size must be times of page size, changed to " LKST_FMT_SZT "\n", (size_t)size);
	lkst_init_buffer_size = size;

	/* maskset id check */
	printk(KERN_INFO "LKST: initial maskset id = %d\n", lkst_init_maskset);
	if (lkst_init_maskset < 0 || lkst_init_maskset > 3){
		printk(KERN_ERR "LKST warning(init): initial maskset id is off limited. set default value\n");
		lkst_init_maskset = LKST_MASKSET_ID_RDEFAULT;
	}
	return lkst_trc_init_main();
}

static void __exit lkst_cleanup_module(void)
{
	int i,j;

#ifdef CONFIG_DEBUG_LKST_ETYPES_PROCFS
	lkst_hook_proc_cleanup();
#endif /*CONFIG_DEBUG_LKST_ETYPES_PROCFS*/
	lkst_device_cleanup();

	/* deregister Dump notifier (if available)*/
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
	unregister_dump_notifier(&dump_nb);
#endif /*defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)*/

	spin_lock(&lkst_evhandler_lock);
	spin_lock(&lkst_maskset_lock);
	
	/* Setup current maskset pointer to RNOTHING */
	lkst_hook_maskset_xchg(LKST_MASKSET_ID_RNOTHING,
			       lkst.current_maskset_id);
	LKST_MASKSET_SET_CURRENT(LKST_MASKSET_ID_RNOTHING);
	
	/* GKHI hook cleanup */
	lkst_hook_cleanup();

	/* Cleanup additional maskset pointer & Free */
	for (i = 3; i < LKST_MASKSET_TBL_MAX; i++) {
		lkst_maskset * maskset;
		if ( (maskset = lkst.maskset_table[i].masksetp) ) {
			LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table, i,
						     "", NULL);
			lkst.used_memory_size -= LKST_MASKSET_TABLE_BYTE;
			vfree(maskset);
		}
	}
	/* Cleanup default maskset pointer (except RNOTHING)*/
	LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
				     LKST_MASKSET_ID_RALL, "", NULL);
	LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
				     LKST_MASKSET_ID_RDEFAULT, "", NULL);

	/* Cleanup event handler */
	for (i = 0; i < LKST_EVHANDLER_TBL_MAX; i++) {
		lkst_evhandler_unregister_core_nl(i);
	}

	/* Cleanup RNOTHING maskset pointer */
	LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
				     LKST_MASKSET_ID_RNOTHING, "", NULL);
	lkst.current_masksetp = NULL;

	spin_unlock(&lkst_maskset_lock);
	spin_unlock(&lkst_evhandler_lock);

	/* Cleanup event buffer_table & Deallcation event buffer */
	for (i = 0; i < LKST_BUFFER_TBL_MAX; i++) {
		struct lkst_event_buffer * buffer;
		for ( j = 0 ; j < LKST_CPU_MAX; j++) {
			if ((buffer = lkst_cpu[j].buffer_table[i].ptr)) {
				lkst_cpu[j].buffer_table[i].ptr = NULL;
				lkst.used_memory_size -= LKST_BUFFER_REAL_SIZE(buffer);
				vfree(buffer);
			}
		}
	}

	LKST_DPRINTK_IF(lkst.used_memory_size!=0, 
			"WARNING!:there are memory leaks!(%d)", lkst.used_memory_size);
	atomic_set(&lkst.static_buffer.recid, 0);
	
	/* Cleanup current buffer */
	for (i = 0; i < LKST_CPU_MAX; i++) {
		lkst_cpu[i].rq_head = LKST_BUFFER_ID_VOID;
		lkst_cpu[i].rq_tail = LKST_BUFFER_ID_VOID;
		lkst_cpu[i].write_buf = LKST_BUFFER_ID_VOID;
	}

	/* LKST initialize flag off */
	lkst_init_flag = LKST_NOT_INIT;

	/* deallocation complete */
	printk(KERN_INFO "LKST :driver removed\n");
}

module_init(lkst_init_module);
module_exit(lkst_cleanup_module);
MODULE_AUTHOR("LKST project <http://lkst.sourceforge.net>");
MODULE_DESCRIPTION("Linux Kernel State Trace driver");
MODULE_LICENSE("GPL");
MODULE_PARM(lkst_init_buffer_size, "i");
MODULE_PARM_DESC(lkst_init_buffer_size, "Default size of each buffer(4096-0x200000, in times of 4k)");
MODULE_PARM(lkst_init_maskset, "i");
MODULE_PARM_DESC(lkst_init_maskset, "a maskset id used in starting up.[0-2]");
#else /*!MODULE*/
core_initcall(lkst_trc_init_main);
#endif /* MODULE */
