/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2003-2004
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/lkst_private.h>
#include <linux/stat.h>
#include <linux/devfs_fs_kernel.h>
#include "lkst_core.h"

// TODO: check architecture independency
#include <asm/uaccess.h>
#include <asm/atomic.h>

/* intra-module symbols */
int lkst_device_init(void);
void lkst_device_cleanup(void);
int lkst_buffer_read(struct lkst_log_buffer *lbuffer);

/* LKST device file_operations */
/* file operators */
static int lkst_open(struct inode *inode, struct file *filp);
static int lkst_close(struct inode *inode, struct file *filp);
static ssize_t lkst_read(struct file *filp, char *buf, 
			 size_t count, loff_t *ppos);
static int lkst_ioctl(struct inode *inode, struct file *filp,
		      unsigned int request, unsigned long arg);

static struct file_operations lkst_cdev_fops = {
	.owner=		THIS_MODULE,
	.ioctl=		lkst_ioctl,
	.open=		lkst_open,
	.read=		lkst_read,
	.release=	lkst_close,
};

/* initialize device interface (/dev/lkst) */
int lkst_device_init(void)
{
	int ret = 0;

	LKST_DPRINTK("Initializing LKST device ...\n");

        /* register LKST device file_operations */
#ifdef CONFIG_DEVFS_FS
	ret = devfs_register_chrdev(LKST_DEV_MAJOR, LKST_DEV_NAME, &lkst_cdev_fops);
	if (ret != 0){
		printk(KERN_ERR "LKST warning(init):cannot register LKST *conventional* device!\n");
		goto out;
	}
#else
	ret = register_chrdev(LKST_DEV_MAJOR, LKST_DEV_NAME, &lkst_cdev_fops);

	if (ret != 0){
		printk(KERN_ERR "LKST error(init):cannot register LKST device!\n");
		goto out;
	}
#endif

#ifdef CONFIG_DEVFS_FS
        lkst_devfs_handle = devfs_register (NULL, LKST_DEV_NAME,
                                            DEVFS_FL_DEFAULT,
                                            LKST_DEV_MAJOR, LKST_DEV_MINOR,
                                            S_IFCHR | S_IRUSR | S_IWUSR,
                                            &lkst_cdev_fops, NULL);
        if ( lkst_devfs_handle == NULL )
		printk(KERN_ERR "LKST error(init):cannot register LKST devfs device!\n");
#ifdef CONFIG_DEBUG_LKST_DEVICE_PER_CPU
        for ( i=0; i<LKST_NR_CPUS; i++) {/* for Hotplug CPU...*/
		char devname[32];
		sprintf(devname, "%s%d", LKST_DEV_NAME, i);
		lkst_devfs_handle_cpu[i] = devfs_register (NULL, devname,
							   DEVFS_FL_DEFAULT,
							   LKST_DEV_MAJOR, i+1,
                                                           S_IFCHR | S_IRUSR | S_IWUSR,
							   &lkst_cdev_fops, NULL
			);
		if ( lkst_devfs_handle_cpu[i] == NULL )
			printk(KERN_ERR "LKST error(init):cannot register LKST devfs device!\n");
        }
#endif /*CONFIG_DEBUG_LKST_DEVICE_PER_CPU*/
#endif

 out:
	LKST_DPRINTK_IF((ret==0), "LKST device init success\n");
	return ret;
}

void lkst_device_cleanup(void)
{
	int i;
	LKST_DPRINTK("Cleanup LKST device\n");
	/* unregister LKST device file_operations */
#ifdef CONFIG_DEVFS_FS
        if (devfs_unregister_chrdev(LKST_DEV_MAJOR, LKST_DEV_NAME))
#else
		if (unregister_chrdev(LKST_DEV_MAJOR, LKST_DEV_NAME))
#endif
			printk(KERN_ERR "LKST error :cannot unregister LKST device!\n");

#ifdef CONFIG_DEVFS_FS
        devfs_unregister ( lkst_devfs_handle );
#ifdef CONFIG_DEBUG_LKST_DEVICE_PER_CPU
        for ( i=0; i<LKST_NR_CPUS; i++) {/* for Hotplug CPU...*/
                devfs_unregister ( lkst_devfs_handle_cpu[i] );
        }
#endif /*CONFIG_DEBUG_LKST_DEVICE_PER_CPU*/
#endif

}

/*
 * Functions related buffers
 */

static int lkst_buffer_set_current_to_owner(int cpu)
{
	unsigned long allowed;
	
	/* !NOTE!: please check sp != NULL before calling this.*/
	if (lkst_current.owner.pid == current->pid) {
		LKST_DPRINTK("This process(pid=%d) is already "
			     "set to read other buffer list.\n", current->pid);
		return -EBUSY;
	}

	/* CPU binding */
	if( lkst_cpu_migrate_save(&allowed, cpu) < 0) //TODO
		return -EINVAL;

	if (lkst_current.owner.pid != LKST_OWNER_PID_VOID && 
	    lkst_current.owner.pid != current->pid) {
		LKST_DPRINTK("Specified buffer list is already set "
			     "reading mode by pid=%d\n",
			     lkst_current.owner.pid);
		lkst_cpu_migrate_restore(allowed);
		return -EBUSY;
	}

	/* Set owner pid to the buffer list of specified CPU */ 
	lkst_current.owner.pid = current->pid;
	return 0;
}

static __inline__ \
int lkst_buffer_list_core(struct lkst_buffer_listparam *lp)
{
	int i, cnt, cnt_max;
	struct lkst_event_buffer *buffer;

	/* !NOTE!: please check (lp != NULL) before calling this.*/
	cnt = 0;
	cnt_max = lp->listent_num;

	/* list buffers */
	for (i = 0; i < LKST_BUFFER_TBL_MAX; i++) {
		if (!LKST_BUFFER_LOAD(i, buffer)) continue;
		if (cnt < cnt_max) {
			lp->listent[cnt].id = buffer->id;
			lp->listent[cnt].next = buffer->next;
			lp->listent[cnt].rq_prev = buffer->rq_prev;
			lp->listent[cnt].rq_next = buffer->rq_next;
			lp->listent[cnt].size = buffer->size*LKST_EVENT_RECORD_BYTE;
			lp->listent[cnt].write_offset = buffer->write_offset;
			lp->listent[cnt].baseid = buffer->baseid;
		}
		cnt++;
	}
	/* record percpu header */
	lp->buffer_num = cnt;
	lp->write_buf = lkst_current.write_buf;
	lp->read_buf = lkst_current.owner.read_buf;
	lp->rq_head = lkst_current.rq_head;
	lp->rq_tail = lkst_current.rq_tail;
	lp->owner = lkst_current.owner.pid;
	lp->read_pos = lkst_current.owner.read_pos;

	return cnt;
}

/* -------------------------
 * functions to read buffers
 */

/*
 * fill header of the log buffer.
 * that read from global header.
 */
static __inline__ void \
lkst_log_header_read_inline(struct lkst_log_buffer * lbuffer) {
	/* get buffer time stamp */
	lbuffer->tsc = lkst.tsc;
	lbuffer->xtime = lkst.xtime;
	lbuffer->cpu_freq = lkst.cpu_freq;
	lbuffer->buf_ver = lkst.buf_ver;
	strncpy(lbuffer->arch, lkst.arch, LKST_ARCH_NAME_LEN);
	lbuffer->endian_big = lkst.endian_big;
}

/*
 * get the next buffer from read queue and dequeue current buffer(if the 
 * next buffer differ from current)
 */

static __inline__ struct lkst_event_buffer * \
lkst_buffer_get_next_readable(int overwritten ) {
	struct lkst_event_buffer *buffer;
	unsigned long flags;
	local_irq_save(flags);

	if ( lkst_current.rq_head != LKST_BUFFER_ID_VOID ) {
		/* then dequeue new buffer */
		if ( LKST_BUFFER_LOAD(lkst_current.rq_head, buffer) ) {
			if ( buffer->baseid != buffer->init_baseid )
				if (lkst_buffer_dequeue_inline(buffer) == 0 ) {
					goto  success;
				}
		}
	} else if ( overwritten ) {
		/* then load same buffer */
		if ( lkst_current.owner.read_buf != LKST_BUFFER_ID_VOID &&
		     LKST_BUFFER_LOAD(lkst_current.owner.read_buf, buffer) ) {
			goto  success;
		}
	}
	local_irq_restore(flags);
	return NULL;
 success:
	lkst_current.owner.read_pos = LKST_BUFFER_INIT_POS - 1;
	lkst_current.owner.read_baseid = 
		LKST_BUFFER_BODY(buffer)[LKST_BUFFER_INIT_POS].log_recid - 1;
	lkst_current.owner.read_buf = buffer->id;
	local_irq_restore(flags);
	return buffer;
}

/* 
 * read entrys from a buffer less than log_num, until breaking 
 * baseid incrementation 
 */
static int __inline__ \
lkst_buffer_entry_read_core(struct lkst_log_record * log, int log_num,
			    struct lkst_event_record * body, int size, 
			    struct lkst_buffer_owner * owner,
			    int * overwritten) { 
	int read_pos;
	int count, baseid, bufferid;
	struct task_struct *task = NULL;
	struct lkst_event_record *entry;
	uid_t uid = 0;
	gid_t gid = 0;
	pid_t pid = LKST_OWNER_PID_VOID;
	pid_t pgrp = 0;

	*overwritten = 0;
	read_pos = owner->read_pos;	/*CARE: read_pos might indicate negative value.*/
	baseid = owner->read_baseid;
	bufferid = owner->read_buf;
	for ( count = 0; count < log_num; count++, log++) {
		read_pos++; baseid++;
		entry = &body[(unsigned)read_pos%size]; //TODO?
		log->posix.log_pid = entry->log_pid;
		log->posix.log_time.tv_sec = (time_t) ((entry->log_time) >> 32);
		log->posix.log_time.tv_nsec = (unsigned long) entry->log_time;
		log->posix.log_thread = -1;		// TODO: write after
		log->posix.log_event_type = entry->log_event_type;
		log->log_arg1 = entry->log_arg1;
		log->log_arg2 = entry->log_arg2;
		log->log_arg3 = entry->log_arg3;
		log->log_arg4 = entry->log_arg4;
		log->posix.log_processor = (posix_log_procid_t)(entry->processor);
		if ( entry->log_recid != baseid ) { 
			baseid--; 
			read_pos--;
			if (body[(unsigned)read_pos%size].log_recid != baseid) {
				*overwritten = 1;
				log->posix.log_recid = LKST_BUFFER_RECID(bufferid, baseid+1);
				log->posix.log_event_type = LKST_ETYPE_LKST_OVWRTN_REC;
				log->posix.log_magic = LKST_LOG_MAGIC_NUM;
				log->posix.log_size = sizeof(lkst_arg_t) * 4;
				log->posix.log_format = PXLOG_BINARY;
				log->posix.log_facility = LOG_KERN;
				log->posix.log_severity = LOG_DEBUG;
				log->posix.log_uid = uid;
				log->posix.log_gid = gid;
				log->posix.log_pgrp = pgrp;
				log->posix.log_flags = 0;
				log++; count++;
			}
			break;
		}
		log->posix.log_recid = LKST_BUFFER_RECID(bufferid, baseid);
		read_pos = (unsigned)read_pos%size; //update read position
		
		/* Is the current (logged) pid same as the last?*/
		if (pid != log->posix.log_pid) {
			/*update uid, gid and pgrp.*/
			pid = log->posix.log_pid;
			read_lock(&tasklist_lock);
			task = find_task_by_pid(pid);
			if (task == NULL) {
				/* write dummy value if task_struct was already deleted */
				uid = -1;	// TODO: write after
				gid = -1;	// TODO: write after
				pgrp = -1;	// TODO: write after
			} else {
				uid = task->uid;
				gid = task->gid;
				pgrp = task->pgrp;
			}
			read_unlock(&tasklist_lock);
		}
		log->posix.log_magic = LKST_LOG_MAGIC_NUM;
		log->posix.log_size = sizeof(lkst_arg_t) * 4;
		log->posix.log_format = PXLOG_BINARY;
		log->posix.log_facility = LOG_KERN;
		log->posix.log_severity = LOG_DEBUG;
		log->posix.log_uid = uid;
		log->posix.log_gid = gid;
		log->posix.log_pgrp = pgrp;
		log->posix.log_flags = 0;
	}
	owner->read_pos = read_pos;
	owner->read_baseid = baseid;
	return count;
}

/*
 * function lkst_buffer_read_all_to_user_core
 * Read from the buffers.
 * This must be called from user space.
 * 
 *  Arguments
 *    logbuffer:	specify user-space buffer.
 *    count:		number of read entries, described as entry counts. not bytes.
 *
 *  returns
 *    result read size:	the size of entries that was read.(counts)
 *    -EAGAIN:		read queue is empty. please read again after a minuts.
 *    -ENOMEM:		failed to allocate temporary buffer.
 *    -EBUSY:		buffer is currently busy.(may be or there is a bug)
 *  
 */
static __inline__ \
int lkst_buffer_read_all_to_user_core(struct lkst_log_record * logbuffer,
				      int count)
{
	struct lkst_event_buffer * buffer;
	struct lkst_log_record * tmp_buf; //TODO: not vmalloc
	int log_pos, read_cnt, log_num, overwritten = 0;

	/* OK, this is *PRIVATE* function, save pointless argument check. */

	/* allocate temprary buffer */
	tmp_buf = (struct lkst_log_record *) vmalloc(LKST_TEMP_BUFFER_BYTE);
	if ( tmp_buf == NULL ) return -ENOMEM;
	log_num = LKST_TEMP_BUFFER_COUNT;
	log_pos = 0;

	/* get the oldest buffer*/
	if ( lkst_current.owner.read_buf != LKST_BUFFER_ID_VOID ) {
		if ( !LKST_BUFFER_LOAD(lkst_current.owner.read_buf, buffer)) {
			log_pos = -EBUSY;
			goto endof_temp_alloc_region;
		}
	} else {
		/* first time, or the buffer before read is deleted */
		if (! (buffer = lkst_buffer_get_next_readable(0))) {
			log_pos = -EAGAIN;
			goto endof_temp_alloc_region;
		}
	}
	/* header is needless to read */
	/* reading loop */
	do {
		if ( (count - log_pos) < log_num )
			log_num = (count - log_pos);

		read_cnt = lkst_buffer_entry_read_core(tmp_buf, log_num,
						       LKST_BUFFER_BODY(buffer), 
						       buffer->size, 
						       &lkst_current.owner,
						       &overwritten);
		if ( read_cnt < log_num ) {	/* end of this buffer */
			buffer = lkst_buffer_get_next_readable(overwritten);
		}
		if ( read_cnt > 0 ) {
			/* temp buffer copy to user space */
			if ( copy_to_user(
				     &logbuffer[log_pos], tmp_buf,
				     read_cnt*sizeof(struct lkst_log_record))){
				printk(KERN_ERR "Failed to copy to user space:"
				       "address %p, size " LKST_FMT_SZT "\n",
				       &logbuffer[log_pos], 
				       (size_t)read_cnt*sizeof(
					       struct lkst_log_record));
				break;//TODO : return correct error code.
			}
			log_pos += read_cnt;
		}
	} while ( buffer != NULL && log_num > 0 );

 endof_temp_alloc_region:
	vfree(tmp_buf);
	return log_pos;
}

/*
 * function lkst_buffer_read_one_to_user_core
 * Read from specified buffer.
 * This must be called from user space.
 * 
 *  Arguments
 *    id:		If specify LKST_BUFFER_ID_VOID, then read static-shared
 *                      buffer, otherwise read specified buffer.
 *    logbuffer:	specify user-space buffer.
 *    count:		number of read *entries*, described as entry counts. 
 *          		Not bytes.
 *
 *  returns
 *    result read size:	the size of entries that was read.(counts)
 *    -EINVAL:		specified buffer can not use.
 *    -ENOMEM:		failed to allocate temporary buffer.
 *  
 */
static __inline__ \
int lkst_buffer_read_one_to_user_core(lkst_buffer_id id, 
				      struct lkst_log_record * logbuffer,
				      int count)
{
	struct lkst_buffer_owner owner;
	struct lkst_log_record * tmp_buf;
	int log_pos, read_cnt, log_num, overwritten=0;
	struct lkst_event_record *body;
	int size;

	/* allocate temprary buffer */
	tmp_buf = (struct lkst_log_record *) vmalloc(LKST_TEMP_BUFFER_BYTE);
	if ( tmp_buf == NULL ) return -ENOMEM;
	log_num = LKST_TEMP_BUFFER_COUNT;
	log_pos = 0;

	/* get buffer body and size*/
	if ( id == LKST_BUFFER_ID_VOID ) { // static shared buffer
		unsigned int start_recid;
		start_recid = (unsigned)(atomic_read(&lkst.static_buffer.recid) + 2);
		if ( start_recid < LKST_STATIC_BUFFER_COUNT )
			start_recid = 0;
		else 
			start_recid %= LKST_STATIC_BUFFER_COUNT;
		body = LKST_STATIC_BUFFER_BODY();
		size = LKST_STATIC_BUFFER_COUNT;
		owner.read_pos = start_recid - 1;
		owner.read_buf = 0;
		owner.read_baseid = body[start_recid].log_recid - 1;
	} else { /* get specified buffer */  // percpu buffer
		struct lkst_event_buffer * buffer;
		if (! LKST_BUFFER_LOAD(id, buffer) ) {
			vfree(tmp_buf);
			return -EINVAL;
		}
		body = LKST_BUFFER_BODY(buffer);
		size = buffer->size;
		owner.read_pos = LKST_BUFFER_INIT_POS - 1;
		owner.read_buf = id;
		owner.read_baseid = body[LKST_BUFFER_INIT_POS].log_recid - 1;
	}
	
	/* reading loop */
	do {
		if ( (count - log_pos) < log_num )
			log_num = (count - log_pos);

		read_cnt = lkst_buffer_entry_read_core(tmp_buf, log_num, 
						       body, size,
						       &owner,
						       &overwritten);
		if ( read_cnt > 0 ) {
			/* temp buffer copy to user space */
			if ( copy_to_user(
				     &logbuffer[log_pos], tmp_buf,
				     read_cnt*sizeof(struct lkst_log_record))){
				printk(KERN_ERR "Failed to copy to user space:"
				       "address %p, size " LKST_FMT_SZT "\n",
				       &logbuffer[log_pos], 
				       (size_t)read_cnt*sizeof(
					       struct lkst_log_record));
				break;//TODO : return correct error code.
			}
			log_pos += read_cnt;
		}
	} while ( log_num > 0 && read_cnt == log_num);

	vfree(tmp_buf);
	return log_pos;
}

/*
 * function lkst_buffer_read_core
 * Read from the specified buffer.
 * This is called from lkst_buffer_read
 * 
 *  Arguments
 *    lbuffer:		the logging buffer.
 *
 *  returns
 *    result read size:	the size of entries.(counts)
 *    -EINVAL:		spacified buffer is not exist.
 *  
 *  !NOTE!: This must be called after CPU migration. 
 * 
 */
static __inline__ \
int lkst_buffer_read_core(struct lkst_log_buffer * lbuffer)
{
	struct lkst_event_buffer * buffer;
	struct lkst_buffer_owner owner;
	int ow;

	if( lbuffer->id == LKST_BUFFER_ID_VOID || 
	    ! LKST_BUFFER_LOAD(lbuffer->id, buffer) ) {
		return -EINVAL;
	}
	lbuffer->cpu = smp_processor_id();
	owner.read_pos = LKST_BUFFER_INIT_POS;
	owner.read_buf = lbuffer->id;
	owner.read_baseid = 
		LKST_BUFFER_BODY(buffer)[LKST_BUFFER_INIT_POS].log_recid;
	lkst_log_header_read_inline(lbuffer);
	return lbuffer->result_read_size = 
		lkst_buffer_entry_read_core(lbuffer->buffer, 
					    lbuffer->read_size,
					    LKST_BUFFER_BODY(buffer),
					    buffer->size,
					    &owner, &ow);
}

/* kernel buffer_read wrap function 
 * this function is exported in lkst_core.c
 */
int lkst_buffer_read(struct lkst_log_buffer *lbuffer)
{
	int ret;
	if (lbuffer == NULL) { 
		printk(KERN_ERR "Warning : argument lbuffer is NULL!\n");
		return -EINVAL;
	}

	lkst_current.reading++; /*read flag up*/
	ret = lkst_buffer_read_core(lbuffer);
	lkst_current.reading--; /*read flag down*/
	return ret ;
}

/*********************************************************************
 * ioctl functions (static *CLOSED* functions)
 ********************************************************************/

/* ioctl trc_status wrap function */
static int lkst_trc_status_ioctl(struct lkst_status_param *param_user)
{
	int retval = 0;
	struct lkst_status_param param;

	retval = lkst_trc_status(&param);

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_status_param))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n", 
			     param_user);
		return -EINVAL;
	}

	return retval;
}

/* ioctl trc_start wrap function */
static int lkst_trc_start_ioctl(void)
{
	return lkst_trc_start();
}

/* ioctl trc_stop wrap function */
static int lkst_trc_stop_ioctl(void)
{
	return lkst_trc_stop();
}

/* ioctl maskset_read wrap function */
static int lkst_maskset_read_ioctl(struct lkst_maskset_param *param_user)
{
	int retval = 0;
	struct lkst_maskset_param param;
	struct lkst_maskset_body *maskset;
	struct lkst_maskset_body *user_maskset;

	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_maskset_param))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	if (param.maskset_size <= 0 ||
	    verify_area(VERIFY_WRITE, param.maskset, param.maskset_size)) {
		LKST_DPRINTK("Invalid argument(maskset/maskset_size)\n");
		return -EINVAL;
	}

	user_maskset = param.maskset;
	maskset = vmalloc(param.maskset_size);
	if (maskset == NULL) {
		LKST_DPRINTK("Cannot allocate kernel memory : size=%d\n", 
			     param.maskset_size);
		return -ENOMEM;
	}

	param.maskset = maskset;
	retval = lkst_maskset_read(&param);

	if (copy_to_user(user_maskset, param.maskset, param.maskset_size)) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     user_maskset);
		return -EINVAL;
	}

	param.maskset = user_maskset;

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_maskset_param))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	vfree(maskset);

	return retval;
}

/* ioctl maskset_write wrap function */
static int lkst_maskset_write_ioctl(struct lkst_maskset_param *param_user)
{
	int retval = 0;
	struct lkst_maskset_param param;
	struct lkst_maskset_body *maskset, *save_user_maskset;

	if (copy_from_user(&param, param_user, 
			   sizeof(struct lkst_maskset_param))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	if (param.maskset_size <= 0 ||
	    verify_area(VERIFY_READ, param.maskset, param.maskset_size)) {
		LKST_DPRINTK("Invalid argument(maskset/maskset_size)\n");
		return -EINVAL;
	}

	maskset = vmalloc(param.maskset_size);
	if (maskset == NULL) {
		LKST_DPRINTK("Cannot allocate kernel memory : size=%d\n",
			     param.maskset_size);
		return -ENOMEM;
	}

	if (copy_from_user(maskset, param.maskset, param.maskset_size)) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param.maskset);
		return -EINVAL;
	}

	save_user_maskset = param.maskset;
	param.maskset = maskset;

	retval = lkst_maskset_write(&param);

	param.maskset = save_user_maskset;

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_maskset_param))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	vfree(maskset);

	return retval;
}

/* ioctl maskset_get_id wrap function */
static int lkst_maskset_get_id_ioctl(struct lkst_maskset_getparam *param_user)
{
	struct lkst_maskset_getparam param;
	int ret;
	
	if (copy_from_user(&param, param_user, 
			   sizeof(struct lkst_maskset_getparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}
	
	if (param.name[0] == '\0') {
		return -EINVAL;
	}
	
	ret = lkst_maskset_get_id(param.name);
	if ( ret < 0 ) 
		return ret;

	param.id = ret;
	if (copy_to_user(param_user, &param, 
			 sizeof(struct lkst_maskset_getparam))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n", param_user);
		return -EINVAL;
	}
	return 0;
}

/* ioctl maskset_set wrap function */
static int lkst_maskset_set_ioctl(lkst_maskset_id id)
{
	return lkst_maskset_xchg(&id);
}

/* ioctl maskset_list wrap function */
static int lkst_maskset_list_ioctl(struct lkst_maskset_listparam *param_user)
{
	int retval = 0;
	struct lkst_maskset_listparam param;
	struct lkst_maskset_listent *listent;
	struct lkst_maskset_listent *user_listent;
	if (copy_from_user(&param, param_user, 
			   sizeof(struct lkst_maskset_listparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}


	if (param.listent_size <= 0 || 
	    verify_area(VERIFY_WRITE, param.listent, param.listent_size)) {
		LKST_DPRINTK("Invalid argument(listent/listent_size)\n");
		return -EINVAL;
	}
	
	user_listent = param.listent;
	listent = vmalloc(param.listent_size);
	if (listent == NULL) {
		LKST_DPRINTK("Cannot allocate kernel memory : size=%d\n",
			     param.listent_size);
		return -ENOMEM;
	}

	param.listent = listent;

	retval = lkst_maskset_list(&param);

	if (copy_to_user(user_listent, param.listent, param.listent_size)) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     user_listent);
		return -EINVAL;
	}

	param.listent = user_listent;

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_maskset_listparam))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	vfree(listent);

	return retval;
}

/* ioctl maskset_delete wrap function */
static int lkst_maskset_delete_ioctl(lkst_maskset_id id)
{
	return lkst_maskset_delete(id);
}

/* ioctl evhandler_list wrap function */
static int lkst_evhandler_list_ioctl(struct lkst_evhandler_listparam *param_user)
{
	int retval = 0;
	struct lkst_evhandler_listparam param;
	struct lkst_evhandler_listent *listent;
	struct lkst_evhandler_listent *user_listent;

	if (copy_from_user(&param, param_user, 
			   sizeof(struct lkst_evhandler_listparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}


	if (param.listent_size <= 0 ||
	    verify_area(VERIFY_WRITE, param.listent, param.listent_size)) {
		LKST_DPRINTK("Invalid argument(listent/listent_size)\n");
		return -EINVAL;
	}
	
	user_listent = param.listent;
	listent = vmalloc(param.listent_size);
	if (listent == NULL) {
		LKST_DPRINTK("Cannot allocate kernel memory : size=%d\n",
			     param.listent_size);
		return -ENOMEM;
	}

	param.listent = listent;

	retval = lkst_evhandler_list(&param);

	if (copy_to_user(user_listent, param.listent, param.listent_size)) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     user_listent);
		return -EINVAL;
	}

	vfree(listent);

	return retval;
}

/* ioctl evhandler_ctrl wrap function */
static int lkst_evhandler_ctrl_ioctl(struct lkst_evhandler_ctrl_param *param_user)
{
	int retval = 0;
	struct lkst_evhandler_ctrl_param param;
	char *buf = NULL;
	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_evhandler_ctrl_param))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}


	if (param.bufsize < 0 ||
	    verify_area(VERIFY_READ, param.buf, param.bufsize)) {
		LKST_DPRINTK("Invalid argument(buf/bufsize)\n");
		return -EINVAL;
	}

	buf = param.buf; /* save */
	if (param.bufsize != 0) {
		param.buf = vmalloc(param.bufsize);
		if (param.buf == NULL) {
			LKST_DPRINTK("Cannot allocate kernel memory : "
				     "size=%d\n", param.bufsize);
			return -ENOMEM;
		}
		if (copy_from_user(param.buf, buf, param.bufsize)) {
			LKST_DPRINTK("Reading fail at user memory address %p\n"
				     , buf);
			vfree(param.buf);
			return -EINVAL;
		}
	}
	retval = lkst_evhandler_ctrl(&param);

	/* copy return value */
	{ void *ptmp = param.buf; param.buf = buf; buf = ptmp;} /*restore*/
	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_evhandler_ctrl_param))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
		vfree(buf);
		return -EINVAL;
	}
	/* copy contents of arguments */
	if (param.bufsize != 0) {
		if (copy_to_user(param.buf, buf, param.bufsize) ) {
			LKST_DPRINTK("Writing fail at user memory address %p\n",
				     param.buf);
			retval = -EINVAL;
		}
		vfree(buf);
	}
	return retval;
}

/* ioctl evhandler_get_id wrap function */
static int lkst_evhandler_get_id_ioctl(struct lkst_evhandler_getparam *param_user)
{
	struct lkst_evhandler_getparam param;
	
	if (copy_from_user(&param, param_user, 
			   sizeof(struct lkst_evhandler_getparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}
	
	if (param.name[0] == '\0') {
		return -EINVAL;
	}
	
	param.id = lkst_evhandler_get_id(param.name);
	
	if (copy_to_user(param_user, &param, 
			 sizeof(struct lkst_evhandler_getparam))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n", param_user);
		return -EINVAL;
	}
	return 0;
}

/* ioctl buffer_read wrap function */
static int lkst_buffer_read_ioctl(struct lkst_log_buffer *lbuffer_user)
{
	int retval = 0;
	struct lkst_log_buffer lbuffer;

	if (copy_from_user(&lbuffer, lbuffer_user,
			   sizeof(struct lkst_log_buffer))) {
	  	LKST_DPRINTK("Reading fail at user memory address %p\n", 
			     lbuffer_user);
		return -EINVAL;
	}
	if (lbuffer.read_size <= 0 ||
	    verify_area(VERIFY_WRITE, lbuffer.buffer, lbuffer.read_size )) {
		LKST_DPRINTK("Invalid argument(buffer/read_size)\n");
		return -EINVAL;
	}
	
	if ( lbuffer.id == LKST_BUFFER_ID_VOID && lbuffer.cpu != -1 ){
		LKST_DPRINTK("Invalid argument(buffer id)\n");
		return -EINVAL;
	}
	/*cpu migration*/{
		unsigned long allowed;
		if( lbuffer.id != LKST_BUFFER_ID_VOID ) {
			if ( lkst_cpu_migrate_save(&allowed, lbuffer.cpu) < 0)
				return -EINVAL;
		}

		lkst_current.reading++;/*read flag up*/
		retval = lkst_buffer_read_one_to_user_core(
				lbuffer.id, lbuffer.buffer, 
				lbuffer.read_size/LKST_LOG_RECORD_BYTE);
		if (retval >= 0) 
			lbuffer.result_read_size = retval*LKST_LOG_RECORD_BYTE;
		lkst_current.reading--;/*read flag down*/

		if( lbuffer.id != LKST_BUFFER_ID_VOID ) {
			lkst_cpu_migrate_restore(allowed);
		}
	}
	lkst_log_header_read_inline(&lbuffer);
	if (copy_to_user(lbuffer_user, &lbuffer,
			 sizeof(struct lkst_log_buffer))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     lbuffer_user);
		return -EINVAL;
	}
	return retval;
}

/* ioctl buffer_create wrap function */
static int lkst_buffer_create_ioctl(struct lkst_buffer_param *param_user)
{
	int retval = 0;
	struct lkst_buffer_param param;

	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_buffer_param))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	/*migrate cpu*/{
		unsigned long allowed;
	
		LKST_DPRINTK("Migrating cpu=%d...",param.cpu);
		if ( lkst_cpu_migrate_save(&allowed, param.cpu) < 0) {
			LKST_DPRINTK("failed!\n");
			return -EINVAL;
		}
		LKST_DPRINTK("OK(%d)\n",smp_processor_id());

		retval = lkst_buffer_create(param.id, param.size, 
					    param.next);
		if (retval >= 0 && retval <= LKST_BUFFER_ID_MAX) {
			struct lkst_event_buffer * buffer;
			if (!LKST_BUFFER_LOAD(retval, buffer))
				LKST_ENCOUNTER_A_BUG;//TODO
			param.id = (lkst_buffer_id)retval;
			param.result_size= buffer->size*LKST_EVENT_RECORD_BYTE;
		}
		lkst_cpu_migrate_restore(allowed);
	}

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_buffer_param))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	return retval;
}

/* ioctl buffer_delete wrap function */
static int lkst_buffer_delete_ioctl(struct lkst_buffer_delparam *user_param)
{
	struct lkst_buffer_delparam param;
	int retval;
	if (copy_from_user(&param, user_param, 
			   sizeof(struct lkst_buffer_delparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     user_param);
		return -EINVAL;
	}
	/*cpu migration*/{
		unsigned long allowed;
		if(lkst_cpu_migrate_save(&allowed, param.cpu)<0)
			return -EINVAL;
		retval = lkst_buffer_delete(param.id);
		lkst_cpu_migrate_restore(allowed);
	}
	return retval;
}

/* ioctl buffer_shift wrap function */
static int lkst_buffer_shift_ioctl(int cpu)
{
	int retval;
	/*cpu migration*/{
		unsigned long allowed;
		if(lkst_cpu_migrate_save(&allowed, cpu)<0)
			return -EINVAL;
		retval = lkst_buffer_shift();
		lkst_cpu_migrate_restore(allowed);
	}
	return retval;
}

/* ioctl buffer_jump wrap function */
static int lkst_buffer_jump_ioctl(struct lkst_buffer_jumpparam *user_param)
{
	struct lkst_buffer_jumpparam param;
	int retval;
	if (copy_from_user(&param, user_param,
			   sizeof(struct lkst_buffer_jumpparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     user_param);
		return -EINVAL;
	}
	/*cpu migration*/{
		unsigned long allowed;
		if(lkst_cpu_migrate_save(&allowed, param.cpu)<0)
			return -EINVAL;
		retval = lkst_buffer_jump(param.dest);
		lkst_cpu_migrate_restore(allowed);
	}
	return retval;
}

/* ioctl buffer_list wrap function */
static int lkst_buffer_list_ioctl(struct lkst_buffer_listparam *user_param)
{
	int retval = 0;
	struct lkst_buffer_listparam param;
	struct lkst_buffer_listent *listent;
	struct lkst_buffer_listent *user_listent;

	if (copy_from_user(&param, user_param,
			   sizeof(struct lkst_buffer_listparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     user_param);
		return -EINVAL;
	}
	if (param.listent_num < 0 ){
		LKST_DPRINTK("Invalid argument(listent_num)\n");
		return -EINVAL;
	}
	
	if (param.listent_num > 0) {
		if ( verify_area(VERIFY_WRITE, param.listent, 
				 param.listent_num*
				 sizeof(struct lkst_buffer_listent)) ){
			LKST_DPRINTK("Invalid argument(listent)\n");
			return -EINVAL;
		}
		user_listent = param.listent;
		listent = vmalloc(param.listent_num*sizeof(struct lkst_buffer_listent));
		if (listent == NULL) {
			LKST_DPRINTK("Cannot allocate kernel memory : size=%d\n", 
				     param.listent_num*sizeof(struct lkst_buffer_listent));
			return -ENOMEM;
		}
	} else {
		user_listent = NULL;
		listent = NULL;
	}
	
	/*migrate cpu*/{
		unsigned long allowed;
		if(lkst_cpu_migrate_save(&allowed, param.cpu)<0)
			return -EINVAL;

		param.listent = listent;
		retval = lkst_buffer_list_core(&param);
		
		lkst_cpu_migrate_restore(allowed);
	}
	param.listent = user_listent;
	if (copy_to_user(user_param, &param,
			 sizeof(struct lkst_buffer_listparam))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     user_param);
		return -EINVAL;
	}
	if (param.listent_num > 0) {
		if (param.listent_num < param.buffer_num)
			param.buffer_num = param.listent_num;
		if (copy_to_user(user_listent, listent,
				 param.buffer_num*sizeof(struct lkst_buffer_listent))){
			LKST_DPRINTK("Writing fail at user memory address %p\n",
				     user_listent);
			return -EINVAL;
		}
		vfree(listent);
	}

	return retval;
}

/* ioctl buffer_link function */
static int lkst_buffer_link_ioctl(struct lkst_buffer_linkparam *user_param)
{
	struct lkst_buffer_linkparam param;
	int ret = 0;

	/* get user parameter */
	if (copy_from_user(&param, user_param,
			   sizeof(struct lkst_buffer_linkparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     user_param);
		return -EINVAL;
	}

	if (param.id == param.next) {
		LKST_DPRINTK("Source and Distinaion are same buffer.\n");
		return -EINVAL;
	}
	
	/*cpu migration*/{
		unsigned long allowed;
		struct lkst_event_buffer * buffer;
		if(lkst_cpu_migrate_save(&allowed, param.cpu)<0)
			return -EINVAL;
		if ( LKST_BUFFER_LOAD(param.id, buffer)) {
			buffer->next = param.next;
		} else {
			ret = -EINVAL;
		}
		lkst_cpu_migrate_restore(allowed);
	}
	return ret;
}

/* ioctl buffer_setrmod wrap function */
static int lkst_buffer_setrmod_ioctl(struct lkst_buffer_srmodparam *param_user)
{
	int retval = 0;
	struct lkst_buffer_srmodparam param;


	if (copy_from_user(&param, param_user,
			   sizeof(struct lkst_buffer_srmodparam))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	retval = lkst_buffer_set_current_to_owner(param.cpu);
	if (retval < 0) {
		return retval;
	}
	
	/* get buffer time stamp */
	param.tsc = lkst.tsc;
	param.xtime.tv_sec = lkst.xtime.tv_sec;
	param.xtime.tv_usec = lkst.xtime.tv_usec;
	param.cpu_freq = lkst.cpu_freq;
	param.buf_ver = lkst.buf_ver;
	strncpy(param.arch, lkst.arch, LKST_ARCH_NAME_LEN);
	param.endian_big = lkst.endian_big;

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_buffer_srmodparam))) {
		LKST_DPRINTK("Writing fail at user memory address %p\n",
			     param_user);
		return -EINVAL;
	}

	return retval;
}

/* ioctl entry_log wrap function */
static int lkst_entry_log_ioctl(struct lkst_entry_args *arg_user)
{
	struct lkst_entry_args arg;

	if (copy_from_user(&arg, arg_user, sizeof(struct lkst_entry_args))) {
		LKST_DPRINTK("Reading fail at user memory address %p\n",
			     arg_user);
		return -EINVAL;
	}

	(void) lkst_entry_log(arg.event_type,
			      arg.arg1, arg.arg2, arg.arg3, arg.arg4);

	return (0);
}

/* LKST device open function */
static int lkst_open(struct inode *inode, struct file *filp)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
#ifdef CONFIG_DEBUG_LKST_DEVICE_PER_CPU
{	int cpu = MINOR(inode->i_rdev) - 1;
	int retval;
	/* Instead of using setrmod, you can open /dev/lkst0-31 
	 * those minor numbers are from 1 to 32, to read buffers
	 * of each cpus.*/
	/* But /dev/lkst remains for backward-compatibility and
	 * general interface (for masksets/event-handlers)*/
	if (cpu != -1 ) {
		if (!lkst_cpu_is_online(cpu))
			return -ENODEV;
		retval = lkst_buffer_set_current_to_owner(cpu);
		if (retval < 0)
			return retval;

		filp->private_data = (void *)((lkst_ptr_t)cpu);
	}else 
		filp->private_data = (void *)-1;
}
#endif /*CONFIG_DEBUG_LKST_DEVICE_PER_CPU*/

	return 0;
}


/* LKST device close function */
static int lkst_close(struct inode *inode, struct file *filp)
{
	int i;

	/* Clear owner pid of buffer list */
	for (i = 0; i < LKST_CPU_MAX; i++) {
		if (lkst_cpu[i].owner.pid == current->pid) {
			lkst_cpu[i].owner.pid = LKST_OWNER_PID_VOID;
			/* CPU unbinding */
			lkst_cpu_migrate_restore(cpu_online_map);
			break;
		}
	}

	return 0;
}


/* LKST device read function */
static ssize_t lkst_read(struct file *filp, char *buf, 
			 size_t count, loff_t *ppos)
{
	ssize_t read_count;
	int retval;
	
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (lkst_current.owner.pid != current->pid ) 
		return -EBUSY;

#ifdef CONFIG_DEBUG_LKST_DEVICE_PER_CPU
	if ( filp->private_data != (void *)-1 ) {
		struct lkst_log_buffer_header *head;
		
		head = (struct lkst_log_buffer_header *) kmalloc (
				sizeof(struct lkst_log_buffer_header),
				GFP_KERNEL);
		if (!head) return -ENOMEM;

		/* setting up the header of log buffer*/
		head->header.log_magic = LOGFILE_MAGIC;
		head->header.log_version = 0;

		head->entry.log_size =  2*sizeof(int) + 
			sizeof(char)*LKST_ARCH_NAME_LEN;
		head->entry.log_format = PXLOG_BINARY;

		head->endian_big = lkst.endian_big;
		head->buf_ver = lkst.buf_ver;
		strncpy(head->arch, lkst.arch, LKST_ARCH_NAME_LEN);

		/* copy to user space (header only) */
		copy_to_user(buf, head, sizeof(struct lkst_log_buffer_header));
		kfree(head);
		
		filp->private_data = (void *)-1;
		return sizeof(head);
	}
#endif /*CONFIG_DEBUG_LKST_DEVICE_PER_CPU*/
	
	if (count < sizeof(struct lkst_log_record) || 
	    verify_area(VERIFY_WRITE, buf, count)) {
		LKST_DPRINTK("Invalid argument(buf/count)\n");
		return -EINVAL;
	}
	count /= sizeof(struct lkst_log_record);
	read_count = 0;

	lkst_current.reading++;/*read flag up*/
 read_retry:
	retval = lkst_buffer_read_all_to_user_core(
			(struct lkst_log_record *)buf,
			(int)count);
	if (retval > 0) read_count += retval;
	if ((retval > 0 && retval < (int)count) || retval == -EAGAIN) {
		if(!(filp->f_flags & O_NONBLOCK)){
			lkst_current.owner.ready = 0;
			interruptible_sleep_on(&lkst_current.owner.wait);
			if (!signal_pending(current)) {
				buf += retval*sizeof(struct lkst_log_record);
				count -= retval;
				goto read_retry;
			}else {
				/* We cannot rewind the read process.
				 * But, at least, we can send read size
				 * back to user process, instead of -EINTR.*/
			}
		}
	} else if ( retval < 0 ){
		lkst_current.reading--;/*read flag down*/
		return retval;
	}
	lkst_current.reading--;/*read flag down*/
	
	read_count *= sizeof(struct lkst_log_record);
	return read_count; /* we must return in byte*/
}

/* LKST device ioctl function */
static int lkst_ioctl(struct inode *inode, struct file *filp,
	       unsigned int request, unsigned long arg)
{
	int retval = 0;		/* return value */

	/* Only administrator can use LKST.*/
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	switch (request) {
	case LKST_IOC_TRC_STATUS:
		retval = 
		    lkst_trc_status_ioctl((struct lkst_status_param *) arg);
		break;
	case LKST_IOC_TRC_START:
		retval = lkst_trc_start_ioctl();
		break;
	case LKST_IOC_TRC_STOP:
		retval = lkst_trc_stop_ioctl();
		break;
	case LKST_IOC_MASKSET_READ:
		retval =
		    lkst_maskset_read_ioctl((struct lkst_maskset_param *) arg);
		break;
	case LKST_IOC_MASKSET_WRITE:
		retval =
		   lkst_maskset_write_ioctl((struct lkst_maskset_param *) arg);
		break;
	case LKST_IOC_MASKSET_SET:
		retval = lkst_maskset_set_ioctl((lkst_maskset_id) arg);
		break;
	case LKST_IOC_MASKSET_LIST:
		retval =
		    lkst_maskset_list_ioctl((struct lkst_maskset_listparam *) arg);
		break;
	case LKST_IOC_MASKSET_DELETE:
		retval = lkst_maskset_delete_ioctl((lkst_maskset_id) arg);
		break;
	case LKST_IOC_EVHANDLER_LIST:
		retval =
		    lkst_evhandler_list_ioctl((struct lkst_evhandler_listparam *) arg);
		break;
	case LKST_IOC_EVHANDLER_CTRL:
		retval =
		    lkst_evhandler_ctrl_ioctl((struct lkst_evhandler_ctrl_param *) arg);
		break;
	case LKST_IOC_BUFFER_READ:
		retval = 
		    lkst_buffer_read_ioctl((struct lkst_log_buffer *) arg);
		break;
	case LKST_IOC_BUFFER_CREATE:
		retval =
		    lkst_buffer_create_ioctl((struct lkst_buffer_param *) arg);
		break;
	case LKST_IOC_BUFFER_SHIFT:
		retval = lkst_buffer_shift_ioctl((int)arg);
		break;
	case LKST_IOC_BUFFER_JUMP:
		retval = lkst_buffer_jump_ioctl((struct lkst_buffer_jumpparam *)arg);
		break;
	case LKST_IOC_BUFFER_DELETE:
		retval = lkst_buffer_delete_ioctl((struct lkst_buffer_delparam*) arg);
		break;
	case LKST_IOC_BUFFER_LIST:
		retval = lkst_buffer_list_ioctl((struct lkst_buffer_listparam *) arg);
		break;
	case LKST_IOC_BUFFER_LINK:
		retval = lkst_buffer_link_ioctl((struct lkst_buffer_linkparam *) arg);
		break;
	case LKST_IOC_BUFFER_SETRMOD:
		retval = lkst_buffer_setrmod_ioctl((struct lkst_buffer_srmodparam *) arg);
		break;

	case LKST_IOC_ENTRY_LOG:
		retval =
		    lkst_entry_log_ioctl((struct lkst_entry_args *) arg);
		break;
	case LKST_IOC_MASKSET_GETID:
		retval = lkst_maskset_get_id_ioctl(
				(struct lkst_maskset_getparam *) arg);
		break;
	case LKST_IOC_EVHANDLER_GETID:
		retval = lkst_evhandler_get_id_ioctl(
				(struct lkst_evhandler_getparam *) arg);
		break;
	default:
		LKST_DPRINTK("Invalid command no.%d\n", request);
		return -EINVAL;
	}

	return retval;
}

