/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2001-2002
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by K.HATASAKI <keisukeh@crl.hitachi.co.jp>
 *
 ************************************************************/

#ifndef _LINUX_LKST_BUFFER_H
#define _LINUX_LKST_BUFFER_H


#include <linux/lkst.h>
#ifdef __KERNEL__
#include <linux/types.h>
#include <linux/time.h>
#else /* #ifndef __KERNEL__ */
#include <pthread.h>
#include <sys/time.h>
#endif	/* __KERNEL__ */

		/* mode for ioctl(LKST_IOC_BUFFER_SETRMOD) command */
#define RAW	0x01
#define STD	0x02
#define SORTED	0x04

		/* size of lkst_event_record. It must be power of 2 */
#define LKST_SIZEOF_LKST_EVENT_RECORD	(64)

		/*  following 4 lines should be deleted
		 *  if they has been already defined. */
typedef int posix_log_recid_t;
typedef int posix_log_facility_t;
typedef int posix_log_severity_t;
typedef int posix_log_procid_t;

#ifdef __KERNEL__
typedef unsigned long int pthread_t;
#endif /*__KERNEL__ */


#ifndef PXLOG_BINARY
#define PXLOG_BINARY	1
#endif /* PXLOG_BINARY */

#ifndef LOG_KERN
#define LOG_KERN	0
#endif /* LOG_KERN */

#ifndef LOG_DEBUG
#define LOG_DEBUG	7
#endif /* LOG_DEBUG */


#define LOGFILE_MAGIC   0xbeefface
#define LOGREC_MAGIC    0xfeefface
#define LKST_LOG_MAGIC_NUM 0x76758384

#define LKST_ARCH_NAME_LEN 72

/* Logfile Header for evlog compatibility. */
typedef struct log_header {
	uint		log_magic;	/* Magic number indicating log file */
	long		log_version;	/* Event logging version */
	posix_log_recid_t last_recid;	/* Last recid in log: used by evlogd */
	off_t		reserved1;
	uint		log_generation;	/* Changes during log maintenance */
} log_header_t;

struct posix_log_entry {
	unsigned int log_magic;
	posix_log_recid_t log_recid;	/* ID of the event record */
	size_t log_size;	/* size of the event record variable data */
	int log_format;		/* format of variable data */
	int log_event_type;	/* event identification code */
	posix_log_facility_t log_facility;	/* event facility code */
	posix_log_severity_t log_severity;	/* event severity code */
	uid_t log_uid;	/* effective user ID associated with the event */
	gid_t log_gid;	/* effective group ID associated with the event */
	pid_t log_pid;	/* process ID associated with event */
	pid_t log_pgrp;	/* process group associated with event */
	struct timespec log_time;	/* event time stamp */
	unsigned int log_flags;		/* bitmap of event flag */
	pthread_t log_thread;		/* thread ID associated with event */
	posix_log_procid_t log_processor; /* processor ID associated with event */
};

struct lkst_log_record {
	struct posix_log_entry posix;	/* log form specified by POSIX */
	lkst_arg_t log_arg1;
			/* 1st argument acquired at a trace point */
	lkst_arg_t log_arg2;
			/* 2nd argument acquired at a trace point */
	lkst_arg_t log_arg3;
			/* 3rd argument acquired at a trace point */
	lkst_arg_t log_arg4;
			/* 4th argument acquired at a trace point */
};

struct lkst_log_buffer {
	int cpu;			/* cpu number */
	size_t read_size;		/* size of event records to read */
	size_t result_read_size;	/* size of read event records */
	struct timeval xtime;		/* xtime */
	lkst_tsc_t tsc;			/* machine cycle */ 
					/* These two will be used to calculate 
					 * time of events in real time. */
	lkst_cpu_freq_t	cpu_freq;	/* cpu clockspeed in kHz */ 
	struct lkst_log_record *buffer;
			/* address of a buffer to store event records */

	int endian_big;			/* byte order, 0 if little endian */
	int buf_ver;			/* LKST buffer version */
	char arch[LKST_ARCH_NAME_LEN];	/* Architecture name */
	lkst_buffer_id id;		/* event buffer ID */
};

struct lkst_log_buffer_header {
	struct log_header header;
	struct posix_log_entry entry;
	int endian_big;			/* byte order, 0 if little endian */
	int buf_ver;			/* LKST buffer version */
	char arch[LKST_ARCH_NAME_LEN];	/* Architecture name */	
};

struct lkst_buffer_param {
	int cpu;		/* cpu number */
	size_t size;		/* size of event buffer */
	size_t result_size;	/* result size of event buffer */
	lkst_buffer_id id;	/* event buffer ID */
	lkst_buffer_id next;	/* event buffer ID of next buffer */
};

struct lkst_buffer_listent {
	size_t size;			/* buffer size */
	int write_offset;		/* offset to write */
	int baseid;			/* base counter */
	lkst_buffer_id id, next;	/* buffer ID of own/next buffer */
	lkst_buffer_id rq_prev, rq_next;	/* buffer ID of read-queue previous/next buffer */
};

struct lkst_buffer_listparam {
	int cpu;			/* cpu number of the buffer */
	int listent_num;		/* num of the listent */
	int buffer_num;			/* num of the buffers owned by this CPU*/
	lkst_buffer_id write_buf;	/* current writing buffer */
	lkst_buffer_id read_buf;	/* current reading buffer */
	lkst_buffer_id rq_head;
	lkst_buffer_id rq_tail;
	pid_t owner;
	int read_pos;
	struct lkst_buffer_listent *listent;
				/* area to store the list of event buffers 
				 * this size must be lkstent_num*sizeof(struct lkst_buffer_listent)*/
};

struct lkst_buffer_linkparam {
	int cpu;			/* cpu number */
	lkst_buffer_id id, next;	/* link parameter */
};

struct lkst_buffer_delparam {
	int cpu;			/* cpu number */
	lkst_buffer_id id;		/* buffer id to be deleted */
};

struct lkst_buffer_jumpparam {
	int cpu;			/* cpu number */
	lkst_buffer_id dest;		/* destination buffer id */
};

struct lkst_buffer_srmodparam {
	int cpu;			/* cpu number */
	int mode;			/* read mode */
	struct timeval xtime;		/* xtime */
	lkst_tsc_t tsc;			/* machine cycle */ 
					/* These two will be used to calculate 
					 * time of events in real time. */
	lkst_cpu_freq_t	cpu_freq;	/* cpu clockspeed in kHz */ 
	int endian_big;			/* byte order, 0 if little endian */
	int buf_ver;			/* LKST buffer version */
	char arch[LKST_ARCH_NAME_LEN];	/* Architecture name */
};

#ifdef __KERNEL__

int lkst_buffer_shift(void);
int lkst_buffer_jump(lkst_buffer_id next);

/* !NEVER use these functions in event handlers nor in interrupts !*/
int lkst_buffer_create(lkst_buffer_id id, size_t size, lkst_buffer_id next);
int lkst_buffer_delete(lkst_buffer_id id);
int lkst_buffer_read(struct lkst_log_buffer *lbuffer);

#endif /*__KERNEL__*/

#define LKST_BUFFER_LISTENT_SIZE(n)					\
	     (sizeof(struct lkst_buffer_listent) * (n))

#define LKST_BUFFER_SIZE_LOCAL2EXT(size) 				\
		((size)*sizeof(struct lkst_log_record)/LKST_SIZEOF_LKST_EVENT_RECORD)

#endif				/* _LINUX_LKST_BUFFER_H */

/* END */
